unit MWKEditSpans;
      {******************************************************************}
      { MWKEditSpans                                                     }
      {                                                                  }
      { home page : http://www.mwcs.de                                   }
      { email     : martin.walter@mwcs.de                                }
      {                                                                  }
      { date      : 02-12-2007                                           }
      { version   : v1.3                                                 }
      {                                                                  }
      { Use of this file is permitted for commercial and non-commercial  }
      { use, as long as the original author is credited.                 }
      { This file (c) 2005, 2007 Martin Walter                           }
      {                                                                  }
      { This Software is distributed on an "AS IS" basis, WITHOUT        }
      { WARRANTY OF ANY KIND, either express or implied.                 }
      {                                                                  }
      { *****************************************************************}

interface

uses
  MSXML2_TLB_Light,
  Windows, Classes, Graphics;

const
  SUPERSUBFRAC = 0.5;
  
  cHyphen = ''; // #170
  cProtectSpace = ''; // #167
//  cTagSpace = ''; // #
  cNewParagraph = ''; // #244
  cNewLine = '';
  cTabChar = #142;

  cTab = #9;
  //WideChar($21B5);

  cUpperQuote = WideChar(8220);
  cLowerQuote = WideChar(8222);
  cLeftQuote = WideChar(8216);
  cRightQuote = WideChar(8217);

type
  TSpanValues = record
    IsBold, IsItalic, IsUnderline, IsStrikeOut, IsSpaced, IsSuper, IsSub,
    IsIndex, IsNewLine: Boolean;
    Color, BackColor: TColor;
    FontName: WideString;
    FontHeight: Currency;
    ShadowFontName: WideString;
    CharSet: Integer;
    Indent: Integer;
  end;

  TSpan = class(TObject)
  private
    FFontName: WideString;
    FShadowFontName: WideString;
    FFontHeight: Currency;
    FBold: Boolean;
    FItalic: Boolean;
    FUnderline: Boolean;
    FStrikeOut: Boolean;
    FSpaced: Boolean;
    FSuper: Boolean;
    FSub: Boolean;
    FIndex: Boolean;
    FColor: TColor;
    FBackColor: TColor;
    FSelStart: Integer;
    FSelLength: Integer;
    FSpells: array of Integer;
    FSpellCount: Integer;
    FText: WideString;
    FUpdating: Boolean;
    FNewLine: Boolean;
    FNewParagraph: Boolean;
    FLineBreak: Boolean;
    FIsTab: Boolean;
    FLength: Integer;
    FShowTags: Boolean;
    FShowSpellErrors: Boolean;
    FRightToLeft: Boolean;
    FIndent: Integer;

    procedure CalcWidth; virtual;

    function GetBackColor: TColor;
    function GetBold: Boolean;
    function GetColor: TColor;
    function GetFontName: WideString;
    function GetFontHeight: Currency;
    function GetIndex: Boolean;
    function GetItalic: Boolean;
    function GetWidth: Integer;
    function GetSelLength: Integer;
    function GetSelStart: Integer;
    function GetSpaced: Boolean;
    function GetStrikeOut: Boolean;
    function GetSub: Boolean;
    function GetSuper: Boolean;
    function GetText: WideString;
    function GetUnderline: Boolean;
    function GetUpdating: Boolean;
    function GetLength: Integer;
    function GetNewLine: Boolean;
    function GetNewParagraph: Boolean;
    function GetLineBreak: Boolean; virtual;
    function GetValueRec: TSpanValues;
    function GetRightToLeft: Boolean;
    function GetShowTags: Boolean;
    function GetShowSpellErrors: Boolean;
    function GetIsTab: Boolean;
    function GetIndent: Integer;
    function GetShadowFontName: WideString;
    function GetHeight: Integer;
    function GetBaseline: Integer;
    function GetStartFromBaseline: Integer;
    procedure SetBackColor(const Value: TColor);
    procedure SetBold(const Value: Boolean);
    procedure SetColor(const Value: TColor);
    procedure SetFontName(const Value: WideString);
    procedure SetFontHeight(const Value: Currency);
    procedure SetIndex(const Value: Boolean);
    procedure SetItalic(const Value: Boolean);
    procedure SetSelLength(const Value: Integer);
    procedure SetSelStart(const Value: Integer);
    procedure SetSpaced(const Value: Boolean);
    procedure SetStrikeOut(const Value: Boolean);
    procedure SetSub(const Value: Boolean);
    procedure SetSuper(const Value: Boolean);
    procedure SetText(const Value: WideString); virtual;
    procedure SetUnderline(const Value: Boolean);
    procedure SetUpdating(const Value: Boolean);
    procedure SetNewLine(const Value: Boolean);
    procedure SetNewParagraph(const Value: Boolean);
    procedure SetLineBreak(const Value: Boolean); virtual;
    procedure SetValueRec(const Value: TSpanValues);
    procedure SetRightToLeft(const Value: Boolean);
    procedure SetShowTags(const Value: Boolean);
    procedure SetShowSpellErrors(const Value: Boolean);
    procedure SetIndent(const Value: Integer);
    procedure SetWidth(const Value: Integer);
    procedure SetShadowFontName(const Value: WideString);
  protected
    FWidth: Integer;
    FHeight: Integer;
    FStartFromBaseline: Integer;
    FBaseline: Integer;
  public
    constructor Create;
    destructor Destroy; override;

    function CreateFont(const WithSuperSub: Boolean): HFont;
    function MeasureWidth(Width: Integer; WithLineBreak: Boolean): Integer; virtual;
    function CalcTextSize(const Font: HFont; const Text: WideString): TSize; virtual;
    function CalcTextWidth(const Text: WideString): Integer; virtual;
    function CalcBaseline(Font: HFont = 0): Integer; virtual;
    procedure SetTextWithoutChange(const Value: WideString); virtual;

    function LastSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; virtual;
    function NextSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; virtual;
    procedure DeleteSelected; virtual;
    function Part(Width: Integer): TSpan; virtual;
    function Split(Position: Integer): TSpan; virtual;
    procedure Assign(const Source: TSpan); virtual;
    procedure Paint(const Target: TCanvas; X, Y, LineHeight, Baseline: Integer;
      Focused: Boolean); virtual;
    function CanAdd(const Source: TSpan): Boolean; virtual;
    procedure Join(const Source: TSpan); virtual;
    function SameValues(const Rec: TSpanValues): Boolean; virtual;
    procedure ClearSpells;
    procedure AddSpell(const Start, Length: Integer);

    property RightToLeft: Boolean read GetRightToLeft write SetRightToLeft;
    property Width: Integer read GetWidth write SetWidth;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property Text: WideString read GetText write SetText;
    property FontName: WideString read GetFontName write SetFontName;
    property ShadowFontName: WideString read GetShadowFontName write SetShadowFontName;
    property FontHeight: Currency read GetFontHeight write SetFontHeight;
    property Bold: Boolean read GetBold write SetBold;
    property Italic: Boolean read GetItalic write SetItalic;
    property Underline: Boolean read GetUnderline write SetUnderline;
    property StrikeOut: Boolean read GetStrikeOut write SetStrikeOut;
    property Spaced: Boolean read GetSpaced write SetSpaced;
    property Color: TColor read GetColor write SetColor;
    property BackgroundColor: TColor read GetBackColor write SetBackColor;
    property Super: Boolean read GetSuper write SetSuper;
    property Sub: Boolean read GetSub write SetSub;
    property Index: Boolean read GetIndex write SetIndex;
    property IsTab: Boolean Read GetIsTab;
    property NewLine: Boolean read GetNewLine write SetNewLine;
    property LineBreak: Boolean read GetLineBreak write SetLineBreak;
    property NewParagraph: Boolean read GetNewParagraph write SetNewParagraph;
    property Values: TSpanValues read GetValueRec write SetValueRec;
    property Updating: Boolean read GetUpdating write SetUpdating;
    property Length: Integer read GetLength;
    property ShowTags: Boolean read GetShowTags write SetShowTags;
    property ShowSpellErrors: Boolean read GetShowSpellErrors write SetShowSpellErrors;
    property Indent: Integer read GetIndent write SetIndent;
    property Height: Integer read GetHeight;
    property DeltaY: Integer read GetStartFromBaseline;
    property Baseline: Integer read GetBaseline;
  published
  end;

  TGraphicSpan = class(TSpan)
  private
    FPicture: Pointer;
    FImageHeight: Integer;

    procedure SetText(const Value: WideString); override;

    procedure CalcWidth; override;
  public
    constructor Create;
    destructor Destroy; override;

    function MeasureWidth(Width: Integer; WithLineBreak: Boolean): Integer; override;
    function CalcTextSize(const Font: HFont; const Text: WideString): TSize; override;
    function CalcTextWidth(const Text: WideString): Integer; override;
    function GetAsText(WithMarker: Boolean): WideString;
    procedure SetTextWithoutChange(const Value: WideString); override;

    function LastSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; override;
    function NextSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; override;

    procedure DeleteSelected; override;
    function Part(Width: Integer): TSpan; override;
    function Split(Position: Integer): TSpan; override;
    procedure Assign(const Source: TSpan); override;
    procedure Paint(const Target: TCanvas; X, Y, LineHeight, Baseline: Integer;
      Focused: Boolean); override;
    function CanAdd(const Source: TSpan): Boolean; override;
    procedure Join(const Source: TSpan); override;
    function SameValues(const Rec: TSpanValues): Boolean; override;
    procedure SetPicture(Image: Pointer);
    function IsSVG: Boolean;
    function GetSVG: IXMLDOMDocument3;
    property Picture: Pointer read FPicture;
  published
  end;

  TSectionSpan = class(TSpan)
  private
    FDisplay: WideString;
    procedure SetText(const Value: WideString); override;
    procedure CalcWidth; override;
    function GetLineBreak: Boolean; override;
    procedure SetLineBreak(const Value: Boolean); override;
  public
    constructor Create;
    function LastSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; override;
    function NextSpace(const Text: WideString; const Pos: Integer;
      var Ch: WideChar): Integer; override;

    procedure SetTextWithoutChange(const Value: WideString); override;
    procedure DeleteSelected; override;
    function Part(Width: Integer): TSpan; override;
    function Split(Position: Integer): TSpan; override;
    procedure Assign(const Source: TSpan); override;
    procedure Paint(const Target: TCanvas; X, Y, LineHeight, Baseline: Integer;
      Focused: Boolean); override;
    function CanAdd(const Source: TSpan): Boolean; override;
    procedure Join(const Source: TSpan); override;
    function SameValues(const Rec: TSpanValues): Boolean; override;
  end;

  TSpanList = class(TList)
  private
    function GetItem(const Index: Integer): TSpan;
    procedure SetItem(const Index: Integer; const Value: TSpan);
  public
    procedure Clear; override;
    procedure Add(Span: TSpan);
    procedure Delete(Index: Integer);
    function First: TSpan;
    function Last: TSpan;
    property Items[const Index: Integer]: TSpan read GetItem write SetItem; default;
  end;


implementation

uses
  SysUtils, StrUtils, Math,
  MWKSysUtils, MWKEditGraphics;


function InvertColor(const Color: TColor): Integer;
begin
  Result := not (ColorToRGB(Color));
  Result := Result and $00FFFFFF;
end;

function Brighten(const Color: TColor; Amount: Double): Integer;
var
  RGB: TRGBQuad;
begin
  Move(Color, RGB, SizeOf(RGB));
  RGB.rgbBlue := Round(RGB.rgbBlue * (1 - Amount) + 255 * Amount);
  RGB.rgbGreen := Round(RGB.rgbGreen * (1 - Amount) + 255 * Amount);
  RGB.rgbRed := Round(RGB.rgbRed * (1 - Amount) + 255 * Amount);

  Move(RGB, Result, SizeOf(Result));
  Result := Result and $00FFFFFF;
end;

function Reverse(const S: WideString): WideString;
var
  C, L: Integer;
begin
  L := Length(S);
  SetLength(Result, L);
  for C := 0 to L - 1 do
    Result[L - C] := S[C + 1];
end;               

{ TSpan }

constructor TSpan.Create;
begin
  inherited;
  FSelStart := -1;
  FSelLength := 0;
  FSpellCount := 0;
end;

procedure TSpan.DeleteSelected;
begin
  if FSelLength > 0 then
  begin
    if FSelLength = FLength then
      FText := ''
    else
      FText := Copy(FText, 1, FSelStart - 1) + Copy(FText, FSelStart + FSelLength, MaxInt);
    FLength := System.Length(FText);
    if (FNewParagraph or FNewLine) then
    begin
      if (Length = 0) or
         (FNewLine and (FText[Length] <> #14)) or
         (FNewParagraph and (FText[Length] <> #13)) then
      begin
        FNewParagraph := False;
        FNewLine := False;
      end;
    end;

    FSelLength := 0;
    if not FUpdating then
      CalcWidth;
  end;
end;

function TSpan.Part(Width: Integer): TSpan;
var
  Space: Integer;
  Ch: WideChar;
  Font, OldFont: HFont;
  DC: HDC;
  Fit: Integer;
  Size: TSize;
begin
  Result := nil;
  if Width >= FWidth then
    Exit;


  Font := CreateFont(True);
  if Font = 0 then
  begin
    Result := nil;
    Exit;
  end;
  DC := GetDC(0);
  OldFont := SelectObject(DC, Font);

  if FNewLine or FNewParagraph then
    GetTextExtentExPointW(DC, PWideChar(FText), FLength - 1, Width, @Fit, nil, Size)
  else
    GetTextExtentExPointW(DC, PWideChar(FText), FLength, Width, @Fit, nil, Size);

  SelectObject(DC, OldFont);
  DeleteObject(Font);
  ReleaseDC(0, DC);

  if Fit = 0 then
  begin
    Result := Pointer(1); // Span passt nicht in die Zeile
    Exit;
  end;

  Space := LastSpace(FText, Fit, Ch);

  if Space = 0 then
  begin
    Result := Split(Fit + 1);
    FNewLine := False;
    FNewParagraph := False;
    FLineBreak := True;
    Exit;
  end;

  while (Space < FLength) and (FText[Space + 1] = ' ') do
    Inc(Space);

  Result := Split(Space + 1);
  FNewLine := False;
  FNewParagraph := False;
  FLineBreak := True;
end;

function TSpan.Split(Position: Integer): TSpan;
begin
  Result := nil;
  if (Position < 2) or (Position > Length) then
    Exit;

  Result := TSpan.Create;
  Result.Assign(Self);
  Result.SetTextWithoutChange(Copy(FText, Position, MaxInt));
  Text := Copy(FText, 1, Position - 1);
  FLineBreak := False;

  if (FSelLength = 0) then
    Exit;

  if (FSelStart + FSelLength - 1 <= FLength) then
  begin
    Result.SelStart := 0;
    Result.SelLength := 0;
    Exit;
  end;

  if FSelStart > FLength then
  begin
    FSelStart := 0;
    FSelLength := 0;
    Result.FSelStart := Result.FSelStart - FLength;
    Exit;
  end;

  Result.FSelStart := 1;
  Result.FSelLength := FSelLength - (FLength - FSelStart + 1);

  FSelLength := FLength - FSelStart + 1;
end;

procedure TSpan.Assign(const Source: TSpan);
begin
  if Source = nil then
    Exit;

  FRightToLeft := Source.RightToLeft;
  FFontName := Source.FontName;
  FShadowFontName := Source.ShadowFontName;
  FFontHeight := Source.FontHeight;
  FBold := Source.Bold;
  FItalic := Source.Italic;
  FUnderline := Source.Underline;
  FStrikeOut := Source.StrikeOut;
  FSpaced := Source.Spaced;
  FSuper := Source.Super;
  FSub := Source.Sub;
  FIndex := Source.Index;
  FColor := Source.Color;
  FBackColor := Source.BackgroundColor;
  FText := Source.Text;
  FLength := Source.Length;
  FNewLine := Source.NewLine;
  FNewParagraph := Source.NewParagraph;
  FLineBreak := Source.LineBreak;
  FSelStart := Source.SelStart;
  FSelLength := Source.SelLength;
  FShowTags := Source.ShowTags;
  FShowSpellErrors := Source.ShowSpellErrors;
  FIsTab := Source.IsTab;
  FIndent := Source.Indent;
  if not FUpdating then
    CalcWidth;
end;

destructor TSpan.Destroy;
begin
  ClearSpells;
  inherited;
end;

function TSpan.LastSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Result := Min(Pos, Length);
  while Result > 0 do
  begin
    if (Text[Result] = ' ') or
       (Text[Result] = cProtectSpace) or
       (Text[Result] = cTab) or
       (Text[Result] = #13) or
       (Text[Result] = #14) or
       (Text[Result] = '-') or
       (Text[Result] = cHyphen) or
       (Text[Result] = '+') or
       (Text[Result] = '/') or
       (Text[Result] = '*') or
       (Text[Result] = '&') or
       (Text[Result] = '(') or
       (Text[Result] = ')') or
       (Text[Result] = '{') or
       (Text[Result] = '}') or
       (Text[Result] = '[') or
       (Text[Result] = ']') or
       (Text[Result] = ',') or
       (Text[Result] = '.') or
       (Text[Result] = ';') or
       (Text[Result] = ':') or
       (Text[Result] = '!') or
       (Text[Result] = '?') or
       (Text[Result] = '\') or
       (Text[Result] = '~') or
       (Text[Result] = '"') or
       (Text[Result] = cLowerQuote) or
       (Text[Result] = cUpperQuote) or
       (Text[Result] = '''') or
       (Text[Result] = '_') or
       (Text[Result] = '<') or
       (Text[Result] = '>') or
       (Text[Result] = '=') then
    begin
      Ch := Text[Result];
      Exit;
    end;
    Dec(Result);
  end;
end;

function TSpan.NextSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Ch := #0;
  Result := Max(1, Min(Pos, Length));
  while Result <= Length do
  begin
    if (Text[Result] = ' ') or
       (Text[Result] = cProtectSpace) or
       (Text[result] = cTab) or
       (Text[Result] = #13) or
       (Text[Result] = #14) or
       (Text[Result] = '-') or
       (Text[Result] = cHyphen) or
       (Text[Result] = '+') or
       (Text[Result] = '/') or
       (Text[Result] = '*') or
       (Text[Result] = '&') or
       (Text[Result] = '(') or
       (Text[Result] = ')') or
       (Text[Result] = '{') or
       (Text[Result] = '}') or
       (Text[Result] = '[') or
       (Text[Result] = ']') or
       (Text[Result] = ',') or
       (Text[Result] = '.') or
       (Text[Result] = ';') or
       (Text[Result] = ':') or
       (Text[Result] = '!') or
       (Text[Result] = '?') or
       (Text[Result] = '\') or
       (Text[Result] = '~') or
       (Text[Result] = '"') or
       (Text[Result] = cLowerQuote) or
       (Text[Result] = cUpperQuote) or
       (Text[Result] = '_') or
       (Text[Result] = '<') or
       (Text[Result] = '>') or
       (Text[Result] = '=') then
    begin
      Ch := Text[Result];
      Exit;
    end;
    Inc(Result);
  end;
end;

function TSpan.CreateFont(const WithSuperSub: Boolean): HFont;
var
  LogFont: TLogFontW;
  FontName: WideString;
begin
  FillChar(LogFont, SizeOf(LogFont), 0);
  if (FSub or FSuper) and WithSuperSub then
    LogFont.lfHeight := Round(FFontHeight * SUPERSUBFRAC)
  else
    LogFont.lfHeight := Round(FFontHeight);
  LogFont.lfWidth := 0;
  LogFont.lfEscapement := 0;
  LogFont.lfOrientation := 0;
  if FBold then
    LogFont.lfWeight := FW_BOLD
  else
    LogFont.lfWeight := FW_NORMAL;
  LogFont.lfItalic := Byte(FItalic);
  LogFont.lfUnderline := Byte(FUnderline);
  LogFont.lfStrikeOut := Byte(FStrikeOut);
  logfont.lfCharSet := DEFAULT_CHARSET;
  LogFont.lfOutPrecision := OUT_DEFAULT_PRECIS;
  LogFont.lfClipPrecision := CLIP_DEFAULT_PRECIS;
  LogFont.lfQuality := DEFAULT_QUALITY;
  LogFont.lfPitchAndFamily := DEFAULT_PITCH or FF_DONTCARE;
  if FIsTab then
    FontName := 'Wingdings 3'
  else
  begin
    if FShadowFontName <> '' then
      FontName := FShadowFontName
    else
      FontName := FFontName;
  end;

  if FontName = '' then
    FontName := 'Arial';
  Move(FontName[1], LogFont.lfFaceName, System.Length(FontName) * SizeOf(WideChar));
  LogFont.lfFaceName[System.Length(FontName)] := #0;
  Result := CreateFontIndirectW(LogFont);
end;

function TSpan.CalcBaseline(Font: HFont): Integer;
var
  OldFont: HFont;
  DC: HDC;
  Metric: TTextMetricW;
  CreateNewFont: Boolean;
begin
  CreateNewFont := Font = 0;
  if CreateNewFont then
    Font := CreateFont(False);
  if Font = 0 then
  begin
    Result := 0;
    Exit;
  end;
  DC := GetDC(0);
  OldFont := SelectObject(DC, Font);
  GetTextMetricsW(DC, Metric);
  SelectObject(DC, OldFont);
  ReleaseDC(0, DC);

  if CreateNewFont then
    DeleteObject(Font);
  Result := Metric.tmAscent;
end;

function TSpan.CalcTextSize(const Font: HFont; const Text: WideString): TSize;
var
  OldFont: HFont;
  DC: HDC;
  Size: TSize;
  OldExtra: Integer;
  L: Integer;
  T: WideString;
  Empty: Boolean;
begin
  DC := GetDC(0);
  if FSpaced then
  begin
    if FFontHeight < 0 then
      OldExtra := SetTextCharacterExtra(DC, -Round(FFontHeight / 3))
    else
      OldExtra := SetTextCharacterExtra(DC, Round(FFontHeight / 3));
  end else
    OldExtra := 0;
  OldFont := SelectObject(DC, Font);
  L := System.Length(Text);
  if (L > 0) and ((Text[L] = #13) or (Text[L] = #14)) then
    Dec(L);

  T := Copy(Text, 1, L);
  Empty := T = '';
  if Empty then
    T := 'Mp';

  if (T[1] = #13) or
     (T[1] = #14) then
    T := Copy(T, 2, MaxInt);

  if not FShowTags then
  begin
    T := StringReplace(T, cHyphen, '');
    T := StringReplace(T, cProtectSpace, ' ');
  end;

  GetTextExtentPoint32W(DC, PWideChar(T), System.Length(T), Size);

  SelectObject(DC, OldFont);

  if FSpaced then
    SetTextCharacterExtra(DC, OldExtra);
  ReleaseDC(0, DC);

  if Empty then
    Size.cx := 0;
  Result := Size;
end;

function TSpan.CalcTextWidth(const Text: WideString): Integer;
var
  Font: HFont;
  Size: TSize;
  SuperSubHeight: Integer;
begin
  Font := CreateFont(True);
  if Font = 0 then
  begin
    Result := 0;
    Exit;
  end;

  if FRightToLeft then
    Size := CalcTextSize(Font, Reverse(Text))
  else
    Size := CalcTextSize(Font, Text);

  FBaseline := CalcBaseline(Font);
  DeleteObject(Font);
  Result := Size.cx;
  FHeight := Size.cy;
  FStartFromBaseline := -FBaseline;
  SuperSubHeight := FHeight;

  if Super or Sub then
  begin
    if Sub then
      FStartFromBaseline := -SuperSubHeight div 2;
    Font := CreateFont(False);

    if FRightToLeft then
      Size := CalcTextSize(Font, Reverse(Text))
    else
      Size := CalcTextSize(Font, Text);
    FBaseline := CalcBaseline(Font);
    DeleteObject(Font);
    FHeight := Size.cy;

    if Sub then
    begin
      if FBaseline + FStartFromBaseline + SuperSubHeight > FHeight then
        FStartFromBaseline := FHeight - SuperSubHeight - FBaseline;
    end;

    if Super then
      FStartFromBaseline := - FBaseline;
  end;

  if FIsTab then
    Result := FWidth;
end;

procedure TSpan.CalcWidth;
begin
  FWidth := CalcTextWidth(FText);
end;

function TSpan.GetBackColor: TColor;
begin
  Result := FBackColor;
end;

function TSpan.GetBaseline: Integer;
begin
  Result := FBaseline;
end;

function TSpan.GetBold: Boolean;
begin
  Result := FBold;
end;

function TSpan.GetColor: TColor;
begin
  Result := FColor;
end;

function TSpan.GetStartFromBaseline: Integer;
begin
  Result := FStartFromBaseline;
end;

function TSpan.GetFontName: WideString;
begin
  Result := FFontName;
end;

function TSpan.GetHeight: Integer;
begin
  Result := FHeight;
end;

function TSpan.GetFontHeight: Currency;
begin
  Result := FFontHeight;
end;

function TSpan.GetIndent: Integer;
begin
  Result := FIndent;
end;

function TSpan.GetIndex: Boolean;
begin
  Result := FIndex;
end;

function TSpan.GetIsTab: Boolean;
begin
  Result := FIsTab;
end;

function TSpan.GetItalic: Boolean;
begin
  Result := FItalic;
end;

function TSpan.GetWidth: Integer;
begin
  Result := FWidth;
end;

function TSpan.GetSelLength: Integer;
begin
  Result := FSelLength;
end;

function TSpan.GetSelStart: Integer;
begin
  Result := FSelStart;
end;

function TSpan.GetSpaced: Boolean;
begin
  Result := FSpaced;
end;

function TSpan.GetStrikeOut: Boolean;
begin
  Result := FStrikeOut;
end;

function TSpan.GetSub: Boolean;
begin
  Result := FSub;
end;

function TSpan.GetSuper: Boolean;
begin
  Result := FSuper;
end;

function TSpan.GetText: WideString;
begin
  Result := FText;
end;

function TSpan.GetUnderline: Boolean;
begin
  Result := FUnderline;
end;

function TSpan.GetUpdating: Boolean;
begin
  Result := FUpdating;
end;

function TSpan.GetLength: Integer;
begin
  Result := FLength;
end;

procedure TSpan.Paint(const Target: TCanvas; X, Y, LineHeight, Baseline: Integer;
  Focused: Boolean);
var
  DC: HDC;
  Text: WideString;
  Top, Bottom: Integer;
  StartY: Integer;

  OldExtra: Integer;
  Font, OldFont: HFont;

  procedure Prepare;
  begin
    DC := Target.Handle;
    Text := FText;

    if FShowTags then
    begin
      if FNewLine then
        Text[Length] := cNewLine;

      if FIsTab then
        Text := cTabChar;

      if FNewParagraph then
        Text[Length] := cNewParagraph;
    end;

    if FRightToLeft then
    begin
      Text := Reverse(Text);
      if FNewParagraph and FShowTags then
        X := X - CalcTextWidth(cNewParagraph);
      if FNewLine and FShowTags then
        X := X - CalcTextWidth(cNewLine);
    end;

    Top := Y;
    Bottom := Y + LineHeight;
    StartY := Top + Baseline - FBaseline;

    if FSub then
    begin
      StartY := Top + Baseline - Round(FHeight * SUPERSUBFRAC / 2);
      if StartY + FHeight * SUPERSUBFRAC >= Bottom then
        StartY := Bottom - Round(FHeight * SUPERSUBFRAC) - 1;
    end;

    SetTextColor(DC, ColorToRGB(FColor));
    SetBkColor(DC, ColorToRGB(FBackColor));
    SetBkMode(DC, OPAQUE);
  end;

  function SetShowTags(const S: WideString): WideString;
  begin
    Result := S;
    if not FShowTags then
    begin
      Result := StringReplace(Result, cHyphen, '');
      Result := StringReplace(Result, cProtectSpace, ' ');
      Result := StringReplace(Result, cNewParagraph, '');
      Result := StringReplace(Result, cNewLine, '');
      Result := StringReplace(Result, cTabChar, '');
    end;
  end;

  procedure PaintSuperSubBackground(Color, Left, Width: Integer);
  var
    Y1, Y2: Integer;
    TextHeight: Integer;
    Brush: HBrush;
  begin
    TextHeight := Round(FHeight * SUPERSUBFRAC);
    if FSuper then
    begin
      Y1 := Top + TextHeight;
      Y2 := Top + LineHeight;
    end
    else
    begin
      Y1 := Top;
      Y2 := Top + Baseline - TextHeight div 2;
    end;

    Brush := CreateSolidBrush(Color);
    try
      FillRect(DC, Rect(Left, Y1, Left + Width, Y2), Brush);
      if FSub then
      begin
        Y1 := Y2;
        Y2 := Bottom;
        FillRect(DC, Rect(Left, Y1, Left + Width, Y2), Brush);
      end;
    finally
      DeleteObject(Brush);
    end;
  end;

  function PaintTabBackground(TabWidth: Integer; Color: Integer): Integer;
  var
    TabSpace: Integer;
    Brush: HBrush;
  begin
    TabSpace := (FWidth - TabWidth) div 2;
    if (TabSpace > 0) or (not FShowTags) then
    begin
      Brush := CreateSolidBrush(Color);
      try
        if FShowTags then
        begin
          FillRect(DC, Rect(X, Top, X + TabSpace, Bottom - 1), Brush);
          FillRect(DC, Rect(X + TabWidth + TabSpace, Top, X + FWidth, Bottom - 1), Brush);
          FrameRect(DC, Rect(X, Y + LineHeight, X + FWidth, Bottom), Brush);
        end else
          FillRect(DC, Rect(X, Top, X + FWidth, Bottom), Brush);
      finally
        DeleteObject(Brush);
      end;
      Result := TabSpace;
    end else
      Result := 0;
  end;

  function PaintText(const S: WideString; const Left: Integer;
    const Color, BackColor: Integer): Integer;
  var
    TabSpace: Integer;
    Size: TSize;
    Width: Integer;
    Text: WideString;
    Distance: Integer;
    Brush: HBrush;
  begin
    Text := SetShowTags(S);

    Size := CalcTextSize(Font, Text);
    Width := Size.cx;


    if FSuper or FSub then
      PaintSuperSubBackground(BackColor, Left, Width);

    if FIsTab then
      TabSpace := PaintTabBackground(Width, BackColor)
    else
      TabSpace := 0;

{    Distance := FHeight - Size.cy;
    if Distance > 0 then
    begin
      Brush := CreateSolidBrush(BackColor);
      FillRect(DC, Rect(Left + TabSpace, Top,
        Left + TabSpace + Width, Bottom), Brush);
      DeleteObject(Brush);
    end;}

    Distance := LineHeight - FHeight;
    if Distance > 0 then
    begin
      Brush := CreateSolidBrush(BackColor);
      FillRect(DC, Rect(Left + TabSpace, Top,
        Left + TabSpace + Width, StartY), Brush);

      FillRect(DC, Rect(Left + TabSpace, StartY + FHeight,
        Left + TabSpace + Width, Bottom), Brush);
      DeleteObject(Brush);
    end;

    
    SetTextColor(DC, Color);
    SetBkColor(DC, BackColor);
    TextOutW(DC, Left + TabSpace, StartY, PWideChar(Text), System.Length(Text));
    Result := Width;
  end;

  function PaintNonSelected(const S: WideString; const X: Integer): Integer;
  begin
    Result := PaintText(S, X, ColorToRGB(FColor), ColorToRGB(FBackColor));
  end;

  function PaintSelected(const S: WideString; const X: Integer): Integer;
  begin
    if Focused then
      Result := PaintText(S, X, InvertColor(FColor), InvertColor(FBackColor))
    else
      Result := PaintText(S, X, InvertColor(FColor), Brighten(InvertColor(FBackColor), 0.7));
  end;

  procedure PaintWithSelection;
  var
    Part: WideString;
    PosX: Integer;
  begin
    PosX := X;
    // Paint non selected start
    if FRightToLeft then
      Part := Copy(Text, 1, Length - SelStart + 1 - SelLength)
    else
      Part := Copy(Text, 1, SelStart - 1);

    if Part <> '' then
      PosX := PosX + PaintNonSelected(Part, PosX);

    // Paint selected part
    if FRightToLeft then
      Part := Copy(Text, Length - SelStart + 1 - SelLength + 1, SelLength)
    else
      Part := Copy(Text, SelStart, SelLength);

    if Part <> '' then
      PosX := PosX + PaintSelected(Part, PosX);

    // Paint non selected end
    if FRightToLeft then
      Part := Copy(Text, Length - SelStart + 2, MaxInt)
    else
      Part := Copy(Text, SelStart + SelLength, MaxInt);

    if Part <> '' then
      PaintNonSelected(Part, PosX);
  end;

  procedure PaintIndex;
  var
    P: TPoint;
    Pen, OldPen: HPen;
  begin
    if FIndex then
    begin
      Pen := CreatePen(PS_DOT, 1, ColorToRGB(clBlue));
      OldPen := SelectObject(DC, Pen);

      MoveToEx(DC, X, Baseline, @P);
      LineTo(DC, X + FWidth, Baseline);

      SelectObject(DC, OldPen);
      DeleteObject(Pen);
    end;
  end;

  procedure PaintSpellErrors;
  var
    C, D: Integer;
    SpellStart: Integer;
    SpellLength: Integer;
    Start, Width, Height: Integer;
    Left: Integer;
    YPos, DeltaY: Integer;
  begin
    if FShowSpellErrors then
    for C := 0 to FSpellCount div 2 - 1 do
    begin
      SpellStart := FSpells[C * 2];
      SpellLength := FSpells[C * 2 + 1];

      Start := CalcTextSize(Font, Copy(FText, 1, SpellStart - 1)).cx;
      Width := CalcTextSize(Font, Copy(FText, SpellStart, SpellLength)).cx;
      if Super or Sub then
        Height := CalcBaseline(Font)
      else
        Height := FBaseline;

      Left := X + Start;

      YPos := Left mod 4;
      if YPos < 2 then
        DeltaY := 1
      else
        DeltaY := -1;

      if YPos > 2 then
        YPos := YPos mod 2;

      for D := 0 to Width - 1 do
      begin
        Inc(YPos, DeltaY);
        if (YPos = 0) or (YPos = 2) then
          DeltaY := -DeltaY;
        SetPixel(DC, Left, StartY + Height + YPos, ColorToRGB(clRed));
        Inc(Left);
      end;
    end;
  end;

  procedure PaintSpaceTags;
  var
    Brush, SelBrush, OldBrush: HBrush;
    Pen, SelPen, OldPen: HPen;
    PosX, PosY, Left: Integer;
    PosSpace: Integer;
    SpaceWidth: Integer;
    Radius: Double;
    Size: TSize;
  begin
    if FShowTags then
    begin
      PosSpace := Pos(' ', Text);
      if PosSpace > 0 then
      begin
        Size := CalcTextSize(Font, ' ');
        SpaceWidth := Size.cx;
        Radius := Size.cy / 15;
        PosX := SpaceWidth div 2 - Round(Radius);
        PosY := StartY + FHeight div 2 - Round(Radius);

        Brush := CreateSolidBrush(ColorToRGB(FColor));
        SelBrush := CreateSolidBrush(InvertColor(FColor));
        Pen := CreatePen(PS_SOLID, 1, ColorToRGB(FColor));
        SelPen := CreatePen(PS_SOLID, 1, InvertColor(FColor));
        OldBrush := SelectObject(DC, Brush);
        OldPen := SelectObject(DC, Pen);
        while PosSpace > 0 do
        begin
          if (PosSpace < SelStart) or (PosSpace > SelStart + SelLength - 1) then
          begin
            SelectObject(DC, Brush);
            SelectObject(DC, Pen);
          end
          else
          begin
            SelectObject(DC, SelBrush);
            SelectObject(DC, SelPen);
          end;

          Left := X + CalcTextSize(Font, Copy(Text, 1, PosSpace - 1)).cx + PosX;

          Ellipse(DC, Left, PosY, Round(Left + Radius * 2), Round(PosY + Radius * 2));

          PosSpace := PosEx(' ', Text, PosSpace + 1);
        end;
        SelectObject(DC, OldPen);
        SelectObject(DC, OldBrush);
        DeleteObject(SelPen);
        DeleteObject(Pen);
        DeleteObject(SelBrush);
        DeleteObject(Brush);
      end;
    end;
  end;

begin
  Prepare;

  Font := CreateFont(True);
  if Font = 0 then
    Exit;

  OldFont := SelectObject(DC, Font);
  try
    if FSpaced then
    begin
      if FFontHeight < 0 then
        OldExtra := SetTextCharacterExtra(DC, -Round(FFontHeight / 3))
      else
        OldExtra := SetTextCharacterExtra(DC, Round(FFontHeight / 3));
    end else
      OldExtra := 0;

    if FSelLength = 0 then
      PaintNonSelected(Text, X)
    else
      PaintWithSelection;

    PaintIndex;

    PaintSpellErrors;

    PaintSpaceTags;

    if FSpaced then
      SetTextCharacterExtra(DC, OldExtra);
  finally
    SelectObject(DC, OldFont);
    DeleteObject(Font);
  end;
end;

function TSpan.CanAdd(const Source: TSpan): Boolean;
begin
  Result := (not (FNewLine or FNewParagraph)) and
    (Source.FFontName = FFontName) and
    (Source.FShadowFontName = FShadowFontName) and
    (Source.FFontHeight = FFontHeight) and
    (Source.FBold = FBold) and
    (Source.FItalic = FItalic) and
    (Source.FUnderline = FUnderline) and
    (Source.FStrikeOut = FStrikeOut) and
    (Source.FSpaced = FSpaced) and
    (Source.FSuper = FSuper) and
    (Source.FSub = FSub) and
    (Source.FIndex = FIndex) and
    (Source.FColor = FColor) and
    (Source.FBackColor = FBackColor) and
    (not (Source.IsTab or IsTab)) and
    (not ((Source is TGraphicSpan) or (Self is TGraphicSpan))) and
    (not ((Source is TSectionSpan) or (Self is TSectionSpan)));
end;

procedure TSpan.Join(const Source: TSpan);
var
  OldLength: Integer;
begin
  FLineBreak := False;
  OldLength := Length;
  FText := FText + Source.FText;
  FLength := System.Length(FText);
  if FSelStart + FSelLength - 1 = OldLength then
    Inc(FSelLength, Source.FSelLength);

  if (FSelLength = 0) and (Source.FSelStart > 0) and (Source.FSelLength > 0) then
  begin
    FSelStart := OldLength + Source.FSelStart;
    FSelLength := Source.FSelLength;
  end;
  FNewLine := Source.NewLine;
  FNewParagraph := Source.NewParagraph;
  FLineBreak := Source.FLineBreak;
  CalcWidth;
end;

function TSpan.SameValues(const Rec: TSpanValues): Boolean;
begin
  Result :=
    (FBold = Rec.IsBold) and
    (FItalic = Rec.IsItalic) and
    (FUnderline = Rec.IsUnderline) and
    (FStrikeOut = Rec.IsStrikeOut) and
    (FSpaced = Rec.IsSpaced) and
    (FSuper = Rec.IsSuper) and
    (FSub = Rec.IsSub) and
    (FIndex = Rec.IsIndex) and
    (FColor = Rec.Color) and
    (FBackColor = Rec.BackColor) and
    (FFontName = Rec.FontName) and
    (FFontHeight = Rec.FontHeight) and
    (FShadowFontName = Rec.ShadowFontName) and
    (FIndent = Rec.Indent);
end;

procedure TSpan.ClearSpells;
begin
  SetLength(FSpells, 0);
  FSpellCount := 0;
end;

procedure TSpan.AddSpell(const Start, Length: Integer);
begin
  SetLength(FSpells, FSpellCount + 2);
  FSpells[FSpellCount] := Start;
  FSpells[FSpellCount + 1] := Length;
  Inc(FSpellCount, 2);
end;

procedure TSpan.SetBackColor(const Value: TColor);
begin
  FBackColor := Value;
end;

procedure TSpan.SetBold(const Value: Boolean);
begin
  if FBold = Value then
    Exit;
  FBold := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetColor(const Value: TColor);
begin
  FColor := Value;
end;

procedure TSpan.SetFontName(const Value: WideString);
begin
  if FFontName = Value then
    Exit;
  FFontName := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetFontHeight(const Value: Currency);
begin
  if FFontHeight = Value then
    Exit;
  FFontHeight := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetIndent(const Value: Integer);
begin
  if Value >= 0 then
    FIndent := Value;
end;

procedure TSpan.SetIndex(const Value: Boolean);
begin
  FIndex := Value;
end;

procedure TSpan.SetItalic(const Value: Boolean);
begin
  if FItalic = Value then
    Exit;
  FItalic := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetSelLength(const Value: Integer);
begin
  if Value < 0 then
    Exit;
  FSelLength := Value;
end;

procedure TSpan.SetSelStart(const Value: Integer);
begin
  FSelStart := Value;
end;

procedure TSpan.SetSpaced(const Value: Boolean);
begin
  if FSpaced = Value then
    Exit;
  FSpaced := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetStrikeOut(const Value: Boolean);
begin
  FStrikeOut := Value;
end;

procedure TSpan.SetSub(const Value: Boolean);
begin
  if FSub = Value then
    Exit;
  FSub := Value;
  FSuper := False;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetSuper(const Value: Boolean);
begin
  if FSuper = Value then
    Exit;
  FSuper := Value;
  FSub := False;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetText(const Value: WideString);
begin
  if FText = Value then
    Exit;
  FText := Value;
  FLength := System.Length(FText);

  FNewLine := False;
  FNewParagraph := False;
  if Value <> '' then
  begin
    if (Value[FLength] = #13) then
      FNewParagraph := True;

    if (Value[FLength] = #14) then
      FNewLine := True;

    if FNewLine or FNewParagraph then
      FLineBreak := True;
  end;

  FIsTab := Value = cTab;

  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetUnderline(const Value: Boolean);
begin
  FUnderline := Value;
end;

procedure TSpan.SetUpdating(const Value: Boolean);
begin
  FUpdating := Value;
end;

function TSpan.GetNewLine: Boolean;
begin
  Result := FNewLine;
end;

function TSpan.GetNewParagraph: Boolean;
begin
  Result := FNewParagraph;
end;

procedure TSpan.SetNewLine(const Value: Boolean);
begin
  FNewLine := Value;
  if Value then
  begin
    FNewParagraph := False;
    FLineBreak := True;
  end;
end;

procedure TSpan.SetNewParagraph(const Value: Boolean);
begin
  FNewParagraph := Value;
  if FNewParagraph then
  begin
    FNewLine := False;
    FLineBreak := True;
  end;
end;

function TSpan.GetLineBreak: Boolean;
begin
  Result := FLineBreak;
end;

procedure TSpan.SetLineBreak(const Value: Boolean);
begin
  FLineBreak := Value;
end;

function TSpan.GetValueRec: TSpanValues;
begin
  Result.IsBold := FBold;
  Result.IsItalic := FItalic;
  Result.IsUnderline := FUnderline;
  Result.IsStrikeOut := FStrikeOut;
  Result.IsSpaced := FSpaced;
  Result.IsSuper := FSuper;
  Result.IsSub := FSub;
  Result.IsIndex := FIndex;
  Result.IsNewLine := FNewLine;
  Result.Color := FColor;
  Result.BackColor := FBackColor;
  Result.FontName := FFontName;
  Result.FontHeight := FFontHeight;
  Result.ShadowFontName := FShadowFontName;
  Result.Indent := FIndent;
end;

procedure TSpan.SetValueRec(const Value: TSpanValues);
begin
  FBold := Value.IsBold;
  FItalic := Value.IsItalic;
  FUnderline := Value.IsUnderline;
  FStrikeOut := Value.IsStrikeOut;
  FSpaced := Value.IsSpaced;
  FSuper := Value.IsSuper;
  FIndent := Value.Indent;
  if FSuper then
    FSub := False;
  FSub := Value.IsSub;
  if FSub then
    FSuper := False;
  FIndex := Value.IsIndex;
  FNewLine := Value.IsNewLine;
  FColor := Value.Color;
  FBackColor := Value.BackColor;
  FFontName := Value.FontName;
  FFontHeight := value.FontHeight;
  FShadowFontName := Value.ShadowFontName;
  If FNewLine then
    FLineBreak := True;

  if not Updating then
    CalcWidth;
end;

procedure TSpan.SetWidth(const Value: Integer);
begin
  FWidth := Value;
end;

function TSpan.MeasureWidth(Width: Integer; WithLineBreak: Boolean): Integer;
var
  Laenge: Integer;
  Position, LastPos: Integer;
  Text: WideString;
  Font: HFont;
begin
  if Width >= FWidth then
  begin
    if WithLineBreak or (not (FNewLine or FNewParagraph)) then
      Result := FLength + 1
    else
      Result := FLength;
    Exit;
  end;

  if Width = 0 then
  begin
    Result := 1;
    Exit;
  end;

  Font := CreateFont(True);
  if Font = 0 then
  begin
    Result := 0;
    Exit;
  end;

  LastPos := FLength;
  Position := FLength div 2;
  repeat
    Text := Copy(FText, 1, Position);
    Laenge := CalcTextSize(Font, Text).cx;
    if Laenge < Width then
      Position := (LastPos + Position) div 2;
    if Laenge > Width then
    begin
      LastPos := Position;
      Position := Position div 2;
    end;
    if Laenge = Width then
      Break;

  until Abs(Position - LastPos) < 2;

  Text := Copy(FText, 1, Position);
  Laenge := CalcTextSize(Font, Text).cx;
  if Laenge < Width then
    Result := LastPos
  else
    Result := Position;

  Text := Copy(FText, 1, Result - 1);
  Laenge := CalcTextSize(Font, Text).cx;

  Dec(Width, Laenge);
  Laenge := CalcTextSize(Font, FText[Result]).cx;

  if Width >= Laenge div 2 then
    Inc(Result);

  DeleteObject(Font);
end;

function TSpan.GetShadowFontName: WideString;
begin
  Result := FShadowFontName;
end;

function TSpan.GetRightToLeft: Boolean;
begin
  Result := FRightToLeft;
end;

procedure TSpan.SetShadowFontName(const Value: WideString);
begin
  if FShadowFontName = Value then
    Exit;
  FShadowFontName := Value;
  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetRightToLeft(const Value: Boolean);
begin
  if FRightToLeft = Value then
    Exit;

  FRightToLeft := Value;

  if not FUpdating then
    CalcWidth;
end;

procedure TSpan.SetTextWithoutChange(const Value: WideString);
begin
  if FText = Value then
    Exit;
  FText := Value;
  FLength := System.Length(FText);

  if FNewLine then
  begin
    if FLength = 0 then
    begin
      FText := #14;
      FLength := 1;
    end
    else
    if (FText[FLength] <> #14) then
    begin
      FText := FText + #14;
      Inc(FLength);
    end;
  end;

  if FNewParagraph then
  begin
    if FLength = 0 then
    begin
      FText := #13;
      FLength := 1;
    end
    else
    if (FText[FLength] <> #13)  then
    begin
      FText := FText + #13;
      Inc(FLength);
    end;
  end;

  if not FUpdating then
    CalcWidth;
end;

function TSpan.GetShowTags: Boolean;
begin
  Result := FShowTags;
end;

procedure TSpan.SetShowTags(const Value: Boolean);
begin
  if Value = FShowTags then
    Exit;

  FShowTags := Value;
  CalcWidth;
end;

function TSpan.GetShowSpellErrors: Boolean;
begin
  Result := FShowSpellErrors;
end;

procedure TSpan.SetShowSpellErrors(const Value: Boolean);
begin
  FShowSpellErrors := Value;
end;

{ TGraphicSpan }

procedure TGraphicSpan.Assign(const Source: TSpan);
begin
  if Source is TGraphicSpan then
  begin
    FreeAndNil(FPicture);
    FPicture := CloneImage(TGraphicSpan(Source).FPicture);
  end;
  inherited;
end;

function TGraphicSpan.CalcTextSize(const Font: HFont;
  const Text: WideString): TSize;
begin
  Result := inherited CalcTextSize(Font, 'XX');
  if Text = '' then
    Result.cx := 0
  else
    Result.cx := FWidth;
end;

function TGraphicSpan.CalcTextWidth(const Text: WideString): Integer;
begin
  if Text = '' then
    Result := 0
  else
    Result := FWidth;
end;

procedure TGraphicSpan.CalcWidth;
var
  Font, OldFont: HFont;
  DC: HDC;
  Size: TSize;
  SuperSubHeight: Integer;
begin
  Font := CreateFont(True);
  if Font = 0 then
    Exit;

  DC := GetDC(0);
  OldFont := SelectObject(DC, Font);

  GetTextExtentPoint32(DC, PChar('Mp'), 2, Size);

  FImageHeight := Size.cy;
  FHeight := FImageHeight;
  FBaseline := CalcBaseline(Font);
  FStartFromBaseline := -FBaseline;
  SuperSubHeight := FHeight;

  SelectObject(DC, OldFont);
  DeleteObject(Font);

  FWidth := CalcImageWidth(FPicture, Size.cy);

  if Super or Sub then
  begin
    if Sub then
      FStartFromBaseline := -SuperSubHeight div 2;
    Font := CreateFont(False);

    OldFont := SelectObject(DC, Font);

    GetTextExtentPoint32(DC, PChar('Mp'), 2, Size);

    FBaseline := CalcBaseline(Font);
    SelectObject(DC, OldFont);
    DeleteObject(Font);

    FHeight := Size.cy;

    if Sub then
    begin
      if FBaseline + FStartFromBaseline + SuperSubHeight > FHeight then
        FStartFromBaseline := FHeight - SuperSubHeight - FBaseline;
    end;

    if Super then
      FStartFromBaseline := - FBaseline;
  end;
  ReleaseDC(0, DC);
end;

function TGraphicSpan.CanAdd(const Source: TSpan): Boolean;
begin
  Result := False;
end;

constructor TGraphicSpan.Create;
begin
  inherited;
  FPicture := nil;
  FText := '';
end;

procedure TGraphicSpan.DeleteSelected;
begin
  if FSelLength = 1 then
  begin
    FreeAndNil(FPicture);
    FSelLength := 0;
    FText := '';
    if not FUpdating then
      CalcWidth;
  end;
end;

destructor TGraphicSpan.Destroy;
begin
  FreeAndNil(FPicture);
  inherited;
end;

function TGraphicSpan.GetAsText(WithMarker: Boolean): WideString;
begin
  Result := ImageToText(FPicture, WithMarker);
end;

function TGraphicSpan.GetSVG: IXMLDOMDocument3;
begin
  if IsSVG then
    Result := GetSVGXML(FPicture)
  else
    Result := nil;
end;

function TGraphicSpan.IsSVG: Boolean;
begin
  Result := IsImageSVG(FPicture);
end;

procedure TGraphicSpan.Join(const Source: TSpan);
begin
end;

function TGraphicSpan.LastSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Result := -2;
end;

function TGraphicSpan.MeasureWidth(Width: Integer;
  WithLineBreak: Boolean): Integer;
begin
  if Width > FWidth / 2 then
    Result := 2
  else
    Result := 1;
end;

function TGraphicSpan.NextSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Ch := #0;
  Result := -2;
end;

procedure TGraphicSpan.Paint(const Target: TCanvas; X, Y, LineHeight, Baseline: Integer;
  Focused: Boolean);
var
  Brush: HBrush;
  Top, Bottom, StartY: Integer;
begin
  Brush := CreateSolidBrush(ColorToRGB(FBackColor));
  FillRect(Target.Handle, Rect(X, Y, X + FWidth, Y + LineHeight), Brush);
  DeleteObject(Brush);

  Top := Y;
  Bottom := Y + LineHeight;
  StartY := Top + Baseline - FBaseline;

  if FSub then
  begin
    StartY := Top + Baseline - Round(FHeight * SUPERSUBFRAC / 2);
    if StartY + FHeight * SUPERSUBFRAC >= Bottom then
      StartY := Bottom - Round(FHeight * SUPERSUBFRAC) - 1;
  end;

  PaintImage(FPicture, Target, X, StartY, FWidth, FImageHeight, FSelLength = 1, Focused);
end;

function TGraphicSpan.Part(Width: Integer): TSpan;
begin
  Result := nil;
end;

function TGraphicSpan.SameValues(const Rec: TSpanValues): Boolean;
begin
  Result := False;
end;

procedure TGraphicSpan.SetPicture(Image: Pointer);
begin
  FreeAndNil(FPicture);
  FPicture := CloneImage(Image);
  if FPicture <> nil then
    FText := '';
  FLength := System.Length(FText);
  CalcWidth;
end;

procedure TGraphicSpan.SetText(const Value: WideString);
begin
end;

procedure TGraphicSpan.SetTextWithoutChange(const Value: WideString);
begin
  if Value = '' then
    SetPicture(nil);
end;

function TGraphicSpan.Split(Position: Integer): TSpan;
begin
  Result := nil;
end;

{ TSpanList }

procedure TSpanList.Add(Span: TSpan);
begin
  inherited Add(Span);
end;

procedure TSpanList.Clear;
var
  C: Integer;
begin
  for C := 0 to Count - 1 do
    GetItem(C).Free;

  inherited;
end;

procedure TSpanList.Delete(Index: Integer);
begin
  GetItem(Index).Free;
  inherited Delete(Index);
end;

function TSpanList.First: TSpan;
begin
  Result := TSpan(inherited First);
end;

function TSpanList.GetItem(const Index: Integer): TSpan;
begin
  Result := TSpan(Get(Index));
end;

function TSpanList.Last: TSpan;
begin
  Result := TSpan(inherited Last);
end;

procedure TSpanList.SetItem(const Index: Integer; const Value: TSpan);
begin
  Put(Index, Value);
end;

{ TSectionSpan }

procedure TSectionSpan.Assign(const Source: TSpan);
begin
  inherited;

end;

procedure TSectionSpan.CalcWidth;
begin
  if (FText <> '') and FShowTags then
    FWidth := CalcTextWidth(FDisplay)
  else
    FWidth := 0;
end;

function TSectionSpan.CanAdd(const Source: TSpan): Boolean;
begin
  Result := False;
end;

constructor TSectionSpan.Create;
begin
  inherited;
  FDisplay := '===Abschnitt===';
  FText := '';
  FLength := 1;
  CalcWidth;
end;

procedure TSectionSpan.DeleteSelected;
begin
  if FSelLength = 1 then
  begin
    FSelLength := 0;
    FText := '';
    FLength := 0;
    if not FUpdating then
      CalcWidth;
  end;
end;

function TSectionSpan.GetLineBreak: Boolean;
begin
  Result := True;
end;

procedure TSectionSpan.Join(const Source: TSpan);
begin
end;

function TSectionSpan.LastSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Result := -2;
end;

function TSectionSpan.NextSpace(const Text: WideString; const Pos: Integer;
  var Ch: WideChar): Integer;
begin
  Ch := #0;
  Result := -2;
end;

procedure TSectionSpan.Paint(const Target: TCanvas; X, Y, LineHeight,
  Baseline: Integer; Focused: Boolean);
begin
  if (FText <> '') and FShowTags then
  begin
    FText := FDisplay;
    FLength := System.Length(FText);
    CalcWidth;
    if FSelLength > 0 then
      FSelLength := FLength;

    inherited;

    FText := '';
    FLength := 1;
    if FSelLength > 0 then
      FSelLength := 1;
  end;
end;

function TSectionSpan.Part(Width: Integer): TSpan;
begin
  Result := nil;
end;

function TSectionSpan.SameValues(const Rec: TSpanValues): Boolean;
begin
  Result := False;
end;

procedure TSectionSpan.SetLineBreak(const Value: Boolean);
begin
  inherited SetLineBreak(True);
end;

procedure TSectionSpan.SetText(const Value: WideString);
begin
  inherited;

end;

procedure TSectionSpan.SetTextWithoutChange(const Value: WideString);
begin
  if Value = '' then
  begin
    FText := '';
    FLength := 0;
  end;
end;

function TSectionSpan.Split(Position: Integer): TSpan;
begin
  Result := nil;
end;

end.
