unit SpellChecker;

interface

uses
  WideStringList;

type
  TCustomSpellChecker = class(TObject)
  private
    FSuggestions: TWideStringList;
    FDictionaries: TWideStringList;
    FCurrentDictionary: Integer;
    FDictionaryDir: WideString;

    procedure SetCurrentDictionary(Value: Integer);
    procedure SetDictionaryDir(const ADirectory: WideString);
  protected
    FLoaded: Boolean;
    procedure LoadDictionaries; virtual; abstract;
    procedure RealizeDictionary; virtual; abstract;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function SpelledCorrect(const AWord: WideString): Boolean; virtual; abstract;
    procedure FillSuggestions(const AWord: WideString); virtual; abstract;
    procedure Replace(const AWord, AReplacer: WideString); virtual; abstract;
    procedure Add(const AWord: WideString); virtual; abstract;
    procedure Ignore(const AWord: WideString); virtual; abstract;

    property Suggestions: TWideStringList read FSuggestions;
    property Dictionaries: TWideSTringList read FDictionaries;
    property DictionaryDir: WideString read FDictionaryDir write SetDictionaryDir;
    property CurrentDictionary: Integer read FCurrentDictionary write SetCurrentDictionary;
    property Loaded: Boolean read FLoaded;
  end;

implementation

uses
  SysUtils, MWKSysUtils;

{ TCustomSpellChecker }

constructor TCustomSpellChecker.Create;
begin
  inherited;
  FSuggestions := TWideStringList.Create;
  FDictionaries := TWideStringList.Create;
  FCurrentDictionary := -1;
end;

destructor TCustomSpellChecker.Destroy;
begin
  FSuggestions.Free;
  FDictionaries.Free;
  inherited;
end;

procedure TCustomSpellChecker.SetCurrentDictionary(Value: Integer);
begin
  if (Value = FCurrentDictionary) then
    Exit;

  if (Value < 0) or (Value >= FDictionaries.Count) then
    Exception.Create('Index out of bounds');

  FCurrentDictionary := Value;
  RealizeDictionary;
end;

procedure TCustomSpellChecker.SetDictionaryDir(const ADirectory: WideString);
begin
  FDictionaryDir := AddTrailingPathDelimiter(ADirectory);
  LoadDictionaries;
end;

end.
