unit MWKEdit;
      {******************************************************************}
      { MWKEdit                                                          }
      {                                                                  }
      { home page : http://www.mwcs.de                                   }
      { email     : martin.walter@mwcs.de                                }
      {                                                                  }
      { date      : 21-03-2008                                           }
      { version   : v0.91-beta                                          }
      {                                                                  }
      { Use of this file is permitted for commercial and non-commercial  }
      { use, as long as the original author is credited.                 }
      { This file (c) 2005, 2008 Martin Walter                           }
      {                                                                  }
      { This Software is distributed on an "AS IS" basis, WITHOUT        }
      { WARRANTY OF ANY KIND, either express or implied.                 }
      {                                                                  }
      { *****************************************************************}

interface

uses
  Windows, Messages, Classes, Controls, StdCtrls, ExtCtrls, Forms, Graphics, ActiveX,
  MSXML2_TLB_Light, XMLHelp,
  UndoRedoIntf, WideStringListIntf, WideStringList,
  SpellChecker, SpellCheckDialog,
  MWKEditSpans;

type
  TFileFormat = (ffXML, ffRTF, ffAnsi, ffUnicode);

  TChangeType = (ctBold, ctItalic, ctUnderline, ctStrikeOut, ctSpaced, ctSuper,
    ctSub, ctIndex, ctColor, ctBackColor, ctFontName, ctFontHeight);

  TNewCaretPos = (ncLeft, ncRight, ncUp, ncDown, ncHome, ncEnd, ncFirst, ncLast,
    ncPageDown, ncPageUp, ncPrevWord, ncNextWord);

  TSpanRange = record
    Start: Integer;
    Stop: Integer;
  end;
  PSpanRange = ^TSpanRange;

  TClipFormats = array of TClipFormat;

  TWideKeyPressEvent = procedure(Sender: TObject; var Key: WideChar) of object;

  TInsertState = (isInsert, isOverWrite);

  IMWKDropTarget = interface(IInterface)
  ['{2EDAD5B2-8BE4-4DAF-BAAC-59F4F750767E}']
    function GetActive: Boolean;
    procedure SetActive(const Value: Boolean);
    property Active: Boolean read GetActive write SetActive;
  end;

  TSpellChecker = type String;

  TSpellCheckerClass = class of TCustomSpellChecker;

  TMWKEdit = class(TCustomControl)
  private
    FDropTarget: IMWKDropTarget;
    FOLEDragDrop: Boolean;
    FOLEFormat: TClipFormat;
    FShowDragCaret: Boolean;
    FDragCaretPos: TPoint;
    FDragSpan: Integer;
    FPosInDragSpan: Integer;
    FIndentSize: Integer;

    IsAligning: Boolean;
    FBorderStyle: TBorderStyle;
    ClientWidth: Integer;
    FClientRect: TRect;
    FPlain: Boolean;
    FRightToLeft: Boolean;
    FUpdateCount: Integer;

    FFontName: TFontName;
    FFontHeight: Currency;
    FFontSize: Currency;
    FBold: Boolean;
    FItalic: Boolean;
    FUnderline: Boolean;
    FStrikeOut: Boolean;
    FSpaced: Boolean;
    FSuper: Boolean;
    FSub: Boolean;
    FIndex: Boolean;
    FFontColor: TColor;
    FBackColor: TColor;
    FReadOnly: Boolean;
    FInsertState: TInsertState;
    FOnChange: TNotifyEvent;
    FOnSelectionChange: TNotifyEvent;
    FOnKeyPress: TWideKeyPressEvent;
    FOnToggleInsert: TNotifyEvent;
    FSelStart: Integer;
    FSelLength: Integer;
    FText: WideString;

    FOldFontName: TFontName;
    FOldSuper: Boolean;
    FOldSub: Boolean;
    FOldFontHeight: Currency;
    FOldCaretStart: Integer;
    FOldCaretHeight: Integer;

    FShowTags: Boolean;
    FShowSpellErrors: Boolean;
    FItems: TSpanList;
    FLineStarts: TList;
    FTextHeight: Integer;
    FMaxLineSize: Integer;
    FLinePos: Integer;
    FVerticalScrollBar: TScrollBar;
    FHorizontalScrollBar: TScrollBar;
    FStartLine: Integer;
    FEndLine: Integer;
    FShiftDown: Boolean;
    FCtrlDown: Boolean;
    MSelStart: Integer;
    FCaretHeight: Integer;
    FCaretX: Integer;
    FOldCaretX: Integer;
    FLineCount: Integer;
    FLastEmpty: Boolean;
    FBrush: TBrush;
    FSpellCheckForm: TSpellCheckForm;
    FSpellCheckerName: TSpellChecker;
    FSpellChecker: TCustomSpellChecker;

    FPosInSpan: Integer;
    FCurrentSpan: Integer;
    FCurrentLine: Integer;

    FPosInMouseSpan: Integer;
    FCurrentMouseSpan: Integer;

    FCommingFromRight: Boolean;
    FNegativeSelection: Boolean;
    FWantTabs: Boolean;
    FTabWidth: Integer;

    FLeftDown: Boolean;
    MDouble: Boolean;

    FUndoRedo: IUndoRedo;
    FRollBack: Integer;

    FSelStartBeforeInsert: Integer;
    FSelLengthBeforeInsert: Integer;
    FPendingInsertFinish: Boolean;
    FTextBeforeInsert: IXMLDOMDocument;
    FInsertTimer: TTimer;
    FOldInsertAction: TObject;
    FGlobalIndent: Integer;
    FFixedFont: Boolean;
    FFixedColor: Boolean;
    FFixedSize: Boolean;
    FLineBreak: Boolean;
    FMultiLine: Boolean;
    FShowVerticalScrollBar: Boolean;
    FShowHorizontalScrollBar: Boolean;
    FShowScrollBars: Boolean;
    procedure OnInsertTimer(Sender: TObject);
    procedure FinishInsert;
    function GetRedoText: WideString;
    function GetCanRedo: Boolean;
    function GetCanUndo: Boolean;
    function GetUndoText: WideString;
    procedure BeginRollBack;
    procedure EndRollBack;

    procedure SetOLEDragDrop(const Value: Boolean);

    procedure WMEraseBackGround(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMSetFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMCut(var Message: TWMCut); message WM_CUT;
    procedure WMCopy(var Message: TWMCopy); message WM_COPY;
    procedure WMPaste(var Message: TWMPaste); message WM_PASTE;
    procedure WMClear(var Message: TWMClear); message WM_CLEAR;
    procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;

    function SupportsOLEData(const DataObj: IDataObject;
      const GrfKeyState: LongInt; const PT: TPoint; var dwEffect: LongInt): HRESULT;
    function InsertOLEData(const DataObj: IDataObject;
      const GrfKeyState: LongInt; PT: TPoint; var dwEffect: LongInt): HRESULT;
    function ModifyOLEDropEffect(const GrfKeyState: LongInt; const PT: TPoint;
      var dwEffect: LongInt): HRESULT;

    function OLEHasSupportedFormat(const DataObj: IDataObject;
      const Formats: array of TClipFormat; var Format: TClipFormat): Boolean;
    function GetMyOLEFormats: TClipFormats;

    procedure ShowDragCaret(PT: TPoint);
    procedure HideDragCaret;

    procedure CalcLineCountAndStart;
    procedure RemoveEmptySpans;
    procedure SetVerticalScrollBar;
    procedure SetHorizontalScrollBar;
    procedure SetScrollBarPositions;
    procedure ToggleScrollBar;
    procedure CalcMousePos(X, Y: Integer; MouseDown: Boolean);
    procedure CalcDragMousePos(X, Y: Integer);
    function GetMousePosAbsolute: Integer;
    procedure SetSpanValues(ChangeType: TChangeType);
    function FindStartLine: Integer;
    function FindEndLine: Integer;
    function FindLineOfSpan(Index: Integer): Integer;
    function GetCurrentSpanNum(var Position: Integer): Integer;
    function GetSpanAndPosition(var Position: Integer): Integer;

    procedure SetXMLAttributes(const Node: IXMLDOMNode; const Span: TSpan);
    procedure SetXMLIndent(const Node: IXMLDOMNode; Indent: Integer);
    function CollectAsXML(Selected, Plain: Boolean): WideString;
    procedure AddToXML(const XML: IXMLDOMDocument; Selected: Boolean);

    function ParseXML(const XML: IXMLDOMDocument; Insert: Integer;
      var BreakCount: Integer; WithCurrentStyle: Boolean): Integer;

    procedure ReBuildIndent;
    procedure ReBuildLines(StartLine, EndLine: Integer);
    function GetWordStart(Position: Integer): Integer;
    function GetWordLength(Position: Integer): Integer;
    function GetWord(WithSpace: Boolean; var Start, Length: Integer): WideString;
    procedure RemoveSelection;
    procedure ClearSpells;
    procedure SetSpell(const AStart, ALength: Integer);
    procedure CalcSpellCheck(const WithPaint: Boolean);
    procedure CalcSelection;
    procedure CalcCurrentLine;
    function ScrollInSight(Point: TPoint): Boolean;
    procedure PaintLine(Index, X, Y: Integer);
    function Shrink(Rect: TRect; Size: Integer = 1): TRect;
    function Expand(Rect: TRect; Size: Integer = 1): TRect;
    function IsSpelledCorrect(AStart: Integer; var ALength: Integer): Boolean;
    procedure CalcAreas;
    function GetCursorPos: Integer;
    function GetEndOfLinePos(WithLineBreak: Boolean): Integer;
    function GetStartOfLinePos: Integer;
    function GetPrevWordPos: Integer;
    function GetNextWordPos: Integer;
    procedure GetCaret(var X, Y: Integer);
    procedure CalcCaretPos(var Top, Height: Integer);
    function SetCaretPos(CalcHeight: Boolean): TPoint;
    procedure SetCaretTo(Where: TNewCaretPos);
    function GetHeightOfCurrentFont: Integer;
    procedure SetSelection(Where: TNewCaretPos);
    procedure Delete(Next: Boolean);
    procedure DeleteSelected(WithEvent, WithRebuild: Boolean); overload;
    procedure InsertXML(const XML: IXMLDOMDocument);
    procedure DefineSelection(const Start, Length: Integer);
    procedure RealizeFontHeight;

    function GetBackColor: TColor;
    function GetBold: Boolean;
    function GetColor: TColor;
    function GetIndex: Boolean;
    function GetItalic: Boolean;
    function GetOnChange: TNotifyEvent;
    function GetPlainText: WideString;
    function GetSelectedPlainText: WideString;
    function GetSelectedText: WideString;
    function GetSelLength: Integer;
    function GetSelStart: Integer;
    function GetLength: Integer;
    function GetSpaced: Boolean;
    function GetStrikeOut: Boolean;
    function GetSub: Boolean;
    function GetSuper: Boolean;
    function GetUnderline: Boolean;
    function GetInsertState: TInsertState;
    function GetSpellChecker: TSpellChecker;
    procedure SetSpellChecker(const Value: TSpellChecker);
    procedure SetBackColor(const Value: TColor);
    procedure SetBold(const Value: Boolean);
    procedure SetColor(const Value: TColor);
    procedure SetIndex(const Value: Boolean);
    procedure SetItalic(const Value: Boolean);
    procedure SetOnChange(const Value: TNotifyEvent);
    procedure SetSelectedText(const Value: WideString);
    procedure SetSelLength(const Value: Integer);
    procedure SetSelStart(const Value: Integer);
    procedure SetSpaced(const Value: Boolean);
    procedure SetStrikeOut(const Value: Boolean);
    procedure SetSub(const Value: Boolean);
    procedure SetSuper(const Value: Boolean);
    procedure SetXMLText(const XML: IXMLDOMDocument);
    procedure SetUnderline(const Value: Boolean);
    procedure SetInsertState(const Value: TInsertState);
    function GetBorderStyle: TBorderStyle;
    procedure SetBorderStyle(const Value: TBorderStyle);
    function GetOnKeyPress: TWideKeyPressEvent;
    procedure SetOnKeyPress(const Value: TWideKeyPressEvent);
    function GetOnSelectionChange: TNotifyEvent;
    procedure SetOnSelctionChange(const Value: TNotifyEvent);

    procedure SetSizeFromHeight;
    function GetPlain: Boolean;
    procedure SetPlain(const Value: Boolean);
    procedure SetPlainText(const Value: WideString);
    function GetReadOnly: Boolean;
    procedure SetReadOnly(const Value: Boolean);
    function GetFontColor: TColor;
    function GetFontHeight: Currency;
    function GetFontName: TFontName;
    function GetFontSize: Currency;
    procedure SetFontColor(const Value: TColor);
    procedure SetFontHeight(const Value: Currency);
    procedure SetFontName(const Value: TFontName);
    procedure SetFontSize(const Value: Currency);
    {function GetRightToLeft: Boolean;
    procedure SetRightToLeft(const Value: Boolean);}

    function GetShowTags: Boolean;
    procedure SetShowTags(const Value: Boolean);
    function GetOnToggleInsert: TNotifyEvent;
    procedure SetOnToggleInsert(const Value: TNotifyEvent);
    function GetDictionaries: TWideStringList;
    function GetDictionary: Integer;
    procedure SetDictionary(const Value: Integer);
    function SpellCheck(AStart, ALength: Integer; Highlight: Boolean): Integer;
    function GetCurrentLine: Integer;
    function GetHeightOfLine(const Index: Integer): Integer;
    function GetBaselineOfLine(const Index: Integer): Integer;
    function GetLineOfY(const Y: Integer): Integer;
    function GetLineStart(const Index: Integer): Integer;
    function GetShowSpellErrors: Boolean;
    procedure SetShowSpellErrors(const Value: Boolean);
    function GetXML: WideString;
    procedure SetXML(const Value: WideString);
    function GetLineCount: Integer;
    procedure SetTabWidth(const Value: Integer);
    function GetRedoTexts: IWideStringList;
    function GetUndoTexts: IWideStringList;
    function GetIndent: Integer;
    procedure SetIndent(const Value: Integer);
    function CreateXMLFromPlain(S: WideString;
      StyleSpan: TSpan): IXMLDOMDocument;
    procedure SetIndentSize(const Value: Integer);
    function GetIsChanged: Boolean;
    function GetEmpty: Boolean;
    procedure SetFixedColor(const Value: Boolean);
    procedure SetFixedFont(const Value: Boolean);
    procedure SetFixedSize(const Value: Boolean);
    procedure SetLineBreak(const Value: Boolean);
    procedure SetShowScrollBars(const Value: Boolean);
    procedure SetMultiLine(const Value: Boolean);
  protected
    procedure DoOnChange;
    procedure DoOnSelectionChange;
    procedure DoOnVerticalScrollBarChange(Sender: TObject);
    procedure DoOnHorizontalScrollBarChange(Sender: TObject);
    procedure Paint; override;
    procedure CreateWnd; override;
    procedure PaintDragCaret;
    function CalculateSelectedSpanRange: TSpanRange;
    function CalculateParagraphRange: TSpanRange;
    function GetSelectedParagraphs: TSpanRange;
    function GetParagraphIndent(Paragraph: Integer): Integer;
    procedure SetParagraphIndent(Paragraph, Indent: Integer);
    function DoKeyPress(var Message: TWMKey): Boolean;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: WideChar); reintroduce;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure Resize; override;
    procedure CreateParams(var Params: TCreateParams); override;

    function GetSpanValues: TSpanValues;
    procedure UpdateSpanValues(Value: TSpanValues);
    procedure ClearSpanValues;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function Focused: Boolean; override;
    procedure Clear;
    procedure SelectAll;
    procedure ClearSelection;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure DeleteSelected; overload;
    procedure PasteFromClipboard;
    function GetHeightToLine(const Line: Integer): Integer;
    procedure BeginUpdate;
    procedure EndUpdate(WithPaint: Boolean);
    procedure Repaint; override;
    function GetCurrentWord(const Position: Integer): WideString;
    procedure SpellCheckSelected;
    procedure SpellCheckAll;
    procedure AddCharacter(const Ch: WideChar);
    function LoadFromFile(const FileName: WideString): Integer;
    procedure SaveToFile(const FileName: WideString; FileFormat: TFileFormat);
    procedure LoadFromStream(const Stream: TStream);
    procedure SaveToStream(const Stream: TStream; FileFormat: TFileFormat);
    function EncodeToRTF: AnsiString;
    function LoadFromMSWord(const FileName: WideString): Integer;
    procedure Undo;
    procedure Redo;
    procedure ClearUndo;
    procedure InsertPicture(const Picture: TObject); overload;
    procedure InsertPicture(const FileName: WideString); overload;
    procedure InsertSection;
    procedure IncIndent;
    procedure DecIndent;
    procedure Changed;

    class procedure RegisterSpellChecker(const Name: TSpellChecker;
      SpellCheckerClass: TSpellCheckerClass);

    property Empty: Boolean read GetEmpty;
    property IsChanged: Boolean read GetIsChanged;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelectedText: WideString read GetSelectedText write SetSelectedText;
    property SelectedPlainText: WideString read GetSelectedPlainText;
    property Length: Integer read GetLength;
    property Bold: Boolean read GetBold write SetBold;
    property Italic: Boolean read GetItalic write SetItalic;
    property Underline: Boolean read GetUnderline write SetUnderline;
    property StrikeOut: Boolean read GetStrikeOut write SetStrikeOut;
    property Spaced: Boolean read GetSpaced write SetSpaced;
    property Super: Boolean read GetSuper write SetSuper;
    property Sub: Boolean read GetSub write SetSub;
    property Index: Boolean read GetIndex write SetIndex;
    property InsertState: TInsertState read GetInsertState write SetInsertState;
    property Dictionaries: TWideStringList read GetDictionaries;
    property Dictionary: Integer read GetDictionary write SetDictionary;
    property CurrentLine: Integer read GetCurrentLine;
    property CanUndo: Boolean read GetCanUndo;
    property CanRedo: Boolean read GetCanRedo;
    property LineCount: Integer read GetLineCount;
    property UndoText: WideString read GetUndoText;
    property RedoText: WideString read GetRedoText;
    property UndoTexts: IWideStringList read GetUndoTexts;
    property RedoTexts: IWideStringList read GetRedoTexts;
    property Indent: Integer read GetIndent write SetIndent;
    property CurrentSpellChecker: TCustomSpellChecker read FSpellChecker;
  published
    property FontName: TFontName read GetFontName write SetFontName;
    property FontHeight: Currency read GetFontHeight write SetFontHeight;
    property FontSize: Currency read GetFontSize write SetFontSize;
    property FontColor: TColor read GetFontColor write SetFontColor;
    property BackgroundColor: TColor read GetBackColor write SetBackColor;
    property Color: TColor read GetColor write SetColor;
    property FixedColor: Boolean read FFixedColor write SetFixedColor;
    property FixedFont: Boolean read FFixedFont write SetFixedFont;
    property FixedSize: Boolean read FFixedSize write SetFixedSize;
    property Align;
    property Anchors;
    property Enabled;
    property Visible;
    property TabStop;
    property TabOrder;
    property PopupMenu;
    property BorderStyle: TBorderStyle read GetBorderStyle write SetBorderStyle;
    property Plain: Boolean read GetPlain write SetPlain;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property ShowTags: Boolean read GetShowTags write SetShowTags;
    property WantTabs: Boolean read FWantTabs write FWantTabs;
    property TabWidth: Integer read FTabWidth write SetTabWidth;
    property LineBreak: Boolean read FLineBreak write SetLineBreak;
    property MultiLine: Boolean read FMultiLine write SetMultiLine;
    property ShowScrollBars: Boolean read FShowScrollBars write SetShowScrollBars;
    property IndentSize: Integer read FIndentSize write SetIndentSize;
    property ShowSpellErrors: Boolean read GetShowSpellErrors write SetShowSpellErrors;
    property OLEDragDrop: Boolean read FOLEDragDrop write SetOLEDragDrop;
    property SpellChecker: TSpellChecker read GetSpellChecker write SetSpellChecker;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyUp;
    property OnMouseMove;
    property OnMouseDown;
    property OnMouseUp;
    // property RightToLeft: Boolean read GetRightToLeft write SetRightToLeft;
    property PlainText: WideString read GetPlainText write SetPlainText;
    property XML: WideString read GetXML write SetXML;
    property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
    property OnSelectionChange: TNotifyEvent read GetOnSelectionChange write SetOnSelctionChange;
    property OnKeyPress: TWideKeyPressEvent read GetOnKeyPress write SetOnKeyPress;
    property OnToggleInsert: TNotifyEvent read GetOnToggleInsert write SetOnToggleInsert;
  end;

  TMWKEditDropTarget = class(TInterfacedObject, IMWKDropTarget, IDropTarget)
  private
    FEdit: TMWKEdit;
    FEditHandle: THandle;
    FActive: Boolean;
    function GetActive: Boolean;
    procedure SetActive(const Value: Boolean);
  public
    constructor Create(Edit: TMWKEdit);
    destructor Destroy; override;
    function DragEnter(const DataObj: IDataObject; GrfKeyState: LongInt;
      PT: TPoint; var dwEffect: LongInt): HResult; stdcall;
    function DragOver(GrfKeyState: Longint; PT: TPoint;
      var dwEffect: LongInt): HResult; stdcall;
    function DragLeave: HResult; stdcall;
    function Drop(const DataObj: IDataObject; GrfKeyState: LongInt; PT: TPoint;
      var dwEffect: LongInt): HResult; stdcall;
    property Active: Boolean read GetActive write SetActive;
  end;

  PSpellCheckerEntry = ^TSpellCheckerEntry;
  TSpellCheckerEntry = record
    SpellCheckerClass: TSpellCheckerClass;
    Name: TSpellChecker;
  end;

  TSpellCheckerList = class(TList)
  public
    destructor Destroy; override;
    function GetName(const Index: Integer): TSpellChecker;
    function IndexOf(const Name: TSpellChecker): Integer;
    procedure Add(const AName: TSpellChecker; AClass: TSpellCheckerClass);
    procedure Remove(AClass: TSpellCheckerClass);
  end;

function GetSpellCheckers: TSpellCheckerList;

implementation

uses
  SysUtils, Math, ComObj, Variants, Dialogs, Types,
  CreateUndoRedoObject, MWKSysUtils,
  MWKEditGraphics, MWKEditRTF2XML;

type
  TFormatEnum = class
  private
    FFormats: array of TFormatETC;
  public
    constructor Create(const DataObject: IDataObject);
    destructor Destroy; override;
    function HasFormat(const cfFormat: TClipFormat): Boolean;
    function GetFormatETC(const cfFormat: TClipFormat): TFormatETC;
  end;

var
  CF_MWKTEXT: TClipFormat;
  CF_RTF: TClipFormat;
  CF_HTML: TClipFormat;

  SpellCheckers: TSpellCheckerList;

{ Undo/Redo}
{$REGION Undo/Redo}

type
  TInsertAction = class
    Edit: TMWKEdit;
    StartLine: Integer;
    EndLine: Integer;
    PosInSpan: Integer;
    CurrentSpan: Integer;

    SelStart: Integer;
    SelLength: Integer;
    NegativeSelection: Boolean;
    Text: WideString;
    Style: TSpanValues;
    NewLength: Integer;
    Position: Integer;
  end;

  TIndentAction = class
    Start: Integer;
    Stop: Integer;
    Indents: TList;
    constructor Create;
    destructor Destroy; override;
  end;

  TStyleAction = class
  private
  public
    StartLine: Integer;
    EndLine: Integer;
    PosInSpan: Integer;
    CurrentSpan: Integer;

    SelStart: Integer;
    SelLength: Integer;
    NegativeSelection: Boolean;
    Text: WideString;
    Style: TSpanValues;
  end;

var
  UINSERT       : Integer;
  UINSERTIMAGE  : Integer;
  UINSERTSECTION: Integer;
  UDELETE       : Integer;
  UCUT          : Integer;
  USTYLE        : Integer;
  USPELLCHECK   : Integer;
  UINDENT       : Integer;

function GetStandardNames(const Action, Language: Integer): PWideChar; stdcall;
begin
  Result := '';
  
  if Action = UINSERT then
    Result := 'Eingabe';

  if Action = UINSERTIMAGE then
    Result := 'Bild einfgen';

  if Action = UINSERTSECTION then
    Result := 'Abschnittswechsel';

  if Action = UDELETE then
    Result := 'Lschen';

  if Action = UCUT then
    Result := 'Ausschneiden';

  if Action = USTYLE then
    Result := 'Formatierung';

  if Action = USPELLCHECK then
    Result := 'Rechtschreibprfung';

  if Action = UINDENT then
    Result := 'Einrckung';
end;

type
  TInsertRec = record
    OldInsertAction: TInsertAction;
    Edit: TMWKEdit;
    Text: IXMLDOMDocument;
  end;
  PInsertRec = ^TInsertRec;

procedure AddInsert(const Action: IUndoRedoAction; const ActionType: TActionType;
  const P1, P2: Integer); stdcall;
var
  InsertAction: TInsertAction;
  MWKEdit: TMWKEdit;
  S: WideString;
  IR: PInsertRec;
begin
  InsertAction := TInsertAction.Create;

  IR := PInsertRec(P1);
  MWKEdit := IR^.Edit;

  Action.Data := MWKEdit;
  InsertAction.StartLine := MWKEdit.FStartLine;
  InsertAction.EndLine := MWKEdit.FEndLine;
  InsertAction.PosInSpan := MWKEdit.FPosInSpan;
  InsertAction.CurrentSpan := MWKEdit.FCurrentSpan;
  InsertAction.Edit := MWKEdit;

  case ActionType of
    atUndo:
    begin
      IR^.OldInsertAction := InsertAction;
      InsertAction.SelStart := MWKEdit.FSelStart;
      InsertAction.SelLength := MWKEdit.FSelLength;
      InsertAction.NegativeSelection := MWKEdit.FNegativeSelection;
      if P2 < 0 then // We are in Overwrite Mode
      begin
        Inc(MWKEdit.FUpdateCount);
        MWKEdit.BeginRollBack;
        MWKEdit.SelLength := MWKEdit.FSelLength + 1;
        InsertAction.Text := MWKEdit.SelectedText;
        MWKEdit.SelLength := InsertAction.SelLength;
        MWKEdit.EndRollBack;
        Dec(MWKEdit.FUpdateCount);
      end else
        InsertAction.Text := MWKEdit.SelectedText;
      InsertAction.NewLength := P2;
    end;
    atRedo:
    begin
      if Assigned(IR^.Text) then
      begin
        S := IR^.Text.xml;
        Ir^.OldInsertAction.Text := S;
      end;
      Ir^.OldInsertAction.NewLength := MWKEdit.FSelLength;
      InsertAction.SelStart := MWKEdit.FSelStart;
      InsertAction.SelLength := MWKEdit.FSelLength;
      InsertAction.NegativeSelection := False;
      InsertAction.Text := MWKEdit.SelectedText;
      if Assigned(IR^.Text) and (P2 < 0) then
        InsertAction.NewLength := - (MWKEdit.FSelLength + Ir^.OldInsertAction.SelLength)
      else
        InsertAction.NewLength := P2;
    end;
  end;
  InsertAction.Style := MWKEdit.GetSpanValues;
  Action.ObjectList.Add(InsertAction);
end;

procedure UndoInsert(const Action: IUndoRedoAction;
  const ActionType: TActionType); stdcall;
var
  MWKEdit: TMWKEdit;
  InsertAction: TInsertAction;
begin
  InsertAction := Action.ObjectList[0];
  MWKEdit := Action.Data;
  Inc(MWKEdit.FUpdateCount);
  MWKEdit.BeginRollBack;

  case ActionType of
    atUndo:
    begin
      MWKEdit.SelStart := InsertAction.SelStart;
      MWKEdit.SelLength := Abs(InsertAction.NewLength);
      MWKEdit.ClearSpanValues;
      MWKEdit.SelectedText := InsertAction.Text;
      MWKEdit.SelStart := InsertAction.SelStart;
      MWKEdit.SelLength := InsertAction.SelLength;
      MWKEdit.FNegativeSelection := InsertAction.NegativeSelection;
      MWKEdit.UpdateSpanValues(InsertAction.Style);
    end;
    atRedo:
    begin
      MWKEdit.SelStart := InsertAction.SelStart;
      if InsertAction.NewLength < 0 then // We are in Overwrite Mode
        MWKEdit.SelLength := Abs(InsertAction.NewLength + 1) + 1
      else
        MWKEdit.SelLength := InsertAction.NewLength;
      MWKEdit.ClearSpanValues;
      MWKEdit.SelectedText := InsertAction.Text;
      MWKEdit.SelStart := InsertAction.SelStart + InsertAction.SelLength;
      MWKEdit.FNegativeSelection := InsertAction.NegativeSelection;
      MWKEdit.UpdateSpanValues(InsertAction.Style);
    end;
  end;

  MWKEdit.FStartLine := InsertAction.StartLine;
  MWKEdit.FEndLine := InsertAction.EndLine;
  MWKEdit.FPosInSpan := InsertAction.PosInSpan;
  MWKEdit.FCurrentSpan := InsertAction.CurrentSpan;

  MWKEdit.CalcSelection;
  MWKEdit.FCurrentSpan := MWKEdit.GetCurrentSpanNum(MWKEdit.FPosInSpan);

  MWKEdit.EndRollBack;
  Dec(MWKEdit.FUpdateCount);
  MWKEdit.Invalidate;

  MWKEdit.DoOnChange;
  MWKEdit.DoOnSelectionChange;
end;

procedure AddIndent(const Action: IUndoRedoAction; const ActionType: TActionType;
  const P1, P2: Integer); stdcall;
var
  MWKEdit: TMWKEdit;
  C: Integer;
  Range: PSpanRange;
  IndentAction: TIndentAction;
begin
  MWKEdit := TMWKEdit(P1);
  Range := PSpanRange(P2);

  Action.Data := MWKEdit;
  IndentAction := TIndentAction.Create;
  IndentAction.Start := Range.Start;
  IndentAction.Stop := Range.Stop;

  Action.ObjectList.Add(IndentAction);

  if Range^.Start = -1 then
  begin
    IndentAction.Indents.Add(Pointer(MWKEdit.FGlobalIndent));
    Exit;
  end;

  for C := Range^.Start to Range^.Stop do
    IndentAction.Indents.Add(Pointer(MWKedit.GetParagraphIndent(C)));
end;

procedure UndoIndent(const Action: IUndoRedoAction;
  const ActionType: TActionType); stdcall;
var
  MWKEdit: TMWKEdit;
  IndentAction: TIndentAction;
  C: Integer;
begin
  IndentAction := Action.ObjectList[0];

  MWKEdit := Action.Data;
  Inc(MWKEdit.FUpdateCount);
  MWKEdit.BeginRollBack;

  if IndentAction.Start = -1 then
    MWKEdit.FGlobalIndent := Integer(IndentAction.Indents[0])
  else
    for C := IndentAction.Start to IndentAction.Stop do
      MWKEdit.SetParagraphIndent(C, Integer(IndentAction.Indents[C - IndentAction.Start]));

  MWKEdit.ReBuildLines(0, MaxInt);
  MWKEdit.CalcSpellCheck(False);
  MWKEdit.CalcSelection;
  MWKEdit.FCurrentSpan := MWKEdit.GetCurrentSpanNum(MWKEdit.FPosInSpan);
  MWKEdit.EndRollBack;
  Dec(MWKEdit.FUpdateCount);
  MWKEdit.Invalidate;

  MWKEdit.DoOnChange;
  MWKEdit.DoOnSelectionChange;
end;

procedure AddDelete(const Action: IUndoRedoAction; const ActionType: TActionType;
  const P1, P2: Integer); stdcall;
var
  MWKEdit: TMWKEdit;
  InsertAction: TInsertAction;
  Start: Integer;
begin
  InsertAction := TInsertAction.Create;
  MWKEdit := TMWKEdit(P1);
  Action.Data := MWKEdit;
  InsertAction.StartLine := MWKEdit.FStartLine;
  InsertAction.EndLine := MWKEdit.FEndLine;
  InsertAction.PosInSpan := MWKEdit.FPosInSpan;
  InsertAction.CurrentSpan := MWKEdit.FCurrentSpan;

  case ActionType of
    atUndo:
    begin
      if MWKEdit.FSelLength = 0 then
      begin
        Inc(MWKEdit.FUpdateCount);
        MWKEdit.BeginRollBack;
        Start := MWKEdit.FSelStart;
        if not Boolean(P2) then // BackSpace
          MWKEdit.SelStart := MWKEdit.SelStart - 1;
        InsertAction.Position := MWKEdit.SelStart;
        MWKEdit.SelLength := 1;
        InsertAction.Text := MWKEdit.SelectedText;
        MWKEdit.SelStart := Start;
        MWKEdit.SelLength := 0;

        MWKEdit.EndRollBack;
        Dec(MWKEdit.FUpdateCount);
      end else
      begin
        InsertAction.Text := MWKEdit.SelectedText;
        InsertAction.Position := MWKEdit.FSelStart;
      end;
      InsertAction.SelStart := MWKEdit.FSelStart;
      InsertAction.SelLength := MWKEdit.FSelLength;
      InsertAction.NegativeSelection := MWKEdit.FNegativeSelection;
      InsertAction.NewLength := P2;
    end;
    atRedo:
    begin
      InsertAction.Position := MWKEdit.FSelStart;
      InsertAction.SelStart := MWKEdit.FSelStart;
      InsertAction.SelLength := MWKEdit.FSelLength;
      InsertAction.NegativeSelection := False;
      InsertAction.Text := MWKEdit.SelectedText;
      InsertAction.NewLength := P2;
    end;
  end;
  InsertAction.Style := MWKEdit.GetSpanValues;
  Action.ObjectList.Add(InsertAction);
end;

procedure UndoDelete(const Action: IUndoRedoAction;
  const ActionType: TActionType); stdcall;
var
  MWKEdit: TMWKEdit;
  InsertAction: TInsertAction;
begin
  InsertAction := Action.ObjectList[0];
  MWKEdit := Action.Data;
  Inc(MWKEdit.FUpdateCount);
  MWKEdit.BeginRollBack;
  case ActionType of
    atUndo:
    begin
      MWKEdit.SelStart := InsertAction.Position;
      MWKEdit.SelLength := 0;
      MWKEdit.ClearSpanValues;
      MWKEdit.SelectedText := InsertAction.Text;
      MWKEdit.SelStart := InsertAction.SelStart;
      MWKEdit.SelLength := InsertAction.SelLength;
      MWKEdit.FNegativeSelection := InsertAction.NegativeSelection;
      MWKEdit.UpdateSpanValues(InsertAction.Style);
    end;
    atRedo:
    begin
      MWKEdit.SelStart := InsertAction.Position;
      MWKEdit.SelLength := InsertAction.NewLength;
      MWKEdit.ClearSpanValues;
      MWKEdit.SelectedText := InsertAction.Text;
      MWKEdit.SelStart := InsertAction.SelStart + InsertAction.SelLength;
      MWKEdit.FNegativeSelection := InsertAction.NegativeSelection;
      MWKEdit.UpdateSpanValues(InsertAction.Style);
    end;
  end;

  MWKEdit.FStartLine := InsertAction.StartLine;
  MWKEdit.FEndLine := InsertAction.EndLine;
  MWKEdit.FPosInSpan := InsertAction.PosInSpan;
  MWKEdit.FCurrentSpan := InsertAction.CurrentSpan;

  MWKEdit.CalcSelection;
  MWKEdit.FCurrentSpan := MWKEdit.GetCurrentSpanNum(MWKEdit.FPosInSpan);

  MWKEdit.EndRollBack;
  Dec(MWKEdit.FUpdateCount);
  MWKEdit.Invalidate;

  MWKEdit.DoOnChange;
  MWKEdit.DoOnSelectionChange;
end;

procedure AddStyle(const Action: IUndoRedoAction; const ActionType: TActionType;
  const P1, P2: Integer); stdcall;
var
  StyleAction: TStyleAction;
  MWKEdit: TMWKEdit;
begin
  StyleAction := TStyleAction.Create;
  MWKEdit := TMWKEdit(P1);
  Action.Data := MWKEdit;
  StyleAction.SelStart := MWKEdit.SelStart;
  StyleAction.SelLength := MWKEdit.SelLength;
  StyleAction.NegativeSelection := MWKEdit.FNegativeSelection;
  StyleAction.Text := MWKEdit.SelectedText;
  StyleAction.Style := MWKEdit.GetSpanValues;
  Action.ObjectList.Add(StyleAction);
end;

procedure UndoStyle(const Action: IUndoRedoAction;
  const ActionType: TActionType); stdcall;
var
  MWKEdit: TMWKEdit;
  StyleAction: TStyleAction;
begin
  StyleAction := Action.ObjectList[0];
  MWKEdit := Action.Data;

  Inc(MWKEdit.FUpdateCount);
  MWKEdit.BeginRollBack;
  MWKEdit.SelStart := StyleAction.SelStart;
  MWKEdit.SelLength := StyleAction.SelLength;
  MWKEdit.SelectedText := StyleAction.Text;
  MWKEdit.SelStart := StyleAction.SelStart;
  MWKEdit.SelLength := StyleAction.SelLength;
  MWKEdit.FNegativeSelection := StyleAction.NegativeSelection;
  MWKEdit.UpdateSpanValues(StyleAction.Style);

  MWKEdit.CalcSelection;
  MWKEdit.FCurrentSpan := MWKEdit.GetCurrentSpanNum(MWKEdit.FPosInSpan);
  MWKEdit.EndRollBack;
  Dec(MWKEdit.FUpdateCount);
  MWKEdit.Invalidate;

  MWKEdit.DoOnChange;
  MWKEdit.DoOnSelectionChange;
end;
{$ENDREGION}


{ SpellChecker functions }

function GetSpellCheckers: TSpellCheckerList;
begin
  if SpellCheckers = nil then
    SpellCheckers := TSpellCheckerList.Create;
  Result := SpellCheckers;
end;


{ Help Functions }
{$REGION Help Functions}

function StripAllXML(const XML: IXMLDOMDocument): WideString;
var
  S: WideString;
  First: Boolean;

  procedure AddNode(const Node: IXMLDOMNode);
  var
    C: Integer;
    T: WideString;
  begin
    T := '';
    if (Node.nodeName = 'span') then
    begin
      if HasAttribute(Node, 'sb') then
        T := DupeString(' ', GetAttribute(Node, 'sb'));

      T := T + Node.text;

      if HasAttribute(Node, 'sa') then
        T := T + DupeString(' ', GetAttribute(Node, 'sa'));
    end;

    if Node.nodeName = 'g' then
      T := '';
    if Node.nodeName = 'tab' then
      T := cTab;

    if Node.nodeName = 'br' then
      T := #13;

    if Node.nodeName = 'p' then
    begin
      if First then
        First := False
      else
        T := #13;
    end;

    S := S + T;

    if (Node.nodeName = 'p') or (Node.nodeName = 'mwkedit') then
      for C := 0 to Node.childNodes.length - 1 do
        AddNode(Node.childNodes[C]);
  end;

begin
  S := '';
  First := True;
  AddNode(XML.documentElement);
  Result := S;
end;

procedure WriteToClipboardW(Format: Word; const Text: WideString);
var
  Data: THandle;
  DataPtr: Pointer;
  Size: Integer;
begin
  Size := (Length(Text) + 1) * SizeOf(WideChar);
  try
    Data := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, Size);
    try
      DataPtr := GlobalLock(Data);
      try
        Move(PWideChar(Text)^, DataPtr^, Size);
        SetClipboardData(Format, Data);
      finally
        GlobalUnlock(Data);
      end;
    except
      GlobalFree(Data);
    end;
  except
  end;
end;

procedure WriteToClipboardA(Format: Word; const Text: AnsiString);
var
  Data: THandle;
  DataPtr: Pointer;
  Size: Integer;
begin
  Size := (Length(Text) + 1);
  try
    Data := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, Size);
    try
      DataPtr := GlobalLock(Data);
      try
        Move(PChar(Text)^, DataPtr^, Size);
        SetClipboardData(Format, Data);
      finally
        GlobalUnlock(Data);
      end;
    except
      GlobalFree(Data);
    end;
  except
  end;
end;

function ReadFromClipBoardW(Format: Word): WideString;
var
  Data: THandle;
begin
  Result := '';
  if not OpenClipboard(0) then
    Exit;
  Data := GetClipboardData(Format);
  try
    if Data <> 0 then
      Result := PWideChar(GlobalLock(Data));
  finally
    if Data <> 0 then
      GlobalUnlock(Data);
    CloseClipboard;
  end;
end;

function ReadFromClipBoardA(Format: Word): AnsiString;
var
  Data: THandle;
begin
  Result := '';
  if not OpenClipboard(0) then
    Exit;
  Data := GetClipboardData(Format);
  try
    if Data <> 0 then
      Result := PChar(GlobalLock(Data));
  finally
    if Data <> 0 then
      GlobalUnlock(Data);
    CloseClipboard;
  end;
end;

type
  TStreamCharSet = (csAnsi, csUnicode, csUnicodeSwapped, csUtf8);

const
  UNICODE_BOM = WideChar($FEFF);
  UNICODE_BOM_SWAPPED = WideChar($FFFE);
  UTF8_BOM = AnsiString(#$EF#$BB#$BF);

function DetectCharacterSet(const Stream: TStream): TStreamCharSet;
var
  ByteOrderMark: WideChar;
  BytesRead: Integer;
  Utf8Test: array[0..2] of AnsiChar;
begin
  // Byte Order Mark
  ByteOrderMark := #0;
  if (Stream.Size - Stream.Position) >= SizeOf(ByteOrderMark) then
  begin
    BytesRead := Stream.Read(ByteOrderMark, SizeOf(ByteOrderMark));
    if (ByteOrderMark <> UNICODE_BOM) and (ByteOrderMark <> UNICODE_BOM_SWAPPED) then
    begin
      ByteOrderMark := #0;
      Stream.Seek(-BytesRead, soFromCurrent);
      if (Stream.Size - Stream.Position) >= Length(Utf8Test) * SizeOf(AnsiChar) then
      begin
        BytesRead := Stream.Read(Utf8Test[0], Length(Utf8Test) * SizeOf(AnsiChar));
        if Utf8Test <> UTF8_BOM then
          Stream.Seek(-BytesRead, soFromCurrent);
      end;
    end;
  end;

  if ByteOrderMark = UNICODE_BOM then
    Result := csUnicode
  else
    if ByteOrderMark = UNICODE_BOM_SWAPPED then
      Result := csUnicodeSwapped
    else
      if Utf8Test = UTF8_BOM then
        Result := csUtf8
      else
        Result := csAnsi;
end;

procedure StrSwapByteOrder(Str: PWideChar);
var
  P: PWord;
begin
  P := PWord(Str);
  while (P^ <> 0) do
  begin
    P^ := MakeWord(HiByte(P^), LoByte(P^));
    Inc(P);
  end;
end;

function IsHex(const S: WideString): Boolean;
var
  C, Start: Integer;
begin
  if S = '' then
  begin
    Result := False;
    Exit;
  end;

  Result := True;
  if S[1] = '#' then
    Start := 2
  else
    Start := 1;
  for C := Start to Length(S) do
    if not (((S[C] >= '0') and (S[C] <= '9')) or
            ((S[C] >= 'A') and (S[C] <= 'F')) or
            ((S[C] >= 'a') and (S[C] <= 'f'))) then
    begin
      Result := False;
      Break;
    end;
end;

function IsDecimal(const S: WideString): Boolean;
var
  C: Integer;
begin
  Result := True;
  for C := 1 to Length(S) do
    if not ((S[C] >= '0') and (S[C] <= '9')) then
    begin
      Result := False;
      Break;
    end;
end;

function PrepareHex(S: WideString): WideString;
var
  C: Integer;
begin
  if S[1] = '#' then
    S := Copy(S, 2, Length(S));
  if Length(S) > 6 then
    S := Copy(S, 1, 6);
  while Length(S) < 6 do
    S := '0' + S;
  Result := '$';
  for C := 0 to 2 do
    Result := Result + Copy(S, 5 - C * 2, 2);
end;

function ConvertColor(const AColor: WideString): Integer;
var
  S: WideString;
begin
  if IsHex(AColor) and (not IsDecimal(AColor)) then
    S := PrepareHex(AColor)
  else
    S := AColor;
  if SysUtils.TryStrToInt(S, Result) then
    Exit;
end;

function TryStrToFloatD(const S: AnsiString; out Value: Double): Boolean;
var
  S1: AnsiString;
begin
  if DecimalSeparator <> ',' then
    S1 := StringReplace(S, ',', DecimalSeparator)
  else
    S1 := S;
  if DecimalSeparator <> '.' then
    S1 := StringReplace(S1, '.', DecimalSeparator);
  Result := TryStrToFloat(S1, Value);
  if not Result then
    Value := 0;
end;

function GetIDataObjectData(const DataObj: IDataObject; const Format:
  TClipFormat; out Medium: TStgMedium): HRESULT;
var
  LobjEnum: TFormatEnum;
begin
  LobjEnum := TFormatEnum.Create(DataObj);
  try
    if not LobjEnum.HasFormat(Format) then
      Result := E_FAIL
    else
      Result := DataObj.GetData(LobjEnum.GetFormatETC(Format), Medium);
  finally
    LobjEnum.Free;
  end;
end;

function GetSomeData(const PData: Pointer; const HGlobal: THandle;
  Format: TClipFormat; const DataSize: Integer): WideString;
var
  S: AnsiString;
begin
  Result := '';

  if Format in [CF_UNICODETEXT, CF_MWKTEXT] then
    Result := PWideChar(PData)
  else
  begin
    S := PChar(PData);
    Result := S;
  end;
end;
{$ENDREGION}

{ TMWKEdit }
{$Region TMWKEdit}

function TMWKEdit.CreateXMLFromPlain(S: WideString; StyleSpan: TSpan): IXMLDOMDocument;
var
  Paragraph, Node: IXMLDOMNode;
  C: Integer;

  procedure AddSpan(const S: WideString);
  var
    T: WideString;
    Node: IXMLDOMNode;
    C: Integer;
  begin
    if S = '' then
      Exit;
    Node := AddChildNode(Paragraph, 'span');

    C := CountFirstSpaces(S);
    if C > 0 then
      SetAttribute(Node, 'sb', C);

    C := CountLastSpaces(S);
    if C > 0 then
      SetAttribute(Node, 'sa', C);

    T := Trim(S);

    if T <> '' then
    try
      Node.text := T;
    except
      ShowMessage(T);
    end;

    if Assigned(StyleSpan) then
      SetXMLAttributes(Node, StyleSpan);
  end;

  procedure AddTextWithTab(const S: WideString);
  var
    C: Integer;
    St: WideString;
  begin
    St := S;
    C := Pos(cTab, St);
    while C > 0 do
    begin
      AddSpan(Copy(St, 1, C - 1));
      AddChildNode(Paragraph, 'tab');
      System.Delete(St, 1, C);
      C := Pos(cTab, St);
    end;
    if St <> '' then
      AddSpan(St);
  end;
  
  procedure AddText(const S: WideString);
  var
    C: Integer;
    St: WideString;
  begin
    St := S;
    C := Pos(#$B, St);
    while C > 0 do
    begin
      AddTextWithTab(Copy(St, 1, C - 1));
      AddChildNode(Paragraph, 'br');
      System.Delete(St, 1, C);
      C := Pos(#$B, St);
    end;
    if St <> '' then
      AddTextWithTab(St);
  end;

begin
  S := StringReplace(S, #13#10, #13);
  Result := CreateXML(True, 'mwkedit');
  if S = '' then
    Exit;

  Paragraph := AddChildNode(Result.documentElement, 'p');
  if Assigned(StyleSpan) then
    SetXMLAttributes(Paragraph, StyleSpan);
  while S <> '' do
  begin
    if S = #13 then
      Break;
    if S[1] = cTab then
    begin
      Node := AddChildNode(Paragraph, 'tab');
      S := Copy(S, 2, MaxInt);
    end
    else
    begin
      C := Pos(#13, S);
      if C = 0 then
      begin
        AddText(S);
        S := '';
      end
      else
      begin
        AddText(Copy(S, 1, C - 1));
        Paragraph := AddChildNode(Result.documentElement, 'p');
        if Assigned(StyleSpan) then
          SetXMLAttributes(Paragraph, StyleSpan);
        S := Copy(S, C + 1, MaxInt);
      end;
    end;
  end;
end;

procedure TMWKEdit.Clear;
begin
  FinishInsert;
  FUndoRedo.Clear;
  FLineCount := 0;
  FLastEmpty := True;
  FItems.Clear;
  FLineStarts.Clear;
  FLinePos := 0;
  FText := '';
  FSelStart := 1;
  FSelLength := 0;
  FGlobalIndent := 0;
  CalcSelection;
  Invalidate;
  DoOnChange;
  DoOnSelectionChange;
end;

procedure TMWKEdit.SelectAll;
begin
  FinishInsert;
  FSelStart := 1;
  SelLength := MaxInt;
  FLeftDown := False;
end;

procedure TMWKEdit.ClearSelection;
begin
  SelLength := 0;
end;

procedure TMWKEdit.CopyToClipboard;
var
  SFormat, SPlain: WideString;
begin
  FinishInsert;
  if SelLength = 0 then
    Exit;
  SFormat := CollectAsXML(True, False);

  SPlain := CollectAsXML(True, True);

  if not OpenClipboard(0) then
    Exit;
  try
    EmptyClipboard;
    WriteToClipboardW(CF_UNICODETEXT, SPlain);
    WriteToClipboardW(CF_MWKTEXT, SFormat);
  finally
    CloseClipboard;
  end;
end;

constructor TMWKEdit.Create(AOwner: TComponent);
begin
  inherited;
  FLineBreak := True;
  FMultiLine := True;
  FShowScrollBars := True;
  FLineStarts := TList.Create;
  FIndentSize := 30;
  FTextBeforeInsert := nil;
  FInsertTimer := TTimer.Create(Self);
  FInsertTimer.Enabled := False;
  FInsertTimer.Interval := 350;
  FInsertTimer.OnTimer := OnInsertTimer;
  FUndoRedo := CreateUndoRedo;
  UINSERT := FUndoRedo.RegisterAction('INSERTTEXT');
  FUndoRedo.RegisterHandler(UINSERT, GetStandardNames, AddInsert, UndoInsert);
  UINSERTIMAGE := FUndoRedo.RegisterAction('INSERTIMAGE');
  FUndoRedo.RegisterHandler(UINSERTIMAGE, GetStandardNames, AddInsert, UndoInsert);
  UINSERTSECTION := FUndoRedo.RegisterAction('INSERTSECTION');
  FUndoRedo.RegisterHandler(UINSERTSECTION, GetStandardNames, AddInsert, UndoInsert);
  UDELETE := FUndoRedo.RegisterAction('DELETETEXT');
  FUndoRedo.RegisterHandler(UDELETE, GetStandardNames, AddDelete, UndoDelete);
  UCUT := FUndoRedo.RegisterAction('CUTTEXT');
  FUndoRedo.RegisterHandler(UCUT, GetStandardNames, AddDelete, UndoDelete);
  USTYLE := FUndoRedo.RegisterAction('STYLETEXT');
  FUndoRedo.RegisterHandler(USTYLE, GetStandardNames, AddStyle, UndoStyle);
  USPELLCHECK := FUndoRedo.RegisterAction('SPELLCHECKTEXT');
  FUndoRedo.RegisterHandler(USPELLCHECK, GetStandardNames, AddInsert, UndoInsert);
  UINDENT := FUndoRedo.RegisterAction('INDENTPARA');
  FUndoRedo.RegisterHandler(UINDENT, GetStandardNames, AddIndent, UndoIndent);

  FBrush := TBrush.Create;
  FBrush.Color := clWindow;
  CF_MWKTEXT := RegisterClipboardFormat('MWKTEXT');
  CF_RTF := RegisterClipboardFormat('Rich Text Format');
  CF_HTML := RegisterClipboardFormat('HTML Format');

  FFontHeight := -10;
  FFontSize := 8;
  FFontName := 'Arial';
  FFontColor := clWindowText;
  FBackColor := FBrush.Color;
  FBold := False;
  FItalic := False;
  FUnderline := False;
  FStrikeOut := False;
  FSpaced := False;
  FSuper := False;
  FSub := False;
  FIndex := False;
  FTabWidth := 40;
  FItems := TSpanList.Create;
  FLinePos := 0;
  FText := '';

  FShowTags := True;
  FShowSpellErrors := True;
  FBorderStyle := bsSingle;
  FVerticalScrollBar := TScrollBar.Create(Self);
  FVerticalScrollBar.Name := 'EditScrollBarV';
  FVerticalScrollBar.TabStop := False;
  FVerticalScrollBar.Kind := sbVertical;
  FVerticalScrollBar.Parent := Self;
  FVerticalScrollBar.Visible := False;
  FVerticalScrollBar.Align := alRight;
  FVerticalScrollBar.OnChange := DoOnVerticalScrollBarChange;

  FHorizontalScrollBar := TScrollBar.Create(Self);
  FHorizontalScrollBar.Name := 'EditScrollBarH';
  FHorizontalScrollBar.TabStop := False;
  FHorizontalScrollBar.Kind := sbHorizontal;
  FHorizontalScrollBar.Parent := Self;
  FHorizontalScrollBar.Visible := False;
  FHorizontalScrollBar.Align := alBottom;
  FHorizontalScrollBar.OnChange := DoOnHorizontalScrollBarChange;

  FSpellChecker := nil;
  FSpellCheckerName := '';

  FSpellCheckForm := TSpellCheckForm.Create(Self);

  if not (csDesigning in ComponentState) then
    FDropTarget := TMWKEditDropTarget.Create(Self);
  FOLEDragDrop := False;
end;

procedure TMWKEdit.CutToClipboard;

  procedure DoDelete;
  var
    L: Integer;
    OldStart: Integer;
    NewStart: Integer;
  begin
    if FSelLength = 0 then
      Exit;
    if FRollBack = 0 then
    begin
      FinishInsert;
      FUndoRedo.AddActionToUndo(UCUT, atUndo, Integer(Self), 0);
    end;
    L := Max(1, FSelLength);

    DeleteSelected(True, True);
    OldStart := FSelStart;

    if FRollBack = 0 then
    begin
      Inc(FUpdateCount);
      NewStart := FSelStart;
      SelStart := OldStart;
      SelLength := 0;
      FUndoRedo.AddActionToUndo(UCUT, atRedo, Integer(Self), L);
      SelStart := NewStart;
      SelLength := 0;
      Dec(FUpdateCount);
    end;
    Invalidate;
    DoOnSelectionChange;
  end;

begin
  CopyToClipboard;
  DoDelete;
end;

procedure TMWKEdit.DeleteSelected(WithEvent, WithRebuild: Boolean);
var
  C: Integer;
begin
  if FSelLength > 0 then
  begin
    for C := 0 to FItems.Count - 1 do
      FItems[C].DeleteSelected;

    RemoveEmptySpans;

    FSelLength := 0;
    FOldCaretX := -1;
    RemoveSelection;
    if WithRebuild then
    begin
      ReBuildLines(FStartLine - 1, FEndLine + 1);
      CalcSpellCheck(False);
    end;
    BeginUpdate;
    CalcSelection;
    EndUpdate(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    if WithEvent then
      DoOnChange;
    DoOnSelectionChange;
  end;
end;

destructor TMWKEdit.Destroy;
begin
  FInsertTimer.Enabled := False;
  FInsertTimer.Free;
  FUndoRedo := nil;
  SetOLEDragDrop(False);
  FDropTarget := nil;
  FSpellCheckForm.Free;
  FreeAndNil(FSpellChecker);
  FVerticalScrollBar.Free;
  FHorizontalScrollBar.Free;
  FLinePos := 0;
  FText := '';
  FSelStart := 1;
  FSelLength := 0;
  FItems.Free;
  FLineStarts.Free;
  FBrush.Free;
  inherited;
end;

function TMWKEdit.GetRedoText: WideString;
begin
  Result := FUndoRedo.RedoText;
end;

function TMWKEdit.GetRedoTexts: IWideStringList;
begin
  Result := TWideStringList.Create;
  FUndoRedo.FillRedoTexts(Result);
end;

function TMWKEdit.GetBackColor: TColor;
begin
  Result := FBackColor;
end;

function TMWKEdit.GetBaselineOfLine(const Index: Integer): Integer;
var
  C: Integer;
  Start, Stop: Integer;
begin
  Result := 0;
  if (Index < 0) or (Index >= FLineStarts.Count) then
    Exit;

  Start := GetLineStart(Index);
  if (Index < FLineStarts.Count - 1) then
    Stop := GetLineStart(Index + 1) - 1
  else
    Stop := FItems.Count - 1;

  for C := Start to Stop do
    if (C >= 0) and (C < FItems.Count) then
      Result := Max(Result, FItems[C].Baseline);
end;

function TMWKEdit.GetBold: Boolean;
begin
  Result := FBold;
end;

function TMWKEdit.GetColor: TColor;
begin
  Result := FBrush.Color;
end;

function TMWKEdit.GetIndent: Integer;
var
  C: Integer;
begin
  if FCurrentSpan = 0 then
  begin
    Result := FItems[0].Indent;
    Exit;
  end;
  
  for C := FCurrentSpan downto 0 do
  begin
    if FItems[C].NewParagraph then
    begin
      if C < FItems.Count - 1 then
      begin
        Result := FItems[C + 1].Indent;
        Exit;
      end;
    end;
  end;
  Result := FGlobalIndent;
end;

function TMWKEdit.GetIndex: Boolean;
begin
  Result := FIndex;
end;

function TMWKEdit.GetInsertState: TInsertState;
begin
  Result := FInsertState;
end;

function TMWKEdit.GetItalic: Boolean;
begin
  Result := FItalic;
end;

function TMWKEdit.GetLength: Integer;
var
  C: Integer;
begin
  Result := 0;
  for C := 0 to FItems.Count - 1 do
    Result := Result + FItems[C].Length;
end;

function TMWKEdit.GetLineCount: Integer;
begin
  Result := FLineCount;
end;

function TMWKEdit.GetLineOfY(const Y: Integer): Integer;
var
  C, D: Integer;
  Height: Integer;
begin
  D := 0;
  C := 0;
  while (D <= Y) and (C < FLineStarts.Count) do
  begin
    Height := GetHeightOfLine(C);
    if (D + Height <= Y) then
    begin
      Inc(D, Height);
      Inc(C);
    end else
      Break;
  end;
  Result := C;
end;

function TMWKEdit.GetLineStart(const Index: Integer): Integer;
begin
  if Index < 0 then
    Result := 0
  else
    if Index >= FLineStarts.Count then
    begin
      if FLineStarts.Count > 0 then
        Result := Integer(FLineStarts.Last)
      else
        Result := 0;
    end
    else
      Result := Integer(FLineStarts[Index]);
end;

function TMWKEdit.GetOnChange: TNotifyEvent;
begin
  Result := FOnChange;
end;

function TMWKEdit.GetPlainText: WideString;
begin
  Result := CollectAsXML(False, True);
end;

function TMWKEdit.GetSelectedParagraphs: TSpanRange;
var
  SR: TSpanRange;
  C: Integer;
begin
  if (FSelLength = 0) and
     ((FItems.Count > 0) and
      (FCurrentSpan = FItems.Count - 1) and
      (FPosInSpan > FItems.Last.Length) and
      (FItems.Last.NewParagraph)) then
  begin
    Result.Start := -1;
    Exit;
  end;

  if FItems.Count = 0 then
  begin
    Result.Start := -1;
    Exit;
  end;
  
  SR := CalculateParagraphRange;
  Result.Start := 0;
  C := 0;
  while (C < SR.Start) do
  begin
    if FItems[C].NewParagraph then
      Inc(Result.Start);
    Inc(C);
  end;

  Result.Stop := Result.Start;
  while (C < FItems.Count) and (C < SR.Stop) do
  begin
    if FItems[C].NewParagraph then
      Inc(Result.Stop);
    Inc(C);
  end;
end;

function TMWKEdit.GetSelectedPlainText: WideString;
begin
  if SelLength = 0 then
    Result := ''
  else
    Result := CollectAsXML(True, True);
end;

function TMWKEdit.GetSelectedText: WideString;
begin
  if SelLength = 0 then
    Result := ''
  else
    Result := CollectAsXML(True, False);
end;

function TMWKEdit.GetSelLength: Integer;
begin
  Result := FSelLength;
end;

function TMWKEdit.GetSelStart: Integer;
begin
  Result := FSelStart;
end;

function TMWKEdit.GetSpaced: Boolean;
begin
  Result := FSpaced;
end;

function TMWKEdit.GetStrikeOut: Boolean;
begin
  Result := FStrikeOut;
end;

function TMWKEdit.GetSub: Boolean;
begin
  Result := FSub;
end;

function TMWKEdit.GetSuper: Boolean;
begin
  Result := FSuper;
end;

function TMWKEdit.GetUnderline: Boolean;
begin
  Result := FUnderline;
end;

function TMWKEdit.GetUndoText: WideString;
begin
  Result := FUndoRedo.UndoText;
end;

function TMWKEdit.GetUndoTexts: IWideStringList;
begin
  Result := TWideStringList.Create;
  FUndoRedo.FillUndoTexts(Result);
end;

procedure TMWKEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
  Form: TWinControl;

  procedure DoDelete;
  var
    L: Integer;
    OldStart: Integer;
    NewStart: Integer;
  begin
    if FRollBack = 0 then
    begin
      FinishInsert;
      FUndoRedo.AddActionToUndo(UDELETE, atUndo, Integer(Self), Integer(Key = VK_DELETE));
    end;
    L := Max(1, FSelLength);

    if FSelLength > 0 then
      DeleteSelected(True, True)
    else
      Delete(Key = VK_DELETE);
    OldStart := FSelStart;

    if FRollBack = 0 then
    begin
      Inc(FUpdateCount);
      NewStart := FSelStart;
      SelStart := OldStart;
      SelLength := 0;
      FUndoRedo.AddActionToUndo(UDELETE, atRedo, Integer(Self), L);
      SelStart := NewStart;
      SelLength := 0;
      Dec(FUpdateCount);
      DoOnSelectionChange;
    end;

    Invalidate;
  end;

begin
  inherited;

  if ((Key = VK_DELETE) and (ssShift in Shift)) or
     ((Key = Ord('X')) and (ssCtrl in Shift))  then
  begin
    FinishInsert;
    CutToClipboard;
    Exit;
  end;

  if ((Key = VK_INSERT) and (ssCtrl in Shift)) or
     ((Key = Ord('C')) and (ssCtrl in Shift)) then
  begin
    FinishInsert;
    CopyToClipboard;
    Exit;
  end;

  if ((Key = VK_INSERT) and (ssShift in Shift)) or
     ((Key = Ord('V')) and (ssCtrl in Shift))then
  begin
    FinishInsert;
    PasteFromClipboard;
    Exit;
  end;

  if FRightToLeft then
  begin
    if Key = VK_LEFT then
      Key := VK_Right
    else
      if Key = VK_RIGHT then
        Key := VK_LEFT;
  end;

  if ssShift in Shift then
  begin
    if Key = VK_LEFT then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetSelection(ncPrevWord)
      else
        SetSelection(ncLeft);
      Exit;
    end;

    if Key = VK_RIGHT then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetSelection(ncNextWord)
      else
        SetSelection(ncRight);
      Exit;
    end;

    if Key = VK_UP then
    begin
      FinishInsert;
      SetSelection(ncUp);
      Exit;
    end;

    if Key = VK_DOWN then
    begin
      FinishInsert;
      SetSelection(ncDown);
      Exit;
    end;

    if Key = VK_HOME then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetSelection(ncFirst)
      else
        SetSelection(ncHome);
      Exit;
    end;

    if Key = VK_END then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetSelection(ncLast)
      else
        SetSelection(ncEnd);
      Exit;
    end;
  end;

  if not (ssShift in Shift) then
  begin
    if Key = VK_LEFT then
    begin
      if ssCtrl in Shift then
        SetCaretTo(ncPrevWord)
      else
        SetCaretTo(ncLeft);
      Exit;
    end;

    if Key = VK_RIGHT then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetCaretTo(ncNextWord)
      else
        SetCaretTo(ncRight);
      Exit;
    end;

    if Key = VK_UP then
    begin
      FinishInsert;
      SetCaretTo(ncUp);
      Exit;
    end;

    if Key = VK_DOWN then
    begin
      FinishInsert;
      SetCaretTo(ncDown);
      Exit;
    end;

    if Key = VK_HOME then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetCaretTo(ncFirst)
      else
        SetCaretTo(ncHome);
      Exit;
    end;

    if Key = VK_END then
    begin
      FinishInsert;
      if ssCtrl in Shift then
        SetCaretTo(ncLast)
      else
        SetCaretTo(ncEnd);
      Exit;
    end;

    if (Key = 65) and
       (ssCtrl in Shift) then
    begin
      FinishInsert;
      SelectAll;
      Exit;
    end;

    if ((Key = VK_DELETE) or (Key = VK_BACK)) and not FReadOnly then
    begin
      DoDelete;
      Exit;
    end;
  end;

  if Key = VK_TAB then
  begin
    if FWantTabs = (ssCtrl in Shift) then
    begin
      FinishInsert;
      Form := GetParentForm(Self);
      if Assigned(Form) and (Form <> Self) then
        if ssShift in Shift then
          Form.Perform(WM_NEXTDLGCTL, 1, 0)
        else
          Form.Perform(WM_NEXTDLGCTL, 0, 0);
    end else
      if Shift = [] then
        AddCharacter(cTab);
  end;

  if (Key = VK_SHIFT) or (Key = VK_MENU) or (Key = VK_CONTROL) or
     (Key = VK_LSHIFT) or (Key = VK_RSHIFT) or
     (Key = VK_LMENU) or (Key = VK_RMENU) or
     (Key = VK_LCONTROL) or (Key = VK_RCONTROL) then
    Exit;

  if (Key = VK_RETURN) and (ssCtrl in Shift) then
  begin
    Key := 0;
    InsertSection;
  end;

  if (Key = VK_RETURN) and (ssShift in Shift) then
    Key := 14;

  if (Key = 189) and (ssCtrl in Shift) then
    AddCharacter(cHyphen);

  if Key = VK_INSERT then
  begin
    FinishInsert;
    if FInsertState = isInsert then
      FInsertState := isOverWrite
    else
      FInsertState := isInsert;
    if Assigned(FOnToggleInsert) then
      FOnToggleInsert(Self);
    Exit;
  end;
  FShiftDown := ssShift in Shift;
  FCtrlDown := ssCtrl in Shift;
end;

procedure TMWKEdit.WMChar(var Message: TWMChar);
begin
  if not DoKeyPress(Message) then
    inherited;
end;

procedure TMWKEdit.WMCut(var Message: TWMCut);
begin
  CutToClipboard;
end;

procedure TMWKEdit.WMDestroy(var Message: TWMDestroy);
begin
  if Assigned(FDropTarget) then
    FDropTarget.Active := False;
  inherited;
end;

procedure TMWKEdit.WMCopy(var Message: TWMCopy);
begin
  CopyToClipboard;
end;

procedure TMWKEdit.WMPaste(var Message: TWMPaste);
begin
  PasteFromClipboard;
end;

procedure TMWKEdit.WMClear(var Message: TWMClear);
begin
  DeleteSelected(True, True);
end;

procedure TMWKEdit.WMMouseWheel(var Message: TWMMouseWheel);
var
  D: Integer;
begin
  if (not FShowVerticalScrollBar) or
     (Message.Keys and MK_LBUTTON = MK_LBUTTON) or
     (Message.Keys and MK_MBUTTON = MK_MBUTTON) or
     (Message.Keys and MK_RBUTTON = MK_RBUTTON) then
    Exit;
  if (Message.Keys and MK_CONTROL = MK_CONTROL) then
    D := FVerticalScrollBar.LargeChange
  else
    D := FVerticalScrollBar.SmallChange;

  if Message.WheelDelta > 0 then
    D := -D;

  FVerticalScrollBar.Position := FVerticalScrollBar.Position + D;
end;

procedure TMWKEdit.KeyPress(var Key: WideChar);
begin
  if Assigned(FOnKeyPress) then
    FOnKeyPress(Self, Key);

  if Key in [WideChar(#8), WideChar(cTab)] then
    Exit;

  if (Key = #13) and FShiftDown then
    Key := #14;

  if (Key = #12) then
    Key := cHyphen;

  if (Key = ' ') and (FCtrlDown) then
    Key := cProtectSpace;
                                                                   
  if (Key < #32) and (Key <> #12) and (Key <> #13) and (Key <> #14) then
    Exit;

  AddCharacter(Key);
end;

procedure TMWKEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
  FShiftDown := ssShift in Shift;
  FCtrlDown := ssCtrl in Shift;
end;

procedure TMWKEdit.DoEnter;
begin
  SetCaretPos(True);
  inherited;
end;

procedure TMWKEdit.DoExit;
begin
  FinishInsert;
  SetCaretPos(False);
  inherited;
end;

function TMWKEdit.LoadFromFile(const FileName: WideString): Integer;
var
  S: TFileStream;
begin
  Result := 0;
  S := TFileStream.Create(Filename, fmOpenRead or fmShareDenyNone);
  try
    try
      LoadFromStream(S);
    except
      Result := 1;
    end;
  finally
    S.Free;
  end;
end;

procedure TMWKEdit.LoadFromStream(const Stream: TStream);
var
  DataLeft: Integer;
  StreamCharSet: TStreamCharSet;
  SW: WideString;
  SA: AnsiString;
  Title: WideString;
  XML: IXMLDOMDocument;
  RTF: TRTFToXML;
begin
  FinishInsert;
  try
    StreamCharSet := DetectCharacterSet(Stream);

    DataLeft := Stream.Size - Stream.Position;

    case StreamCharSet of
      csAnsi:
      begin
        SetLength(SA, DataLeft div SizeOf(AnsiChar));
        Stream.Read(PAnsiChar(SA)^, DataLeft);
        SW := SA;
      end;

      csUnicode, csUnicodeSwapped:
      begin
        if DataLeft < SizeOf(WideChar) then
          SW := ''
        else
        begin
          SetLength(SW, DataLeft div SizeOf(WideChar));
          Stream.Read(PWideChar(SW)^, DataLeft);
          if StreamCharSet = csUnicodeSwapped then
            StrSwapByteOrder(PWideChar(SW));
        end;
      end;

      csUtf8:
      begin
        SetLength(SA, DataLeft div SizeOf(AnsiChar));
        Stream.Read(PAnsiChar(SA)^, DataLeft);
        SW := UTF8Decode(SA);
      end;
    end;
  finally
  end;

  SW := StringReplace(SW, #13#10, #13);
  SW := StringReplaceMultiple(SW, [#10, #13], [sLineBreak, sLineBreak]);
  RTF := TRTFToXML.Create;
  try
    XML := RTF.Parse(SW, Title);
  finally
    RTF.Free;
  end;

  if not Assigned(XML) then
  begin
    if IsValidXML(SW, 'mwkedit') then
    begin
      XML := CreateXML(True, '');
      XML.loadXML(SW);
    end else
      XML := CreateXMLFromPlain(SW, nil);
  end;

  SetXMLText(XML);

  FUndoRedo.Clear;
  DoOnSelectionChange;
end;

function TMWKEdit.ModifyOLEDropEffect(const GrfKeyState: Integer;
  const PT: TPoint; var dwEffect: Integer): HRESULT;
begin
  Result := S_OK;

  if Bool(GrfKeyState and MK_SHIFT) and (not FReadOnly) then
    dwEffect := DROPEFFECT_MOVE
  else
    dwEffect := DROPEFFECT_COPY;

  if FOLEFormat = CF_HDROP then
    if dwEffect = DROPEFFECT_MOVE then
      dwEffect := DROPEFFECT_LINK;
end;

procedure TMWKEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited;
  if csDesigning in ComponentState then
    Exit;

  if Assigned(GetParentForm(Self)) then
    SetFocus
  else
    Windows.SetFocus(Handle);

  if (Button <> mbLeft) then
    Exit;

  FLeftDown := True;

  if not (ssShift in Shift) then
    SelLength := 0;

  if FRightToLeft then
    X := FClientRect.Right + FClientRect.Left - X;

  CalcMousePos(X, Y, True);

  if not (ssShift in Shift) then
  begin
    FNegativeSelection := False;
    FCommingFromRight := False;
  end;

  MSelStart := GetMousePosAbsolute;

  MDouble := (ssDouble in Shift);
  if (ssDouble in Shift) and not (ssShift in Shift) then
  begin
    FNegativeSelection := False;
    SelStart := GetWordStart(MSelStart);
    SelLength := GetWordLength(FSelStart);
    FOldCaretX := -1;
    Exit;
  end;

  if ssShift in Shift then
  begin
    if FNegativeSelection then
    begin
      X := MSelStart - FSelStart - FSelLength;
      FSelStart := FSelStart + FSelLength;
      MSelStart := FSelStart;
      SelLength := X;
    end
    else
    begin
      SelLength := MSelStart - FSelStart;
      if FNegativeSelection then
        MSelStart := FSelStart + FSelLength;
    end;
  end else
    SelStart := MSelStart;
  FOldCaretX := -1;
end;

procedure TMWKEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  C: Integer;
  OldNegative: Boolean;
begin
  inherited;
  if csDesigning in ComponentState then
    Exit;

  if (X >= FClientRect.Left) and (X < FClientRect.Right) and
     (Y >= FClientRect.Top) and (Y < fClientRect.Bottom) then
    Cursor := crIBeam
  else
    Cursor := crDefault;

  if (not FLeftDown) or MDouble then
    Exit;

  OldNegative := FNegativeSelection;
  if FRightToLeft then
    X := FClientRect.Right + FClientRect.Left - X;

  CalcMousePos(X, Y, False);

  if SelLength = Abs(GetMousePosAbsolute - MSelStart) then
  begin
    FNegativeSelection := OldNegative;
    Exit;
  end;
  FSelStart := MSelStart;

  C := GetMousePosAbsolute - MSelStart;
  SelLength := C;
  if C > 0 then
    FNegativeSelection := False;
end;

procedure TMWKEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  FLeftDown := False;
  inherited;
end;

function TMWKEdit.OLEHasSupportedFormat(const DataObj: IDataObject;
  const Formats: array of TClipFormat; var Format: TClipFormat): Boolean;
var
  LIntLoop: Integer;
  LObjEnum: TFormatEnum;
begin
  Result := False;
  LObjEnum := TFormatEnum.Create(DataObj);
  try
    if System.Length(Formats) > 0 then
      for LIntLoop := Low(Formats) to High(Formats) do
        if LObjEnum.HasFormat(Formats[LIntLoop]) then
        begin
          Format := Formats[LIntLoop];
          Result := True;
          Break;
        end;
  finally
    LObjEnum.Free;
  end;
end;

procedure TMWKEdit.OnInsertTimer(Sender: TObject);
begin
  FInsertTimer.Enabled := False;
  FinishInsert;
end;

procedure TMWKEdit.Paint;
var
  C, Start, Stop: Integer;
  X, Y: Integer;
  CR, OCR: TRect;
  RGN: HRGN;
  Brush: HBrush;
begin
  if FShowVerticalScrollBar then
    Y := -FVerticalScrollBar.Position
  else
    Y := 0;

  if FShowHorizontalScrollBar then
    X := -FHorizontalScrollBar.Position
  else
    X := 0;

  OCR := Canvas.ClipRect;
  CR := FClientRect;

  RGN := CreateRectRGN(CR.Left, CR.Top, CR.Right, CR.Bottom);
  SelectClipRgn(Canvas.Handle, RGN);
  DeleteObject(RGN);

  Start := GetLineOfY(-Y);
  Stop := Min(FLineCount - 1, GetLineOfY(-Y + FClientRect.Bottom) + 1);

  if FItems.Count = 0 then
    Stop := -1;

  if FLastEmpty and (Stop = FLineCount - 1) then
    Dec(Stop);

  Inc(Y, FClientRect.Top + GetHeightToLine(Start - 1));

  if FRightToLeft then
    X := FClientRect.Right - X
  else
    X := FClientRect.Left + X;

  for C := Start to Stop do
  begin
    PaintLine(C, X, Y);
    Inc(Y, GetHeightOfLine(C));
  end;

  CR.Top := Y + 2;
  FillRect(Canvas.Handle, Expand(CR, 2), FBrush.Handle);

  if FShowDragCaret then
    PaintDragCaret;

  RGN := CreateRectRGN(OCR.Left, OCR.Top, OCR.Right, OCR.Bottom);
  SelectClipRgn(Canvas.Handle, RGN);
  DeleteObject(RGN);
  CR := Expand(FClientRect);
  FrameRect(Canvas.Handle, CR, FBrush.Handle);
  CR := Expand(CR);
  FrameRect(Canvas.Handle, CR, FBrush.Handle);

  if FVerticalScrollBar.Visible and FHorizontalScrollBar.Visible then
  begin
    CR := Rect(FVerticalScrollBar.Left, FVerticalScrollBar.Top + FVerticalScrollBar.Height,
      FVerticalScrollBar.Left + FVerticalScrollBar.Width, FHorizontalScrollBar.Top + FHorizontalScrollBar.Height);
    Brush := CreateSolidBrush(ColorToRGB(clMenu));
    FillRect(Canvas.Handle, CR, Brush);
    DeleteObject(Brush);
  end;
end;

procedure TMWKEdit.PaintDragCaret;
begin
  Canvas.Pen.Color := FBackColor;
  Canvas.Pen.Mode := pmXor;
  Canvas.MoveTo(FDragCaretPos.X, FDragCaretPos.Y);
  Canvas.LineTo(FDragCaretPos.X, FDragCaretPos.Y + GetHeightOfLine(FindLineOfSpan(FDragSpan)));
  Canvas.Pen.Mode := pmCopy;
end;

function TMWKEdit.DoKeyPress(var Message: TWMKey): Boolean;
var
  Form: TWinControl;
  Ch: WideChar;
  Key: Char;
begin
  Result := True;
  Form := GetParentForm(Self);
  if Assigned(Form) and (Form <> Self) and TCustomForm(Form).KeyPreview then
  begin
    if not (csNoStdEvents in Form.ControlStyle) then
      with Message do
      begin
        Key := Char(CharCode);
        if Assigned(TForm(Form).OnKeyPress) then
          TForm(Form).OnKeyPress(Form, Key);
        CharCode := Word(Key);
        if Char(CharCode) = #0 then Exit;
      end;
  end;

  if not (csNoStdEvents in ControlStyle) then
    with Message do
    begin
      Ch := WideChar(CharCode);
      KeyPress(Ch);
      CharCode := Word(Ch);
      if Char(CharCode) = #0 then Exit;
    end;
  Result := False;
end;

procedure TMWKEdit.WMEraseBackGround(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TMWKEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  if SelLength > 0 then
    Invalidate;
  if (FSelStart = 0) and (FItems.Count > 0) then
    FSelStart := 1;

  BeginUpdate;
  try
    SetCaretPos(True);
  finally
    EndUpdate(False);
  end;
end;

procedure TMWKEdit.WMKillFocus(var Message: TWMSetFocus);
begin
  inherited;
  FLeftDown := False;
  HideCaret(Handle);
  DestroyCaret;
  Invalidate;
end;

procedure TMWKEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTARROWS or DLGC_WANTALLKEYS or
    DLGC_WANTCHARS or DLGC_WANTTAB;
end;

procedure TMWKEdit.RemoveEmptySpans;
var
  C: Integer;
begin
  for C := FItems.Count - 1 downto 0 do
  begin
    if ((FItems[C] is TGraphicSpan) and
        (TGraphicSpan(FItems[C]).Picture = nil)) or
       (FItems[C].Length = 0) then
      FItems.Delete(C);
  end;
end;

procedure TMWKEdit.SetVerticalScrollBar;
var
  Max, Position: Integer;
begin
  if not FShowVerticalScrollBar then
  begin
    FVerticalScrollBar.OnChange := nil;
    FVerticalScrollBar.PageSize := 0;
    FVerticalScrollBar.SetParams(0, 0, 0);
    FVerticalScrollBar.SmallChange := 1;
    FVerticalScrollBar.LargeChange := 1;
    FVerticalScrollBar.OnChange := DoOnVerticalScrollBarChange;
    Exit;
  end;

  Max := FTextHeight - 1;
  Position := Min(FVerticalScrollBar.Position, Max);
  if (Max < 0) or (Position > Max) then
    Beep;
  FVerticalScrollBar.OnChange := nil;
  FVerticalScrollBar.PageSize := 1;
  FVerticalScrollBar.SetParams(Position, 0, Max);
  FVerticalScrollBar.PageSize := FClientRect.Bottom - FClientRect.Top;
  FVerticalScrollBar.SmallChange := Round(Max / FLineCount);
  FVerticalScrollBar.LargeChange := (FClientRect.Bottom - FClientRect.Top) div FVerticalScrollBar.SmallChange * FVerticalScrollBar.SmallChange;

  if FVerticalScrollBar.Position > FVerticalScrollBar.Max - FVerticalScrollBar.PageSize + 1 then
    FVerticalScrollBar.Position := FVerticalScrollBar.Max - FVerticalScrollBar.PageSize + 1;

  FVerticalScrollBar.OnChange := DoOnVerticalScrollBarChange;
end;

procedure TMWKEdit.SetHorizontalScrollBar;
var
  Max, Position: Integer;
begin
  if not FShowHorizontalScrollBar then
  begin
    FHorizontalScrollBar.OnChange := nil;
    FHorizontalScrollBar.PageSize := 0;
    FHorizontalScrollBar.SetParams(0, 0, 0);
    FHorizontalScrollBar.SmallChange := 1;
    FHorizontalScrollBar.LargeChange := 1;
    FHorizontalScrollBar.OnChange := DoOnHorizontalScrollBarChange;
    Exit;
  end;

  Max := FMaxLineSize - 1;
  Position := Min(FHorizontalScrollBar.Position, Max);
  if (Max < 0) or (Position > Max) then
    Beep;
  FHorizontalScrollBar.OnChange := nil;
  FHorizontalScrollBar.PageSize := 1;
  FHorizontalScrollBar.SetParams(Position, 0, Max);
  FHorizontalScrollBar.PageSize := FClientRect.Right - FClientRect.Left;
  FHorizontalScrollBar.SmallChange := Round(5);
  FHorizontalScrollBar.LargeChange := (FClientRect.Right - FClientRect.Left) div FHorizontalScrollBar.SmallChange * FHorizontalScrollBar.SmallChange;

  if FHorizontalScrollBar.Position > FHorizontalScrollBar.Max - FHorizontalScrollBar.PageSize + 1 then
    FHorizontalScrollBar.Position := FHorizontalScrollBar.Max - FHorizontalScrollBar.PageSize + 1;

  FHorizontalScrollBar.OnChange := DoOnHorizontalScrollBarChange;
end;

procedure TMWKEdit.ToggleScrollBar;
begin
  if IsAligning then
    Exit;

  if FMaxLineSize > ClientWidth then
  begin
    if FShowHorizontalScrollBar then
      SetHorizontalScrollBar;

    if not FShowHorizontalScrollBar then
    begin
      FShowHorizontalScrollBar := not FLineBreak;
      FHorizontalScrollBar.Visible := FShowHorizontalScrollBar and FShowScrollBars;
      SetHorizontalScrollBar;
      IsAligning := True;
      CalcAreas;
      IsAligning := False;
      SetScrollBarPositions;
    end;
  end
  else
  if FShowHorizontalScrollBar then
  begin
    FShowHorizontalScrollBar := False;
    FHorizontalScrollBar.Visible := False;
    SetVerticalScrollBar;
    SetHorizontalScrollBar;
    IsAligning := True;
    CalcAreas;
    IsAligning := False;
    SetScrollBarPositions;
    Exit;
  end;


  if (FTextHeight > FClientRect.Bottom - FCLientRect.Top) then
  begin
    if FShowVerticalScrollBar then
      SetVerticalScrollBar;

    if not FShowVerticalScrollBar then
    begin
      FShowVerticalScrollBar := True;
      FVerticalScrollBar.Visible := FShowVerticalScrollBar and FShowScrollBars;
      SetVerticalScrollBar;
      IsAligning := True;
      CalcAreas;
      IsAligning := False;
    end;
    SetScrollBarPositions;
    Exit;
  end;

  if (FTextHeight <= FClientRect.Bottom - FCLientRect.Top) and
     (FShowVerticalScrollBar) then
  begin
    FShowVerticalScrollBar := False;
    FVerticalScrollBar.Visible := False;
    SetVerticalScrollBar;
    SetHorizontalScrollBar;
    IsAligning := True;
    CalcAreas;
    IsAligning := False;
    SetScrollBarPositions;
    Exit;
  end;
end;

procedure TMWKEdit.Undo;
begin
  FinishInsert;
  FUndoRedo.Undo;
end;

procedure TMWKEdit.UpdateSpanValues(Value: TSpanValues);
begin
  FBold := Value.IsBold;
  FItalic := Value.IsItalic;
  FUnderline := Value.IsUnderline;
  FStrikeOut := Value.IsStrikeOut;
  FSpaced := Value.IsSpaced;
  FSuper := Value.IsSuper;
  FSub := Value.IsSub;
  FIndex := Value.IsIndex;
  if not FFixedColor then
  begin
    FFontColor := Value.Color;
    FBackColor := Value.BackColor;
  end;
  if not FFixedFont then
    FFontName := Value.FontName;
  if not FFixedSize then
    FFontHeight := Value.FontHeight;
  SetSizeFromHeight;
  FGlobalIndent := Value.Indent;
  DoOnSelectionChange;
end;

procedure TMWKEdit.CalcMousePos(X, Y: Integer; MouseDown: Boolean);
var
  MouseLine: Integer;
  C: Integer;
  W: Integer;
  Span: TSpan;
  WithLineBreak: Boolean;
begin
  if FItems.Count = 0 then
  begin
    FCurrentSpan := -1;
    FPosInMouseSpan := 0;
    Exit;
  end;

  Dec(X, FClientRect.Left);
  Dec(Y, FClientRect.Top);
  Inc(Y, FVerticalScrollBar.Position);
  Inc(X, FHorizontalScrollBar.Position);
  if X < 0 then
    X := 0;

  MouseLine := GetLineOfY(Y);

  C := GetLineStart(MouseLine);

  if C = FItems.Count then
  begin
    Span := FItems[C - 1];
    FCurrentMouseSpan := C - 1;
    FPosInMouseSpan := Span.Length + 1;
  end
  else
  begin                    /////
    W := 0;
    Dec(X, FItems[C].Indent * FIndentSize);
    repeat
      Span := FItems[C];

      if (C = FItems.Count - 1) or
         (W + Span.Width >= X) or
         (Span.LineBreak) then
        Break;

      Inc(W, Span.Width);

      Inc(C);
    until Span.LineBreak;

    FCurrentMouseSpan := C;
    Dec(X, W);

    if (FCurrentSpan > C) then
      FNegativeSelection := True;
    WithLineBreak := FNegativeSelection or
      ((not FNegativeSelection) and (FSelLength > 0));

    if MouseDown and (FSelLength = 0) then
      WithLineBreak := False;
    FPosInMouseSpan := Span.MeasureWidth(X, WithLineBreak);
  end;
end;

procedure TMWKEdit.CalcDragMousePos(X, Y: Integer);
var
  MouseLine: Integer;
  C: Integer;
  W: Integer;
  Span: TSpan;
begin
  if FItems.Count = 0 then
  begin
    FDragSpan := -1;
    FPosInDragSpan := 0;
    FDragCaretPos := Point(0, 0);
    Exit;
  end;

  Dec(X, FClientRect.Left);
  if X < 0 then
    X := 0;
  Dec(Y, FClientRect.Top);
  Inc(Y, FVerticalScrollBar.Position);
  Inc(X, FHorizontalScrollBar.Position);

  MouseLine := GetLineOfY(Y);

  C := GetLineStart(MouseLine);

  FDragCaretPos.Y := GetHeightToLine(Min(MouseLine, FLineCount - 1)) - Y;

  if C = FItems.Count then
  begin
    Span := FItems.Last;
    FDragSpan := C - 1;
    FPosInDragSpan := Span.Length + 1;
    FDragCaretPos.X := Span.Width;
  end
  else
  begin
    W := 0;
    repeat
      Span := FItems[C];

      if (C = FItems.Count - 1) or
         (W + Span.Width >= X) or
         (Span.LineBreak) then
        Break;

      Inc(W, Span.Width);

      Inc(C);
    until Span.LineBreak;

    FDragSpan := C;
    Dec(X, W);
    FPosInDragSpan := Span.MeasureWidth(X, False);
    FDragCaretPos.X := FClientRect.Left + W +
      Span.CalcTextWidth(Copy(Span.Text, 1, FPosInDragSpan - 1));
  end;
end;

function TMWKEdit.GetMousePosAbsolute: Integer;
var
  C, W: Integer;
begin
  W := 0;
  if FCurrentMouseSpan > FItems.Count then
  begin
    Result := 0;
    Exit;
  end;
  for C := 0 to FCurrentMouseSpan - 1 do
    Inc(W, FItems[C].Length);
  Result := W + FPosInMouseSpan;
end;

function TMWKEdit.GetMyOLEFormats: TClipFormats;
begin
  SetLength(Result, 6);
  Result[0] := CF_TEXT;
  Result[1] := CF_RTF;
  Result[2] := CF_UNICODETEXT;
  Result[3] := CF_OEMTEXT;
  Result[4] := CF_HTML;
  Result[5] := CF_MWKTEXT;
end;

procedure TMWKEdit.SetSpanValues(ChangeType: TChangeType);
var
  C: Integer;
  Span, NewSpan: TSpan;

  procedure Change;
  begin
    case ChangeType of
      ctBold:
        Span.Bold := FBold;

      ctItalic:
        Span.Italic := FItalic;

      ctUnderline:
        Span.Underline := FUnderline;

      ctStrikeOut:
        Span.StrikeOut := FStrikeOut;

      ctSpaced:
        Span.Spaced := FSpaced;

      ctSuper:
        Span.Super := FSuper;

      ctSub:
        Span.Sub := FSub;

      ctIndex:
        Span.Index := FIndex;

      ctColor:
        Span.Color := FFontColor;

      ctBackColor:
        Span.BackgroundColor := FBackColor;

      ctFontName:
        Span.FontName := FFontName;

      ctFontHeight:
        Span.FontHeight := FFontHeight;
    end;
  end;

begin
  if (csDesigning in ComponentState) or FReadOnly or FPlain then
  begin
    for C := 0 to FItems.Count - 1 do
    begin
      Span := FItems[C];
      Change;
    end;
    Exit;
  end;

  if (FSelLength = 0) and (FSelStart <> 0) then
    Exit;

  if FSelStart = 0 then
  begin
    for C := 0 to FItems.Count - 1 do
    begin
      Span := FItems[C];
      Change;
    end;
    DoOnSelectionChange;
    Exit;
  end;

  C := 0;
  while C < FItems.Count do
  begin
    Span := FItems[C];
    if Span.SelStart > 0 then
    begin
      if (Span.SelStart = 1) and (Span.SelLength = Span.Length) then
        Change;
      if Span.SelStart > 1 then
      begin
        NewSpan := Span.Split(Span.SelStart);
        FItems.Insert(C + 1, NewSpan);
      end else
      if Span.SelLength < Span.Length then
      begin
        NewSpan := Span.Split(Span.SelLength + 1);
        FItems.Insert(C + 1, NewSpan);
        Change;
      end;
    end;
    Inc(C);
  end;
  DoOnSelectionChange;
end;

function TMWKEdit.FindStartLine: Integer;
var
  C: Integer;
  Num: Integer;
  Span: TSpan;
begin
  Num := 0;
  Result := 0;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if Span.LineBreak then
      Inc(Result);
    if Num + Span.Length >= FSelStart then
    begin
      if Span.LineBreak and (Num + Span.Length >= FSelStart) then
        Dec(Result);
      Exit;
    end;
    Num := Num + Span.Length;
  end;
  FSelStart := Num + 1;
end;

procedure TMWKEdit.FinishInsert;
var
  NewStart: Integer;
  IR: TInsertRec;
begin
  if FPendingInsertFinish and (FRollBack = 0) then
  begin
    Inc(FUpdateCount);
    NewStart := FSelStart;
    SelStart := FSelStartBeforeInsert;
    SelLength := NewStart - FSelStartBeforeInsert;
    IR.Edit := Self;
    IR.OldInsertAction := TInsertAction(FOldInsertAction);
    IR.Text := FTextBeforeInsert;
    FUndoRedo.AddActionToUndo(UINSERT, atRedo, Integer(@IR), FSelLengthBeforeInsert);
    IR.Text := nil;
    SelStart := NewStart;
    SelLength := 0;
    Dec(FUpdateCount);
    FPendingInsertFinish := False;
    FTextBeforeInsert := nil;
    CalcSelection;
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnChange;
    DoOnSelectionChange;
  end;
end;

function TMWKEdit.Focused: Boolean;
begin
  Result := inherited Focused;
end;

function TMWKEdit.FindEndLine: Integer;
var
  C: Integer;
  Num: Integer;
  Span: TSpan;
  Position: Integer;
begin
  Num := 0;
  Result := 0;
  if FSelLength > 0 then
    Position := FSelStart + FSelLength - 1
  else
    Position := FSelStart;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if Span.LineBreak then
      Inc(Result);
    if Num + Span.Length >= Position then
    begin
      if Span.LineBreak and (Num + Span.Length >= Position) then
        Dec(Result);
      Exit;
    end;
    Num := Num + Span.Length;
  end;
end;

function TMWKEdit.FindLineOfSpan(Index: Integer): Integer;
begin
  Result := -1;
  if Index >= FItems.Count then
    Exit;

  for Result := FLineStarts.Count downto 0 do
    if Index >= GetLineStart(Result) then
      Exit;
  Result := FLineCount + 1;
end;

function TMWKEdit.GetCurrentSpanNum(var Position: Integer): Integer;
var
  C: Integer;
  CursorPos: Integer;
  Span: TSpan;

  procedure UpdateState;
  begin
    if (FUpdateCount > 0) or FPlain or (csDesigning in ComponentState) then
      Exit;
    FBold := Span.Bold;
    FItalic := Span.Italic;
    FUnderline := Span.Underline;
    FStrikeOut := Span.StrikeOut;
    FSpaced := Span.Spaced;
    FSuper := Span.Super;
    FSub := Span.Sub;
    FIndex := Span.Index;
    if not FFixedFont then
      FFontName := Span.FontName;

    if not FFixedSize then
    begin
      FFontHeight := Span.FontHeight;
      SetSizeFromHeight;
    end;

    if not FFixedColor then
    begin
      FFontColor := Span.Color;
      FBackColor := Span.BackgroundColor;
    end;
  end;

begin
  CursorPos := GetCursorPos;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if CursorPos - Span.Length <= 0 then
    begin
      Result := C;
      FGlobalIndent := Span.Indent;
      Position := CursorPos;

      if (Position = 1) and (C > 0) and
         (not FNegativeSelection) then
        Span := FItems[C - 1];

      UpdateState;
      Exit;
    end;
    Dec(CursorPos, Span.Length);
  end;
  Result := FItems.Count - 1;

  if FItems.Count > 0 then
  begin
    Span := FItems.Last;
    FGlobalIndent := Span.Indent;
    Position := Span.Length + 1;
    UpdateState;
  end;
end;

function TMWKEdit.GetSpanAndPosition(var Position: Integer): Integer;
var
  C: Integer;
  Span: TSpan;
  CursorPos: Integer;
begin

  CursorPos := Position;
  Position := 0;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if CursorPos - Span.Length <= 0 then
    begin
      Result := C;
      Position := CursorPos;
      Exit;
    end;
    Dec(CursorPos, Span.Length);
  end;
  Result := FItems.Count - 1;

  if FItems.Count > 0 then
  begin
    Span := FItems.Last;
    Position := Span.Length + 1;
  end;
end;

function TMWKEdit.GetSpanValues: TSpanValues;
begin
  Result.IsBold := FBold;
  Result.IsItalic := FItalic;
  Result.IsUnderline := FUnderline;
  Result.IsStrikeOut := FStrikeOut;
  Result.IsSpaced := FSpaced;
  Result.IsSuper := FSuper;
  Result.IsSub := FSub;
  Result.IsIndex := FIndex;
  Result.Color := FFontColor;
  Result.BackColor := FBackColor;
  Result.FontName := FFontName;
  Result.FontHeight := FFontHeight;
  Result.Indent := FGlobalIndent;
end;

function TMWKEdit.GetSpellChecker: TSpellChecker;
begin
  Result := FSpellCheckerName;
end;

function TMWKEdit.ParseXML(const XML: IXMLDOMDocument; Insert: Integer;
  var BreakCount: Integer; WithCurrentStyle: Boolean): Integer;
var
  Paragraph: IXMLDOMNode;
  ParaIndent: Integer;
  ParaFont: WideString;
  Node: IXMLDOMNode;
  C, D: Integer;
  Span: TSpan;

  Attributes: TStringList;

  procedure SetAttributes(const Span: TSpan; const Node: IXMLDOMNode);
  var
    C: Integer;
    S: AnsiString;
    D: Double;
  begin
    Span.Updating := True;

    if FPlain then
    begin
      Span.FontName := FFontName;
      Span.ShadowFontName := '';
      Span.FontHeight := FFontHeight;
      Span.Bold := FBold;
      Span.Italic := FItalic;
      Span.Underline := FUnderline;
      Span.StrikeOut := FStrikeOut;
      Span.Spaced := FSpaced;
      Span.Sub := FSub;
      Span.Super := FSuper;
      Span.Index := FIndex;
      Span.Indent := 0;
      Span.Color := FFontColor;
      Span.BackgroundColor := FBackColor;
    end
    else
    begin
      if (HasAttribute(Node, 'font')) and (not FFixedFont) then
        Span.FontName := GetAttribute(Node, 'font')
      else
        Span.FontName := ParaFont;

      if (HasAttribute(Node, 'shadowfont')) and (not FFixedFont) then
        Span.ShadowFontName := GetAttribute(Node, 'shadowfont')
      else
        Span.ShadowFontName := '';

      if (HasAttribute(Node, 'size')) and (not FFixedSize) then
      begin
        S := GetAttribute(Node, 'size');
        if TryStrToFloatD(S, D) then
          Span.FontHeight := D;
      end
      else
        Span.FontHeight := FFontHeight;

      if (HasAttribute(Node, 'color')) and (not FFixedColor) then
        Span.Color := ConvertColor(GetAttribute(Node, 'color'))
      else
        Span.Color := FFontColor;

      if (HasAttribute(Node, 'bg-color')) and (not FFixedColor) then
        Span.BackgroundColor := ConvertColor(GetAttribute(Node, 'bg-color'))
      else
        Span.BackgroundColor := FBackColor;

      Span.Indent := ParaIndent;

      Span.Index := HasAttribute(Node, 'index');

      if HasAttribute(Node, 'deco') then
      begin
        Attributes.CommaText := GetAttribute(Node, 'deco');
        for C := 0 to Attributes.Count - 1 do
        begin
          if Attributes[C] = 'bold' then
            Span.Bold := True;

          if Attributes[C] = 'italic' then
            Span.Italic := True;

          if Attributes[C] = 'underline' then
            Span.Underline := True;

          if Attributes[C] = 'strikeout' then
            Span.StrikeOut := True;

          if Attributes[C] = 'super' then
            Span.Super := True;

          if Attributes[C] = 'sub' then
            Span.Sub := True;

          if Attributes[C] = 'spaced' then
            Span.Spaced := True;
        end;
      end;
    end;
    Span.ShowTags := FShowTags;
    Span.ShowSpellErrors := FShowSpellErrors;
    Span.RightToLeft := FRightToLeft;
    Span.Updating := False;
  end;

  procedure Add;
  begin
    if Span.NewLine and not FMultiLine then
      Exit;
    FItems.Insert(Insert, Span);
    Inc(Insert);
    Inc(Result, Span.Length);
    if Span.LineBreak then
      Inc(BreakCount);
  end;

  procedure AddGraphicSpan(const Node: IXMLDOMNode);
  var
    Picture: TObject;
  begin
    Span.Free;
    if FPlain then
      Exit;
    Span := TGraphicSpan.Create;
    SetAttributes(Span, Node);
    if HasAttribute(Node, 'svg') then
      Picture := TextToImage('1', Node.childNodes[0].xml)
    else
      Picture := TextToImage('2', Node.text);

    TGraphicSpan(Span).SetPicture(Picture);
    FreeAndNil(Picture);
  end;

  procedure AddSectionSpan(const Node: IXMLDOMNode);
  begin
    Span.Free;
    if FPlain then
      Exit;
    Span := TSectionSpan.Create;
    SetAttributes(Span, Node);
  end;

{  procedure JoinEqualSpans;
  var
    C: Integer;
    Span, PrevSpan: TSpan;
  begin
    for C := FItems.Count - 1 downto 1 do
    begin
      Span := FItems[C];
      PrevSpan := FItems[C - 1];
      if PrevSpan.CanAdd(Span) then
      begin
        PrevSpan.Join(Span);
        FItems.Delete(C);
      end;
    end;
  end;}

begin
  BreakCount := 0;
  if Insert = -1 then
    Insert := 0;
  if Insert > FItems.Count then
    Insert := FItems.Count;
  Result := 0;

  Attributes := TStringList.Create;
  for C := 0 to XML.documentElement.childNodes.length - 1 do
  begin
    if Assigned(Paragraph) then
    begin
      if not FMultiLine then
      begin
        Attributes.Free;
        Exit;
      end;
      Span := TSpan.Create;
      Span.NewParagraph := True;
      Span.Text := #13;
      SetAttributes(Span, Paragraph);
      Add;
      Paragraph := nil;
    end;

    if XML.documentElement.childNodes[C].nodeName = 'p' then
    begin
      Paragraph := XML.documentElement.childNodes[C];
      if HasAttribute(Paragraph, 'idt') then
        ParaIndent := GetAttribute(Paragraph, 'idt')
      else
        ParaIndent := 0;

      if HasAttribute(Paragraph, 'font') then
        ParaFont := GetAttribute(Paragraph, 'font')
      else
        ParaFont := FFontName;
        
      for D := 0 to Paragraph.childNodes.length - 1 do
      begin
        Node := Paragraph.childNodes[D];
        Span := TSpan.Create;
        SetAttributes(Span, Node);

        if Node.nodeName = 'br' then
        begin
          Span.NewLine := True;
          Span.Text := #14;
        end;

        if Node.nodeName = 'span' then
        begin
          if HasAttribute(Node, 'sb') then
            Span.Text := DupeString(' ', GetAttribute(Node, 'sb'));

          Span.Text := Span.Text + Node.text;

          if HasAttribute(Node, 'sa') then
            Span.Text := Span.Text + DupeString(' ', GetAttribute(Node, 'sa'));
        end;

        if Node.nodeName = 'tab' then
          Span.Text := cTab;

        if Node.nodeName = 'g' then
          AddGraphicSpan(Node);

        if Node.nodeName = 'section' then
          AddSectionSpan(Node);

        if Span.NewLine and not FMultiLine then
        begin
          Attributes.Free;
          Exit;
        end;
        Add;
      end;
    end;
  end;

  if Assigned(Paragraph) and (not Paragraph.hasChildNodes) then
  begin
    if (Span = nil) or (not Span.NewParagraph) then
    begin
      Span := TSpan.Create;
      Span.NewParagraph := True;
      Span.Text := #13;
      SetAttributes(Span, Paragraph);
      Add;
    end;
  end;

  Attributes.Free;

  //JoinEqualSpans;
end;

procedure TMWKEdit.RealizeFontHeight;
var
  C: Integer;
begin
  BeginUpdate;
  if not FMultiLine then
    ClientHeight := GetHeightOfCurrentFont + FClientRect.Top * 2;
  try
    if FFixedSize then
    begin
      for C := 0 to FItems.Count - 1 do
        FItems[C].FontHeight := FFontHeight;
      FUndoRedo.CancelUndo;
      ReBuildLines(0, MaxInt);
    end else
    begin
      SetSpanValues(ctFontHeight);
      ReBuildLines(FStartLine, FEndLine);
    end;
    CalcSpellCheck(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    SetCaretPos(True);
  finally
    EndUpdate(False);
  end;
  Invalidate;
end;

procedure TMWKEdit.ReBuildIndent;
var
  C: Integer;
  Indent: Integer;
  Span: TSpan;
begin
  Indent := -1;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if Indent = -1 then
      Indent := Span.Indent
    else
      Span.Indent := Indent;

    if Span.NewParagraph then
      Indent := -1;
  end;
end;

procedure TMWKEdit.ReBuildLines(StartLine, EndLine: Integer);
var
  C, Line: Integer;
  W, H: Integer;
  Span, NewSpan: TSpan;

  AvailableSpace: Integer;
  CurrentIndent: Integer;

  procedure CalcTabWidth;
  var
    C: Integer;
  begin
    C := (W div FTabWidth + 1) * FTabWidth;
    if C < ClientWidth then
      Span.Width := C - W
    else
      Span.Width := FTabWidth;
  end;

begin
  if not FLineBreak then
  begin
    StartLine := 0;
    EndLine := MaxInt;
  end;
  
  ReBuildIndent;
  if EndLine < StartLine then
    EndLine := MaxInt;

  C := GetLineStart(StartLine);
  Line := StartLine;

  if C = FItems.Count then
  begin
    CalcLineCountAndStart;
    if FItems.Count = 0 then
    begin
      FVerticalScrollBar.Visible := False;
      FShowVerticalScrollBar := False;
      CalcAreas;
    end;
    CalcSelection;
    Exit;
  end;

  FLineStarts.Clear;
  if FItems.Count > 0 then
    FLineStarts.Add(Pointer(0));
  FLineCount := 1;

  FMaxLineSize := 0;
  CurrentIndent := 0;
  W := 0;
  while C < FItems.Count do
  begin
    H := GetHeightToLine(Line);
    if (not FShowVerticalScrollBar) and (H > FClientRect.Bottom - FCLientRect.Top) then
    begin
      FShowVerticalScrollBar := True;
      FVerticalScrollBar.Visible := FShowScrollBars;
      CalcAreas;
      Exit;
    end;

    Span  := FItems[C];

    if FLineBreak then
      AvailableSpace := ClientWidth - Span.Indent * FIndentSize
    else
      AvailableSpace := MaxInt;

    CurrentIndent := Max(CurrentIndent, Span.Indent * FIndentSize);
    
    if Span.IsTab then
      CalcTabWidth;

    while (W + Span.Width < AvailableSpace) and
       (C < FItems.Count - 1) and Span.CanAdd(FItems[C + 1]) do
    begin
      Span.Join(FItems[C + 1]);
      FItems.Delete(C + 1);
    end;
    if not (Span.NewLine or Span.NewParagraph) then
      Span.LineBreak := False;
    NewSpan := Span.Part(AvailableSpace - W);

    if Assigned(NewSpan) then
    begin
      if NewSpan = Pointer(1) then
      begin
        if C > 0 then
        begin
          FItems[C - 1].LineBreak := True;
          Inc(FLineCount);
          FLineStarts.Add(Pointer(C));
        end;
        Inc(Line);
        if EndLine < MaxInt then
          Inc(EndLine);
        if W = 0 then
        begin
          NewSpan := Span.Split(2);
          if Assigned(NewSpan) then
            FItems.Insert(C + 1, NewSpan);
          Span.LineBreak := True;

          Inc(FLineCount);
          FLineStarts.Add(Pointer(C + 1));
          Inc(C);
        end;
        W := 0;
        Continue;
      end;
      FItems.Insert(C + 1, NewSpan);
      W := 0;
      Inc(Line);
      if EndLine < MaxInt then
        Inc(EndLine);
      Inc(C);
      Continue;
    end;

    Inc(W, Span.Width);
    if (Span.LineBreak) or
       (C = FItems.Count - 1) then
    begin
      Inc(FLineCount);
      FLineStarts.Add(Pointer(C + 1));
      FMaxLineSize := Max(FMaxLineSize, W + CurrentIndent);
      CurrentIndent := 0;
      W := 0;
      Inc(Line);
    end;
    if (Span.NewParagraph or Span.NewLine) and
       (Line > EndLine) then
      Break;
    Inc(C);
  end;

  CalcSelection;
  CalcLineCountAndStart;

  ToggleScrollBar;
end;

procedure TMWKEdit.Redo;
begin
  FUndoredo.Redo;
end;

class procedure TMWKEdit.RegisterSpellChecker(const Name: TSpellChecker;
  SpellCheckerClass: TSpellCheckerClass);
begin
  GetSpellCheckers.Add(Name, SpellCheckerClass);
end;

function TMWKEdit.GetWordStart(Position: Integer): Integer;
var
  C: Integer;
  Span: TSpan;
  CurrentSpan: Integer;
  PosInSpan: Integer;
  Current: Integer;
  WC: WideChar;
begin
  Result := 0;
  Current := GetSpanAndPosition(Position);
  if (Current > FItems.Count - 1) or (FItems.Count = 0) then
    Exit;

  Span := FItems[Current];

  if Position > Span.Length then
    Dec(Position);

  CurrentSpan := Current;
  PosInSpan := Position;
  if (PosInSpan = 1) and (CurrentSpan > 0) then
  begin
    Dec(CurrentSpan);
    Span := FItems[CurrentSpan];
    PosInSpan := Span.Length;
    if Span.LastSpace(Span.Text, Maxint, WC) = Span.Length then
    begin
      for C := 0 to CurrentSpan do
        Inc(Result, FItems[C].Length);
      Exit;
    end;
  end;

  while (Span.Text[PosInSpan] = ' ') or
        (Span.Text[PosInSpan] = cProtectSpace) do
  begin
    Dec(PosInSpan);
    if (PosInSpan <= 1) then
    begin
      if (CurrentSpan > 0) then
      begin
        Dec(CurrentSpan);
        Span := FItems[CurrentSpan];
        PosInSpan := Span.Length;
      end else
      begin
        Result := 0;
        Exit;
      end;
    end;
  end;

  PosInSpan := Span.LastSpace(Span.Text, PosInSpan, WC);
  while (PosInSpan = 0) and (CurrentSpan > 0) do
  begin
    Dec(CurrentSpan);
    Span := FItems[CurrentSpan];
    if Span.NewLine or Span.NewParagraph then
      PosInSpan := Span.Length
    else
      PosInSpan := Span.LastSpace(Span.Text, MaxInt, WC);
  end;

  for C := 0 to CurrentSpan - 1 do
    Inc(Result, FItems[C].Length);
  Inc(Result, PosInSpan);
  if not ((PosInSpan = Position) and (CurrentSpan = Current)) then
    Inc(Result);
end;

function TMWKEdit.GetXML: WideString;
begin
  Result := CollectAsXML(False, False);
end;

procedure TMWKEdit.HideDragCaret;
begin
  PaintDragCaret;
  FShowDragCaret := False;
end;

function TMWKEdit.GetWordLength(Position: Integer): Integer;
var
  C: Integer;
  Span: TSpan;
  CurrentSpan: Integer;
  PosInSpan: Integer;
  Current: Integer;
  WC: WideChar;
begin
  Result := 0;
  Current := GetSpanAndPosition(Position);
  if (Current > FItems.Count - 1) or (FItems.Count = 0) then
    Exit;

  Span := FItems[Current];

  if Position > Span.Length then
  begin
    Result := 1;
    Exit;
  end;

  CurrentSpan := Current;
  PosInSpan := Span.NextSpace(Span.Text, Position, WC);

  while (PosInSpan > Span.Length) and (CurrentSpan < FItems.Count - 1) and
    (WC <> #13) and (WC <> #14) do
  begin
    Inc(CurrentSpan);
    Span := FItems[CurrentSpan];
    PosInSpan := Span.NextSpace(Span.Text, 0, WC);
  end;

  while (PosInSpan - 1 > 0) and (Span.Text[PosInSpan - 1] = '''') do
  begin
    Dec(PosInSpan);
    WC := '''';
  end;

  Result := FItems[Current].Length - Position + 1;
  for C := Current + 1 to CurrentSpan - 1 do
    Inc(Result, FItems[C].Length);

  if CurrentSpan <> Current then
  begin
    Inc(Result, PosInSpan);
    if not ((WC = ' ') or (WC = cProtectSpace)) then
      Dec(Result);
  end
  else
  begin
    Result := PosInSpan - Position;
    if (Result = 0) or (WC = ' ') or (WC = cProtectSpace) then
      Inc(Result);
  end;
end;

function TMWKEdit.GetWord(WithSpace: Boolean; var Start, Length: Integer): WideString;
var
  Span: TSpan;
  Current, Position: Integer;
begin
  if Start = -1 then
  begin
    Start := GetWordStart(FSelStart);
    Length := GetWordLength(Start);
  end;

  Position := Start;
  Current := GetSpanAndPosition(Position);
  Span := FItems[Current];
  Result := Copy(Span.Text, Position, Length);
  Dec(Length, System.Length(Result));

  while Length > 0 do
  begin
    Inc(Current);
    Span := FItems[Current];
    Result := Result + Copy(Span.Text, 1, Length);
    Dec(Length, Min(Span.Length, Length));
  end;

  Length := System.Length(Result);
  if (Length > 0) and (not WithSpace) and
     ((Result[Length] = ' ') or (Result[Length] = cProtectSpace)) then
  begin
    Result := Copy(Result, 1, Length - 1);
    Dec(Length);
  end;
end;

function TMWKEdit.GetCurrentWord(const Position: Integer): WideString;
var
  Start, Length: Integer;
  C: Integer;
  Span: TSpan;
begin
  Start := GetWordStart(Position);
  Length := GetWordLength(Start);
  Result := '';

  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    if Start > Span.Length then
      Dec(Start, Span.Length)
    else
    begin
      if Start + Length - 1 <= Span.Length then
      begin
        Result := Result + Copy(Span.Text, Start, Length);
        Exit;
      end else
      begin
        Result := Result + Copy(Span.Text, Start, MaxInt);
        Start := 1;
        Dec(Length, Span.Length);
      end;
    end;
  end;
end;

function TMWKEdit.SpellCheck(AStart, ALength: Integer; Highlight: Boolean): Integer;
var
  Selected: WideString;

  C, Res: Integer;
  Replacer: WideString;

  procedure StartUndo(Start, Len: Integer);
  var
    FSelStart, FSelLength: Integer;
    L: Integer;
    IR: TInsertRec;
  begin
    if FRollBack > 0 then
      Exit;

    FinishInsert;
    FSelStart := SelStart;
    FSelLength := SelLength;
    try
      Inc(FUpdateCount);
      SelStart := Start;
      SelLength := Len;

      L := System.Length(Replacer);

      IR.Edit := Self;
      FUndoRedo.AddActionToUndo(USPELLCHECK, atUndo, Integer(@IR), L);
      FOldInsertAction := IR.OldInsertAction;

      SelStart := FSelStart;
      SelLength := FSelLength;
    finally
      Dec(FUpdateCount);
    end;
  end;

  procedure EndUndo(Start, OldLen, NewLen: Integer);
  var
    FSelStart, FSelLength: Integer;
    IR: TInsertRec;
  begin
    if FRollBack > 0 then
      Exit;

    FSelStart := SelStart;
    FSelLength := SelLength;
    Inc(FUpdateCount);
    SelStart := Start;
    SelLength := NewLen;

    IR.Edit := Self;
    IR.OldInsertAction := TInsertAction(FOldInsertAction);
    IR.Text := nil;

    FUndoRedo.AddActionToUndo(USPELLCHECK, atRedo, Integer(@IR), OldLen);
    SelStart := FSelStart;
    SelLength := FSelLength;
    Dec(FUpdateCount);
  end;

  procedure DoReplace;
  var
    C: Integer;
    Span: TSpan;
    L, SL: Integer;
    Position: Integer;
    SelLength: Integer;
    OldLength: Integer;
    NewLength: Integer;
    UpdateStart: Integer;
  begin
    Position := AStart;
    C := GetSpanAndPosition(Position);
    UpdateStart := C;
    L := System.Length(Replacer);
    NewLength := L;
    OldLength := ALength;
    StartUndo(AStart, ALength);
    SL := L;
    repeat
      Span := FItems[C];
      SelLength := Min(Span.Length - Position + 1, ALength);
      Dec(ALength, SelLength);

      if (SelLength >= L) or
         (Position + SelLength - 1 < Span.Length) or
         (C = FItems.Count - 1) then
      begin
        Span.Text := Copy(Span.Text, 1, Position - 1) + Replacer +
          Copy(Span.Text, Position + SelLength, MaxInt);
        Replacer := '';
      end
      else
      if SelLength = Span.Length then
      begin
        if (C < FItems.Count - 1) and (ALength <= 0) then
        begin
          Span.Text := Replacer;
          Replacer := '';
        end
        else
        begin
          Span.Text := Copy(Replacer, 1, SelLength);
          Replacer := Copy(Replacer, SelLength + 1, MaxInt);
          Inc(C);
        end;
      end
      else
      begin
        Span.Text := Copy(Span.Text, 1, Position - 1) + Copy(Replacer, 1, SelLength);
          Copy(Span.Text, Position + SelLength, MaxInt);
        Replacer := Copy(Replacer, SelLength + 1, MaxInt);
        Inc(C);
      end;
      L := System.Length(Replacer);
      Position := 1;
    until L = 0;

    EndUndo(AStart, OldLength, NewLength);
    ReBuildLines(UpdateStart, UpdateStart);
    if Highlight then
      DefineSelection(AStart, SL)
    else
      Invalidate;
  end;

  function IsNumber(const S: WideString): Boolean;
  var
    C: Integer;
  begin
    Result := True;
    for C := 1 to System.Length(S) do
      if (S[C] >= '0') and (S[C] <= '9') then
        Exit;
    Result := False;
  end;

begin
  Result := -1;
  if not Assigned(FSpellChecker) or (FSpellChecker.CurrentDictionary = -1) or
     not FSpellChecker.Loaded then
    Exit;

  Selected := GetWord(False, AStart, ALength);

  if IsNumber(Selected) then
  begin
    Result := 0;
    Exit;
  end;

  if not FSpellChecker.SpelledCorrect(Selected) then
  begin
    if Highlight then
      DefineSelection(AStart, ALength);

    FSpellChecker.FillSuggestions(Selected);
    FSpellCheckForm.Init(Selected);
    for C := 0 to FSpellChecker.Suggestions.Count - 1 do
      FSpellCheckForm.Add(FSpellChecker.Suggestions[C]);

    Res := FSpellCheckForm.ShowModal;

    if Res = mrOk then
    begin
      // Ersetzen
      Replacer := FSpellCheckForm.Replacer;
      Result := System.Length(Replacer);
      FSpellChecker.Replace(Selected, Replacer);
      DoReplace;
    end;

    if Res = mrCancel then
    begin
    end;

    if Res = mrYes then
    begin
      FSpellChecker.Add(FSpellCheckForm.Edit.Text);
      Result := System.Length(FSpellCheckForm.Edit.Text);
    end;

    if Res = mrNo then
    begin
      Result := System.Length(Selected);
    end;

    if Res = mrNoToAll then
    begin
      FSpellChecker.Ignore(Selected);
      Result := System.Length(Selected);
    end;
  end else
    Result := 0;
end;

function TMWKEdit.IsSpelledCorrect(AStart: Integer; var ALength: Integer): Boolean;
var
  Selected: WideString;

  function IsNumber(const S: WideString): Boolean;
  var
    C: Integer;
  begin
    Result := True;
    for C := 1 to System.Length(S) do
      if (S[C] >= '0') and (S[C] <= '9') then
        Exit;
    Result := False;
  end;

begin
  Result := True;
  if not Assigned(FSpellChecker) or (FSpellChecker.CurrentDictionary = -1) or
     not FSpellChecker.Loaded then
    Exit;

  Selected := GetWord(False, AStart, ALength);

  if IsNumber(Selected) then
    Exit;

  Result := FSpellChecker.SpelledCorrect(Selected);
end;

procedure TMWKEdit.SpellCheckSelected;
begin
  FinishInsert;
  if (FSelLength = 0) then
  begin
    FNegativeSelection := False;
    SelStart := GetWordStart(FSelStart);
    SelLength := GetWordLength(FSelStart);
  end;
  SpellCheck(-1, -1, True);
  CalcSpellCheck(True);
end;

function TMWKEdit.SupportsOLEData(const DataObj: IDataObject;
  const GrfKeyState: Integer; const PT: TPoint; var dwEffect: Integer): HRESULT;
begin
  Result := S_FALSE;
  if (not FReadOnly) and OLEHasSupportedFormat(DataObj, GetMyOLEFormats,
     FOLEFormat) then
    Result := S_OK;
  if FOLEFormat = CF_HDROP then
    if dwEffect = DROPEFFECT_MOVE then
      dwEffect := DROPEFFECT_LINK;
end;

procedure TMWKEdit.SpellCheckAll;
var
  C, Max, SC: Integer;
  OldPosition, Position, Length: Integer;
begin
  FinishInsert;
  if not Assigned(FSpellChecker) or (FSpellChecker.CurrentDictionary = -1) or
     not FSpellChecker.Loaded then
    Exit;

  Position := 1;
  Max := 0;
  for C := 0 to FItems.Count - 1 do
    Inc(Max, FItems[C].Length);

  while Position < Max do
  begin
    OldPosition := Position;
    Position := GetWordStart(Position);
    Length := GetWordLength(Position); 

    if Length > 2 then
    begin
      SC := SpellCheck(Position, Length, True);
      if SC = -1 then
      begin
        CalcSpellCheck(True);
        SelStart := Position + Length - 1;
        Exit;
      end;

      if SC = 0 then
        SC := Length
      else
        Dec(Length);

      Dec(Max, Length - SC);
    end else
      SC := 2;

    Inc(Position, SC + 1);
    if Position <= OldPosition then
      Position := OldPosition + 1;
  end;
  CalcSpellCheck(True);
  SelStart := Max;
  ShowMessage('Die berprfung wurde abgeschlossen!');
end;

procedure TMWKEdit.RemoveSelection;
var
  C: Integer;
begin
  for C := 0 to FItems.Count - 1 do
  begin
    FItems[C].SelStart := 0;
    FItems[C].SelLength := 0;
  end;
end;

procedure TMWKEdit.ClearSpanValues;
begin
  FBold := False;
  FItalic := False;
  FUnderline := False;
  FStrikeOut := False;
  FSpaced := False;
  FSuper := False;
  FSub := False;
  FIndex := False;
end;

procedure TMWKEdit.ClearSpells;
var
  C: Integer;
begin
  for C := 0 to FItems.Count - 1 do
    FItems[C].ClearSpells;
end;

procedure TMWKEdit.ClearUndo;
begin
  FinishInsert;
  FUndoRedo.Clear;
end;

procedure TMWKEdit.SetSpell(const AStart, ALength: Integer);
var
  C, Num, L: Integer;
  Start, Length: Integer;
  Span: TSpan;
  SpellStart, SpellLength: Integer;
begin
  Start := AStart;
  Length := ALength;
  SpellStart := 0;
  Num := 0;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    L := Span.Length;
    if (Start <> -1) and (Num + L >= Start) then
    begin
      SpellStart := Start - Num;
      Start := -1;
    end;

    if (Start = -1) and (Length > 0) then
    begin
      if SpellStart <= 0 then
        SpellStart := 1;
      SpellLength := Length;
      if SpellLength > (L - SpellStart + 1) then
        SpellLength := (L - SpellStart + 1);
      Span.AddSpell(SpellStart, SpellLength);
      Dec(Length, SpellLength);
      Start := Num + L + 1;
    end;

    Inc(Num, L);
    if Length = 0 then
      Break;
  end;
end;

procedure TMWKEdit.SetSpellChecker(const Value: TSpellChecker);
var
  SpellCheckerEntry: PSpellCheckerEntry;
begin
  if SameText(Value, FSpellCheckerName) then
    Exit;

  if Assigned(FSpellChecker) then
    FreeAndNil(FSpellChecker);

  if Assigned(SpellCheckers) and (SpellCheckers.IndexOf(Value) <> -1) then
  begin
    SpellCheckerEntry := SpellCheckers[SpellCheckers.IndexOf(Value)];
    FSpellChecker := SpellCheckerEntry^.SpellCheckerClass.Create;
    FSpellChecker.DictionaryDir := ExtractFilePath(ParamStr(0)) + 'Dict\';
    FSpellCheckerName := Value;
  end else
    FSpellCheckerName := '';
end;

procedure TMWKEdit.CalcSpellCheck(const WithPaint: Boolean);
var
  C, Max: Integer;
  OldPosition, Position, Length: Integer;
begin
  ClearSpells;
  if not Assigned(FSpellChecker) or (FSpellChecker.CurrentDictionary = -1) or
     not FSpellChecker.Loaded then
  begin
    if WithPaint then
      Invalidate;
    Exit;
  end;

  Position := 1;
  Max := 0;
  for C := 0 to FItems.Count - 1 do
    Inc(Max, FItems[C].Length);

  while Position < Max do
  begin
    OldPosition := Position;
    Position := GetWordStart(Position);
    Length := GetWordLength(Position);

    if Length > 2 then
    begin
      if not IsSpelledCorrect(Position, Length) then
        SetSpell(Position, Length);
    end;

    Inc(Position, Length + 1);
    if Position <= OldPosition then
      Position := OldPosition + 1;
  end;
  if WithPaint then
    Invalidate;
end;

function TMWKEdit.CalculateParagraphRange: TSpanRange;
var
  Start, Stop: Integer;
begin
  if FItems.Count = 0 then
  begin
    Result.Start := -1;
    Result.Stop := -1;
    Exit;
  end;

  if FSelLength = 0 then
  begin
    Start := FCurrentSpan - 1;
    if Start > 0 then
    begin
      while Start > 0 do
      begin
        if FItems[Start].NewParagraph then
        begin
          Inc(Start);
          Break;
        end;
        Dec(Start);
      end;
    end;
    if Start < 0 then
      Start := 0;

    if (FCurrentSpan = FItems.Count - 1) and
       ((FPosInSpan > FItems.Last.Length)) then
    begin
      Result.Start := Start;
      Result.Stop := MaxInt;
      Exit;
    end;

    Stop := Start;
    while (Stop < FItems.Count - 1) and not FItems[Stop].NewParagraph do
      Inc(Stop);
  end else

  begin
    Start := 0;
    while Start < FItems.Count do
    begin
      if FItems[Start].SelLength > 0 then
        Break;
      Inc(Start);
    end;

    repeat
      Dec(Start);
    until (Start = -1) or FItems[Start].NewParagraph;
    Inc(Start);

    Stop := FCurrentSpan;
    while (Stop < FItems.Count - 1) and (FItems[Stop].SelLength > 0) do
      Inc(Stop);
    if (Stop < FItems.Count - 1) or
       ((Stop = FItems.Count - 1) and (FItems[Stop].SelLength = 0)) then
      Dec(Stop);

    while (Stop < FItems.Count - 1) and (not FItems[Stop].NewParagraph) do
      Inc(Stop);
  end;

  Result.Start := Start;
  Result.Stop := Stop;
end;

function TMWKEdit.CalculateSelectedSpanRange: TSpanRange;
var
  Start, Stop: Integer;
  Span: TSpan;
begin
  if (FItems.Count = 0) or (FSelLength = 0) then
  begin
    Result.Start := -1;
    Result.Stop := -1;
    Exit;
  end;

  Start := 0;
  while Start < FItems.Count do
  begin
    if FItems[Start].SelLength > 0 then
      Break;
    Inc(Start);
  end;

  Stop := Start;
  Span := FItems[Stop];
  while (Stop < FItems.Count - 1) and (Span.SelLength > 0) and
    (Span.SelStart + Span.SelLength - 1 = Span.Length) do
  begin
    Inc(Stop);
    Span := FItems[Stop];
    if Span.SelLength = 0 then
    begin
      Dec(Stop);
      Break;
    end;
  end;

  Result.Start := Start;
  Result.Stop := Stop;
end;

procedure TMWKEdit.Changed;
begin
  DoOnChange;
end;

procedure TMWKEdit.CalcSelection;
var
  C, Num, Start, Laenge, L: Integer;
  Span: TSpan;
  PaintStart, PaintStop: Integer;
begin
  if FSelLength = 0 then
  begin
    FStartLine := FindStartLine;
    FEndLine := FStartLine;

    if not ScrollInSight(SetCaretPos(False)) then
      Invalidate;
    CalcCurrentLine;
    Exit;
  end;

  Start := FSelStart;
  Laenge := FSelLength;

  PaintStart := -1;
  PaintStop := -1;
  Num := 0;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    L := Span.Length;
    if (Start <> -1) and (Num + L >= Start) then
    begin
      Span.SelStart := Start - Num;
      Start := -1;
      PaintStart := FindLineOfSpan(C);
    end;

    if (Start = -1) and (Laenge > 0) then
    begin
      if Span.SelStart <= 0 then
        Span.SelStart := 1;
      Span.SelLength := Laenge;
      if Span.SelLength > (L - Span.SelStart + 1) then
        Span.SelLength := (L - Span.SelStart + 1);
      Dec(Laenge, Span.SelLength);
      if Laenge = 0 then
        PaintStop := FindLineOfSpan(C);
    end;

    Inc(Num, L);
    if Laenge = 0 then
      Break;
  end;
  if Laenge > 0 then
  begin
    Dec(FSelLength, Laenge);
    PaintStop := FLineCount - 1;
  end;

  FStartLine := PaintStart;
  FEndLine := PaintStop;

  if not ScrollInSight(SetCaretPos(False)) then
    Invalidate;

  CalcCurrentLine;
end;

procedure TMWKEdit.CalcCaretPos(var Top, Height: Integer);

  function CreateFont(WithSuperSub: Boolean): HFont;
  var
    LogFont: TLogFontW;
    FontName: WideString;
  begin
    FillChar(LogFont, SizeOf(LogFont), 0);
    if (FSub or FSuper) and WithSuperSub then
      LogFont.lfHeight := Round(FFontHeight * SUPERSUBFRAC)
    else
      LogFont.lfHeight := Round(FFontHeight);
    LogFont.lfWidth := 0;
    LogFont.lfEscapement := 0;
    LogFont.lfOrientation := 0;
    if FBold then
      LogFont.lfWeight := FW_BOLD
    else
      LogFont.lfWeight := FW_NORMAL;
    LogFont.lfItalic := Byte(FItalic);
    LogFont.lfUnderline := Byte(FUnderline);
    LogFont.lfStrikeOut := Byte(FStrikeOut);
    logfont.lfCharSet := DEFAULT_CHARSET;
    LogFont.lfOutPrecision := OUT_DEFAULT_PRECIS;
    LogFont.lfClipPrecision := CLIP_DEFAULT_PRECIS;
    LogFont.lfQuality := DEFAULT_QUALITY;
    LogFont.lfPitchAndFamily := DEFAULT_PITCH or FF_DONTCARE;
    FontName := FFontName;

    Move(FontName[1], LogFont.lfFaceName, System.Length(FontName) * SizeOf(WideChar));
    LogFont.lfFaceName[System.Length(FontName)] := #0;
    Result := CreateFontIndirectW(LogFont);
  end;

var
  Font, OldFont: HFont;
  Metric: TTextMetricW;
  DC: HDC;
  LineHeight, Ascent, TextHeight: Integer;
  Size: TSize;
begin
  DC := GetDC(0);
  try
    Font := CreateFont(False);
    OldFont := SelectObject(DC, Font);
    GetTextExtentPoint32W(DC, 'Mp', 2, Size);
    LineHeight := Size.cy;
    TextHeight := Size.cy;
    GetTextMetricsW(DC, Metric);
    Ascent := Metric.tmAscent;
    SelectObject(DC, OldFont);
    DeleteObject(Font);

    if FSuper or FSub then
    begin
      Font := CreateFont(True);
      OldFont := SelectObject(DC, Font);
      GetTextExtentPoint32W(DC, 'Mp', 2, Size);
      TextHeight := Size.cy;
      SelectObject(DC, OldFont);
      DeleteObject(Font);
    end;
  finally
    ReleaseDC(0, DC);
  end;
  Height := TextHeight;
  Top := -Ascent;

  if FSub then
  begin
    Top := - TextHeight div 2;
    if Ascent + Top + TextHeight >= LineHeight then
      Top := LineHeight - TextHeight - Ascent;
  end;
end;

procedure TMWkEdit.CalcCurrentLine;
begin
  if FNegativeSelection then
    FCurrentLine := FindStartLine + 1
  else
    FCurrentLine := FindEndLine + 1;
end;

function TMWkEdit.ScrollInSight(Point: TPoint): Boolean;
begin
  Result := False;

  if Point.X < FClientRect.Left then
  begin
    FHorizontalScrollBar.Position := FHorizontalScrollBar.Position + (Point.X - FClientRect.Left);
    Result := True;
  end
  else
  if Point.X > FClientRect.Right then
  begin
    FHorizontalScrollBar.Position := FHorizontalScrollBar.Position + Point.X - FClientRect.Right;
    Result := True;
  end;

  if not FMultiLine then
    Exit;
  if Point.Y < FClientRect.Top then
  begin
    FVerticalScrollBar.Position := FVerticalScrollBar.Position + (Point.Y - FClientRect.Top);
    Result := True;
  end
  else
  if Point.Y + FCaretHeight > FClientRect.Bottom then
  begin
    FVerticalScrollBar.Position := FVerticalScrollBar.Position + Point.Y + FCaretHeight - FClientRect.Bottom;
    Result := True;
  end;
end;

procedure TMWKEdit.PaintLine(Index, X, Y: Integer);
var
  C: Integer;
  Indent: Integer;
  Span: TSpan;
  R: TRect;
  LineHeight, Baseline: Integer;
begin
  if Index < 0 then
    Exit;

  C := GetLineStart(Index);

  if C = FItems.Count then
    Exit;

  Canvas.Brush.Color := FBackColor;

  if C > FItems.Count then
    Exit;

  Indent := FItems[C].Indent * FIndentSize;

  LineHeight := GetHeightOfLine(Index);
  Baseline := GetBaselineOfLine(Index);

  if Indent > 0 then
  begin
    if FRightToLeft then
    begin
      Dec(X, Indent);
      R := Rect(X, Y, FClientRect.Right, Y + LineHeight);
    end
    else
    begin
      Inc(X, Indent);
      R := Rect(FClientRect.Left, Y, X, Y + LineHeight);
    end;
    FillRect(Canvas.Handle, R, FBrush.Handle);
  end;

  Span := nil; // To avoid compiler warning
  repeat
    if C = FItems.Count then
      Break;
    Span := FItems[C];
    if FRightToLeft then
    begin
      Span.Paint(Canvas, X - Span.Width, Y, LineHeight, Baseline, Focused);
      Dec(X, Span.Width);
    end else
    begin
      Span.Paint(Canvas, X, Y, LineHeight, Baseline, Focused);
      Inc(X, Span.Width);
    end;
    Inc(C);
  until Span.LineBreak;

  if FShowTags then
  begin
    if Span.NewLine then
      C := Span.CalcTextWidth(cNewLine)
    else
      if Span.NewParagraph then
        C := Span.CalcTextWidth(cNewParagraph)
      else
        C := 0;
    if FRightToLeft then
      Dec(X, C)
    else
      Inc(X, C);
  end;

  if FRightToLeft then
    R := Rect(FClientRect.Left, Y, X, Y + LineHeight)
  else
    R := Rect(X, Y, FClientRect.Right, Y + LineHeight);
  FillRect(Canvas.Handle, R, FBrush.Handle);
end;

procedure TMWKEdit.PasteFromClipboard;
var
  S: WideString;
  XML: IXMLDOMDocument;
  Span: TSpan;
  Title: WideString;
  RTF: TRTFToXML;
  OldCursor: TCursor;
begin
  FinishInsert;
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    S := ReadFromClipBoardW(CF_MWKTEXT);
    if S <> '' then
    begin
      XML := CreateFromString(S);
      if not (Assigned(XML) and (XML.documentElement.nodeName = 'mwkedit')) then
        XML := nil;
    end;

    if not Assigned(XML) then
    begin
      S := ReadFromClipBoardA(CF_RTF);
      if S <> '' then
      begin
        RTF := TRTFToXML.Create;
        try
          XML := RTF.Parse(S, Title);
        finally
          RTF.Free;
        end;
      end;
    end;

    if not Assigned(XML) then
    begin
      S := ReadFromClipBoardW(CF_UNICODETEXT);
      if S <> '' then
      begin
        if (FCurrentSpan >= 0) and (FCurrentSpan < FItems.Count) then
          Span := FItems[FCurrentSpan]
        else
          Span := nil;

        XML := CreateXMLFromPlain(S, Span);
      end;
    end;

    if Assigned(XML) then
      InsertXML(XML);
  finally
    Screen.Cursor := OldCursor;
  end;
end;

procedure TMWKEdit.SaveToFile(const FileName: WideString; FileFormat: TFileFormat);
var
  S: TFileStream;
begin
  S := TFileStream.Create(Filename, fmCreate);
  try
    SaveToStream(S, FileFormat);
  finally
    S.Free;
  end;
end;

procedure TMWKEdit.SaveToStream(const Stream: TStream; FileFormat: TFileFormat);
const
  UTF8BOM = AnsiString (#$EF#$BB#$BF);

var
  SW: WideString;
  SA: AnsiString;
  BOM: WideChar;
begin
  FinishInsert;
  if FileFormat = ffRTF then
  begin
    SA := EncodeToRTF;
    Stream.WriteBuffer(PChar(SA)^, System.Length(SA));
    FUndoRedo.Stable := True;
    Exit;
  end;

  if FileFormat = ffUnicode then
  begin
    BOM := UNICODE_BOM;
    Stream.WriteBuffer(BOM, SizeOf(WideChar));
  end;

  SW := CollectAsXML(False, FileFormat <> ffXML);

  if FileFormat = ffUnicode then
    Stream.WriteBuffer(PWideChar(SW)^, System.Length(SW) * SizeOf(WideChar))
  else
  begin
    if FileFormat = ffXML then
    begin
      SA := UTF8Encode(SW);
      Stream.WriteBuffer(UTF8BOM[1], 3);
    end else
      SA := SW;
    Stream.WriteBuffer(PChar(SA)^, System.Length(SA));
  end;
  FUndoRedo.Stable := True;
end;

function TMWKEdit.EncodeToRTF: AnsiString;
var
  FontNames: TStringList;
  Colors: TStringList;
  C: Integer;
  Span: TSpan;

  function GetByte(C: Char): Byte;
  begin
    if (C in ['0'..'9']) then
      Result := Ord(C) - Ord('0')
    else
      if (C in ['A'..'F']) then
        Result := Ord(C) - Ord('A') + 10
      else
        if (C in ['a'..'f']) then
          Result := Ord(C) - Ord('a') + 10
        else
          Result := 0;
  end;

  function CreateHeader: AnsiString;
  var
    C: Integer;
    Header: AnsiString;
    S: AnsiString;
    R, G, B: Byte;
  begin
    // RTF Header
    Header := '{\rtf1\ansi\deff0';

    if FontNames.Count > 0 then
    begin
      Header := Header + '{\fonttbl';
      for C := 0 to FontNames.Count - 1 do
        Header := Header + '{\f' + IntToStr(C) + ' ' + FontNames[C] + ';}';
      Header := Header + '}' + sLineBreak;
    end;

    if Colors.Count > 0 then
    begin
      Header := Header + '{\colortbl';
      for C := 0 to Colors.Count - 1 do
      begin
        S := Colors[C];
        B := GetByte(S[1]) shl 4 + GetByte(S[2]);
        G := GetByte(S[3]) shl 4 + GetByte(S[4]);
        R := GetByte(S[5]) shl 4 + GetByte(S[6]);

        Header := Header + '\red' + IntToStr(R) +
          '\green' + IntToStr(G) +
          '\blue' + IntToStr(B) + ';';
      end;
      Header := Header + '}' + sLineBreak;
    end;

    Header := Header + '{\*\generator ' + Application.Title + ';}';

    Header := Header + '\uc1\fs' + FloatToStr(FontSize * 2);

    Result := Header;
  end;

  function GetSpanContents: AnsiString;
  var
    C, D: Integer;
    Span: TSpan;
    Content: AnsiString;
    FontName: AnsiString;
    Color, BackColor: TColor;
    WasTag: Boolean;
  begin
    Color := -1;
    BackColor := -1;
    WasTag := True;
    for C := 0 to FItems.Count - 1 do
    begin
      Span := FItems[C];
      if Span.FontName <> FontName then
      begin
        Content := Content + '\f' + IntToStr(FontNames.IndexOf(Span.FontName));
        WasTag := True;
      end;

      if Span.Color <> Color then
      begin
        Content := Content + '\cf' +
          IntToStr(Colors.IndexOf(IntToHex(ColorToRGB(Span.Color) and $FFFFFF, 6)));
        WasTag := True;
      end;

      if Span.BackgroundColor <> BackColor then
      begin
        Content := Content + '\cb' +
          IntToStr(Colors.IndexOf(IntToHex(ColorToRGB(Span.BackgroundColor) and $FFFFFF, 6)));
        WasTag := True;
      end;

      if Span.Bold then
      begin
        Content := Content + '\b';
        WasTag := True;
      end;
      if Span.Italic then
      begin
        Content := Content + '\i';
        WasTag := True;
      end;
      if Span.Underline then
      begin
        Content := Content + '\ul';
        WasTag := True;
      end;
      if Span.StrikeOut then
      begin
        Content := Content + '\strike';
        WasTag := True;
      end;
      if Span.Spaced then
      begin
        Content := Content + '\expnd8';
        WasTag := True;
      end;
      if Span.Super then
      begin
        Content := Content + '\super';
        WasTag := True;
      end;
      if Span.Sub then
      begin
        Content := Content + '\sub';
        WasTag := True;
      end;

      for D := 1 to Span.Length do
      begin
        if Span.Text[D] = '\' then
        begin
          Content := Content + '\\';
          WasTag := False;
          Continue;
        end;

        if Span.Text[D] = '{' then
        begin
          Content := Content + '\{';
          WasTag := False;
          Continue;
        end;

        if Span.Text[D] = '}' then
        begin
          Content := Content + '\}';
          WasTag := False;
          Continue;
        end;

        if Span.Text[D] = cProtectSpace then
        begin
          Content := Content + '\~';
          WasTag := False;
          Continue;
        end;

        if Span.Text[D] = cHyphen then
        begin
          Content := Content + '\-';
          WasTag := False;
          Continue;
        end;

        if Ord(Span.Text[D]) < 128 then
        begin
          if (Ord(Span.Text[D]) in [$D, $E]) then
            Continue;
          if WasTag then
            Content := Content + ' ' + Span.Text[D]
          else
            Content := Content + Span.Text[D];
          WasTag := False;
          Continue;
        end;

        if Ord(Span.Text[D]) < 256 then
        begin
          Content := Content + '\''' + IntToHex(Ord(Span.Text[D]), 2);
          WasTag := False;
          Continue;
        end;

        Content := Content + '\u' + IntToStr(Ord(Span.Text[D])) + '?';
        WasTag := False;
      end;

      if Span.Bold then
      begin
        Content := Content + '\b0';
        WasTag := True;
      end;
      if Span.Italic then
      begin
        Content := Content + '\i0';
        WasTag := True;
      end;
      if Span.Underline then
      begin
        Content := Content + '\ul0';
        WasTag := True;
      end;
      if Span.StrikeOut then
      begin
        Content := Content + '\strike0';
        WasTag := True;
      end;
      if Span.Spaced then
      begin
        Content := Content + '\expnd0';
        WasTag := True;
      end;
      if Span.Super or Span.Sub then
      begin
        Content := Content + '\nosupersub';
        WasTag := True;
      end;

      Color := Span.Color;
      BackColor := Span.BackgroundColor;
      FontName := Span.FontName;

      if Span.NewLine then
      begin
        Content := Content + '\line' + sLineBreak;
        WasTag := False;
      end;

      if Span.NewParagraph then
      begin
        Content := Content + '\par' + sLineBreak;
        WasTag := False;
      end;
    end;

    Result := Content;
  end;

begin
  FinishInsert;
  FontNames := TStringList.Create;
  FontNames.Duplicates := dupIgnore;
  FontNames.Sorted := True;

  Colors := TStringList.Create;
  Colors.Duplicates := dupIgnore;
  Colors.Sorted := True;
  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    FontNames.Add(Span.FontName);
    Colors.Add(IntToHex(ColorToRGB(Span.Color) and $FFFFFF, 6));
    Colors.Add(IntToHex(ColorToRGB(Span.BackgroundColor) and $FFFFFF, 6));
  end;

  Result := CreateHeader + GetSpanContents + '}';

  FontNames.Free;
  Colors.Free;
end;

procedure TMWKEdit.DecIndent;
var
  C: Integer;
  Range: TSpanRange;
  SaveIndent: Integer;
begin
  if FReadOnly or FPlain then
    Exit;
  FinishInsert;

  Range := GetSelectedParagraphs;

  if (Range.Start = -1) then
  begin
    if (FGlobalIndent > 0) then
    begin
      FUndoRedo.AddActionToUndo(UINDENT, atUndo, Integer(Self), Integer(@Range));
      Dec(FGlobalIndent);
      FUndoRedo.AddActionToUndo(UINDENT, atRedo, Integer(Self), Integer(@Range));
    end;
  end
  else
  begin
    FUndoRedo.AddActionToUndo(UINDENT, atUndo, Integer(Self), Integer(@Range));

    for C := Range.Start to Range.Stop do
      SetParagraphIndent(C, GetParagraphIndent(C) - 1);

    FUndoRedo.AddActionToUndo(UINDENT, atRedo, Integer(Self), Integer(@Range));

    Range := GetSelectedParagraphs;
    Range.Start := FindLineOfSpan(Range.Start);
    Range.Stop := FindLineOfSpan(Range.Stop) + 1;

    ReBuildLines(Range.Start, Range.Stop);
  end;

  CalcSpellCheck(False);
  SaveIndent := FGlobalIndent;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  FGlobalIndent := SaveIndent;
  if not ScrollInSight(SetCaretPos(True)) then
    Invalidate;
end;

procedure TMWKEdit.SetBackColor(const Value: TColor);
begin
  if (FBackColor = Value) or FPlain or FReadOnly then
    Exit;

  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if FSelLength > 0 then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;

  FBackColor := Value;

  SetSpanValues(ctBackColor);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FBackColor := Value;
end;

procedure TMWKEdit.SetBold(const Value: Boolean);
begin
  if FBold = Value then
    Exit;

  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FBold := Value;

  SetSpanValues(ctBold);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FBold := Value;
end;

procedure TMWKEdit.SetColor(const Value: TColor);
var
  C: Integer;
begin
  if FBrush.Color = Value then
    Exit;

  for C := 0 to FItems.Count - 1 do
    if (FItems[C].BackgroundColor = FBrush.Color) or FPlain or FReadOnly then
      FItems[C].BackgroundColor := Value;

  FBrush.Color := Value;
  Invalidate;
end;

procedure TMWKEdit.SetIndent(const Value: Integer);
var
  C: Integer;
  Range: TSpanRange;
begin
  if (Value < 0) or FReadOnly or FPlain then
    Exit;

  FinishInsert;

  Range := GetSelectedParagraphs;

  FUndoRedo.AddActionToUndo(UINDENT, atUndo, Integer(Self), Integer(@Range));

  if Range.Start = -1 then
    Inc(FGlobalIndent)
  else
    for C := Range.Start to Range.Stop do
      SetParagraphIndent(C, Value);

  FUndoRedo.AddActionToUndo(UINDENT, atRedo, Integer(Self), Integer(@Range));

  Range := CalculateParagraphRange;

  Range.Start := FindLineOfSpan(Range.Start);
  Range.Stop := FindLineOfSpan(Range.Stop) + 1;

  ReBuildLines(Range.Start, Range.Stop);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  if not ScrollInSight(SetCaretPos(True)) then
    Invalidate;
end;

procedure TMWKEdit.SetIndentSize(const Value: Integer);
begin
  if FIndentSize <> Value then
  begin
    FIndentSize := Value;
    ReBuildLines(0, MaxInt);
    CalcSelection;
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    Invalidate;
  end;
end;

procedure TMWKEdit.SetIndex(const Value: Boolean);
begin
  if FIndex = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FIndex := Value;

  SetSpanValues(ctIndex);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FIndex := Value;
end;

procedure TMWKEdit.SetInsertState(const Value: TInsertState);
begin
  FinishInsert;
  if FInsertState = Value then
    Exit;
  FInsertState := Value;
  if Assigned(FOnToggleInsert) then
    FOnToggleInsert(Self);
end;

procedure TMWKEdit.SetItalic(const Value: Boolean);
begin
  if FItalic = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FItalic := Value;

  SetSpanValues(ctItalic);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FItalic := Value;
end;

procedure TMWKEdit.SetLineBreak(const Value: Boolean);
begin
  if Value = FLineBreak then
    Exit;
  if not FMultiLine then
    Exit;
  FLineBreak := Value;
  ReBuildLines(0, MaxInt);
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
end;

procedure TMWKEdit.SetMultiLine(const Value: Boolean);
begin
  if Value = FMultiLine then
    Exit;
  FMultiLine := Value;
  if not FMultiLine then
  begin
    Clear;
    FLineBreak := False;
    ShowScrollBars := False;
    Plain := True;
    ClientHeight := GetHeightOfCurrentFont + FClientRect.Top * 2;
    CalcAreas;
  end;
end;

procedure TMWKEdit.SetOLEDragDrop(const Value: Boolean);
begin
  if FOLEDragDrop <> Value then
  begin
    FOLEDragDrop := Value;
    if Assigned(FDropTarget) then
      FDropTarget.Active := Value;
  end;
end;

procedure TMWKEdit.SetOnChange(const Value: TNotifyEvent);
begin
  FOnChange := Value;
end;

procedure TMWKEdit.SetScrollBarPositions;
begin
  if FVerticalScrollBar.Visible and not FHorizontalScrollBar.Visible then
  begin
    FVerticalScrollBar.Align := alRight;
    Exit;
  end;

  if FHorizontalScrollBar.Visible and not FVerticalScrollBar.Visible then
  begin
    FHorizontalScrollBar.Align := alBottom;
    Exit;
  end;
  
  if FHorizontalScrollBar.Visible and FVerticalScrollBar.Visible then
  begin
    FVerticalScrollBar.Align := alNone;
    FHorizontalScrollBar.Align := alNone;

    FVerticalScrollBar.SetBounds(FClientRect.Right + 1, 0,
      FVerticalScrollBar.Width, FClientRect.Bottom);

    FHorizontalScrollBar.SetBounds(0, FClientRect.Bottom + 1,
      FClientRect.Right, FHorizontalScrollBar.Height);
    Exit;
  end;
end;

procedure TMWKEdit.SetSelectedText(const Value: WideString);
var
  XML: IXMLDOMDocument;
  Span: TSpan;
begin
  FinishInsert;
  if not IsValidXML(Value, 'mwkedit') then
  begin
    if (FCurrentSpan >= 0) and (FCurrentSpan < FItems.Count) then
      Span := FItems[FCurrentSpan]
    else
      Span := nil;
    XML := CreateXMLFromPlain(Value, Span);
    InsertXML(XML);
    Exit;
  end;

  XML := CreateFromString(Value);
  if Assigned(XML) and (XML.documentElement.nodeName = 'mwkedit') then
    InsertXML(XML);
end;

procedure TMWKEdit.SetSelLength(const Value: Integer);
begin
  if FInsertTimer.Enabled then
    FInsertTimer.OnTimer(Self);
  if (FSelLength = Value) or (FSelStart = 0) then
    Exit;

  if FSelLength > 0 then
    RemoveSelection;

  FNegativeSelection := (Value < 0);
  if Value < 0 then
  begin
    FSelStart := FSelStart + Value;
    FSelLength := -Value;
    while FSelStart < 1 do
    begin
      Inc(FSelStart);
      Dec(FSelLength);
    end;
    CalcSelection;
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnSelectionChange;
    Exit;
  end;

  FSelLength := Value;
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  DoOnSelectionChange;
end;

procedure TMWKEdit.SetSelStart(const Value: Integer);
begin
  if FInsertTimer.Enabled then
    FInsertTimer.OnTimer(Self);
    
  if (FSelStart = Value) and (FSelLength = 0) then
    Exit;
  if Value < 0 then
    Exit;

  if FSelLength > 0 then
    RemoveSelection;

  if Value = 0 then
    FSelStart := 1
  else
    FSelStart := Value;
  FSelLength := 0;
  CalcSelection;
  FNegativeSelection := False;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  DoOnSelectionChange;
end;

procedure TMWKEdit.SetSpaced(const Value: Boolean);
begin
  if FSpaced = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FSpaced := Value;

  SetSpanValues(ctSpaced);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FSpaced := Value;
end;

procedure TMWKEdit.SetStrikeOut(const Value: Boolean);
begin
  if FStrikeOut = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FStrikeOut := Value;

  SetSpanValues(ctStrikeOut);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FStrikeOut := Value;
end;

procedure TMWKEdit.SetSub(const Value: Boolean);
begin
  if FSub = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FSub := Value;
  if FSub then
    FSuper := False;

  SetSpanValues(ctSub);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FSub := Value;
  if FSub then
    FSuper := False;

  BeginUpdate;
  try
    SetCaretPos(True);
  finally
    EndUpdate(False);
  end;
end;

procedure TMWKEdit.SetSuper(const Value: Boolean);
begin
  if FSuper = Value then
    Exit;
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FSuper := Value;
  if FSuper then
    FSub := False;

  SetSpanValues(ctSuper);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FSuper := Value;
  if FSuper then
    FSub := False;

  BeginUpdate;
  try
    SetCaretPos(True);
  finally
    EndUpdate(False);
  end;
end;

procedure TMWKEdit.SetXML(const Value: WideString);
var
  XML: IXMLDOMDocument;
begin
  FinishInsert;
  XML := CreateXML(True, '');
  if (XML.loadXML(Value) and (XML.documentElement.nodeName = 'mwkedit')) then
    SetXMLText(XML)
  else
    SetPlainText(Value);
end;

procedure TMWKEdit.SetXMLText(const XML: IXMLDOMDocument);
var
  C: Integer;
begin
  FinishInsert;
  try
    Inc(FUpdateCount);
    FSelStart := 1;
    FSelLength := 0;
    FOldCaretX := -1;

    FItems.Clear;
    FLineCount := 0;
    FlastEmpty := True;
    ParseXML(XML, 0, C, False);
    if FItems.Count > 0 then
      FSelStart := 1
    else
      FSelStart := 0;

    FSelLength := 0;
    ReBuildLines(0, MaxInt);
    CalcSpellCheck(False);

    CalcSelection;
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnChange;
    DoOnSelectionChange;
  finally
    Dec(FUpdateCount);
  end;
  Invalidate;
end;

procedure TMWKEdit.SetUnderline(const Value: Boolean);
begin
  if FUnderline = Value then
    Exit;

  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (ReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FUnderline := Value;

  SetSpanValues(ctUnderline);
  ReBuildLines(FStartLine, FEndLine);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FUnderline := Value;
end;

function TMWKEdit.GetBorderStyle: TBorderStyle;
begin
  Result := FBorderStyle;
end;

procedure TMWKEdit.SetBorderStyle(const Value: TBorderStyle);
begin
  FBorderStyle := Value;
  CalcAreas;
  RecreateWnd;
end;

procedure TMWKEdit.ShowDragCaret(PT: TPoint);
begin
  if FShowDragCaret then
    PaintDragCaret;
  FShowDragCaret := True;
  CalcDragMousePos(PT.X, PT.Y);
  PaintDragCaret;
end;

function TMWKEdit.Shrink(Rect: TRect; Size: Integer = 1): TRect;
begin
  Result := Rect;
  Inc(Result.Left, Size);
  Inc(Result.Top, Size);
  Dec(Result.Right, Size);
  Dec(Result.Bottom, Size);
end;

function TMWKEdit.Expand(Rect: TRect; Size: Integer = 1): TRect;
begin
  Result := Rect;
  Dec(Result.Left, Size);
  Dec(Result.Top, Size);
  Inc(Result.Right, Size);
  Inc(Result.Bottom, Size);
end;

procedure TMWKEdit.CalcAreas;
var
  OldClientWidth,
  OldClientHeight: Integer;
begin
  OldClientHeight := FClientRect.Bottom - FClientRect.Top;
  FClientRect := ClientRect;
  FClientRect := Shrink(FClientRect);

  if FBorderStyle = bsSingle then
    FClientRect := Shrink(FClientRect);

  if FShowVerticalScrollBar and FShowScrollBars then
    Dec(FClientRect.Right, FVerticalScrollBar.Width);

  if FShowHorizontalScrollBar and FShowScrollBars then
    Dec(FClientRect.Bottom, FHorizontalScrollBar.Height);

  OldClientWidth := ClientWidth;
  ClientWidth := FClientRect.Right - FClientRect.Left;

  if ClientWidth <> OldClientWidth then
  begin
    ReBuildLines(0, MaxInt);
    CalcSpellCheck(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    SetVerticalScrollBar;
    SetHorizontalScrollBar;
    SetScrollBarPositions;
    Exit;
  end;

  if FClientRect.Bottom - FClientRect.Top <> OldClientHeight then
  begin
    if FShowVerticalScrollBar <> (FTextHeight > FClientRect.Bottom - FCLientRect.Top) then
    begin
      ReBuildLines(0, MaxInt);
      CalcSpellCheck(False);
      FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    end;
    SetVerticalScrollBar;
    SetHorizontalScrollBar;
    SetScrollBarPositions;
  end;
end;

function TMWKEdit.GetCursorPos: Integer;
begin
  if (FSelLength = 0) or FNegativeSelection then
    Result := FSelStart
  else
    Result := FSelStart + FSelLength;
end;

function TMWKEdit.GetEmpty: Boolean;
begin
  Result := FItems.Count = 0;
end;

function TMWKEdit.GetEndOfLinePos(WithLineBreak: Boolean): Integer;
var
  Span: TSpan;
  C: Integer;
  CursorPos: Integer;
  PosInSpan: Integer;
begin
  C := FCurrentSpan;
  PosInSpan := FPosInSpan;

  CursorPos := GetCursorPos;
  if C = -1 then
  begin
    Result := 0;
    Exit;
  end;
  Span := FItems[C];

  if PosInSpan > Span.Length then
  begin
    Result := CursorPos;
    Exit;
  end;

  if (PosInSpan = 1) and (C > 0) and (FItems[C - 1].LineBreak) then
  begin
    if not (FItems[C - 1].NewLine or FItems[C - 1].NewParagraph) and
       not FCommingFromRight then
    begin
      Result := CursorPos;
      Exit;
    end;

    if (FItems[C - 1].NewLine or FItems[C - 1].NewParagraph) and
       (FSelLength > 0) then
    begin
      Result := CursorPos;
      Exit;
    end;
  end;

  Inc(CursorPos, Span.Length - PosInSpan + 1);
  if Span.LineBreak then
  begin
    if (Span.NewLine or Span.NewParagraph) and (not WithLineBreak) then
      Dec(CursorPos);
    FCommingFromRight := False;
    Result := CursorPos;
    Exit;
  end;

  while C < FItems.Count - 1 do
  begin
    Inc(C);
    Span := FItems[C];
    Inc(CursorPos, Span.Length);

    if Span.LineBreak then
    begin
      if (Span.NewLine or Span.NewParagraph) and (not WithLineBreak) then
        Dec(CursorPos);
      FCommingFromRight := False;
      Result := CursorPos;
      Exit;
    end;
  end;
  Result := CursorPos;
end;

function TMWKEdit.GetStartOfLinePos: Integer;
var
  Span: TSpan;
  C: Integer;
  CursorPos: Integer;
  PosInSpan: Integer;
begin
  C := FCurrentSpan;

  CursorPos := GetCursorPos;
  PosInSPan := FPosInSpan;

  if C = -1 then
  begin
    Result := 0;
    Exit;
  end;
  Span := FItems[C];

  if PosInSpan > Span.Length then
  begin
    if (Span.NewLine or Span.NewParagraph) and (FSelLength = 0) then
    begin
      Result := CursorPos;
      Exit;
    end;
    PosInSpan := Span.Length;
    Dec(CursorPos);
    if (PosInSpan = 1) and (C > 0) and (FItems[C - 1].LineBreak) then
    begin
      Result := CursorPos;
      FCommingFromRight := True;
      Exit;
    end;
  end;

  if (PosInSpan = 1) and (C > 0) and (FItems[C - 1].LineBreak) then
  begin
    if FCommingFromRight then
    begin
      Result := CursorPos;
      Exit;
    end;

    if (FItems[C - 1].NewLine or FItems[C - 1].NewParagraph) and
       (FSelLength = 0) then
    begin
      Result := CursorPos;
      Exit;
    end;

    Dec(C);
    Span := FItems[C];
    Dec(CursorPos);
    PosInSpan := Span.Length;
  end;

  Dec(CursorPos, PosInSpan - 1);
  while C > 0 do
  begin
    Dec(C);
    Span := FItems[C];
    if Span.LineBreak then
      Break;
    Dec(CursorPos, Span.Length);
  end;
  Result := CursorPos;
  FCommingFromRight := True;
end;

function TMWKEdit.GetPrevWordPos: Integer;
var
  C: Integer;
  CursorPos: Integer;
  PosInSpan: Integer;
  Span: TSpan;
  WC: WideChar;
begin
  C := FCurrentSpan;
  PosInSpan := FPosInSpan;

  CursorPos := GetCursorPos;
  Span := FItems[C];
  FCommingFromRight := True;

  if (PosInSpan = 1) and (C > 0) then
  begin
    Dec(C);
    Span := FItems[C];
    PosInSpan := Span.Length + 1;
  end;

  WC := Span.Text[PosInSpan - 1];

  if (WC = ' ') or
     (WC = '-') or
     (WC = cHyphen) or
     (WC = cProtectSpace) then
  begin
    repeat
      Dec(PosInSpan);
      Dec(CursorPos);
      if (PosInSpan = 1) and (C > 0) then
      begin
        Dec(C);
        Span := FItems[C];
        PosInSpan := Span.Length + 1;
      end;
    until Span.Text[PosInSpan - 1] <> WC;
    if Span.Text[PosInSpan - 1] = #13 then
    begin
      Result := CursorPos;
      Exit;
    end;
  end;

  while CursorPos > 1 do
  begin
    Dec(PosInSpan);
    Dec(CursorPos);
    if (Span.Text[PosInSpan] = ' ') or
       (Span.Text[PosInSpan] = '-') or
       (Span.Text[PosInSpan] = cHyphen) or
       (Span.Text[PosInSpan] = cProtectSpace) or
       (Span.Text[PosInSpan] = #13) then
    begin
      if ((Span.Text[PosInSpan] = ' ') or
          (Span.Text[PosInSpan] = '-') or
          (Span.Text[PosInSpan] = cHyphen) or
          (Span.Text[PosInSpan] = cProtectSpace)) and
         (PosInspan < Span.Length) then
        Inc(CursorPos);
      Break;
    end;

    if (PosInSpan = 1) and (C > 0) then
    begin
      Dec(C);
      Span := FItems[C];
      PosInSpan := Span.Length + 1;
      Inc(CursorPos);
    end;
  end;
  Result := CursorPos;
end;

function TMWKEdit.GetNextWordPos: Integer;
var
  C: Integer;
  CursorPos: Integer;
  PosInSpan: Integer;
  Span: TSpan;
  WC: WideChar;
begin
  C := FCurrentSpan;
  PosInSpan := FPosInSpan;

  CursorPos := GetCursorPos;
  Span := FItems[C];

  if (PosInSpan > Span.Length) then
  begin
    Result := CursorPos;
    Exit;
  end;

  if (PosInSpan = Span.Length) and
     (Span.NewLine or Span.NewParagraph) then
  begin
    Result := CursorPos + 1;
    FCommingFromRight := False;
    Exit;
  end;


  while C < FItems.Count do
  begin
    if (Span.Text[PosInSpan] = ' ') or
       (Span.Text[PosInSpan] = '-') or
       (Span.Text[PosInSpan] = cHyphen) or
       (Span.Text[PosInSpan] = cProtectSpace) or
       (Span.Text[PosInSpan] = #13) then
    begin
      WC := Span.Text[PosInSpan];
      if (WC = ' ') or
         (WC = '-') or
         (WC = cHyphen) or
         (WC = cProtectSpace) then
      begin
        repeat
          Inc(CursorPos);
          Inc(PosInSpan);
          if (PosInSpan > Span.Length) then
          begin
            if C < FItems.Count - 1 then
            begin
              Inc(C);
              Span := FItems[C];
              PosInSpan := 1;
            end else
              Break;
          end;
        until Span.Text[PosInSpan] <> WC;
      end;
      Break;
    end;

    Inc(PosInSpan);
    Inc(CursorPos);

    if (PosInSpan > Span.Length) then
    begin
      Inc(C);
      if C < FItems.Count then
      begin
        Span := FItems[C];
        PosInSpan := 1;
        Dec(CursorPos);
      end;
    end;
  end;

  Result := CursorPos;
  FCommingFromRight := False;
end;

function TMWKEdit.GetCanRedo: Boolean;
begin
  Result := not (FUndoRedo.IsLast or FPendingInsertFinish);
end;

function TMWKEdit.GetCanUndo: Boolean;
begin
  Result := not FUndoRedo.IsFirst or FPendingInsertFinish;
end;

procedure TMWKEdit.GetCaret(var X, Y: Integer);
var
  C, L, W, Dist: Integer;
  Span: TSpan;
  CursorPos: Integer;
begin
  CursorPos := GetCursorPos;
  Y := 0;

  L := 0;
  if FItems.Count > 0 then
    W := FItems[0].Indent * FIndentSize
  else
    W := FGlobalIndent * FIndentSize;
  X := W;

  for C := 0 to FItems.Count - 1 do
  begin
    Span := FItems[C];
    Dist := CursorPos - Span.Length;
    if (Dist <= 0) or ((Dist = 1) and Span.LineBreak) then
    begin
      if (Dist = 1) and Span.LineBreak then
      begin
        Y := GetHeightToLine(L);
        if (C < FItems.Count - 1) then
          X := TSpan(FItems[C + 1]).Indent * FIndentSize
        else
          X := FGlobalIndent * FIndentSize;
        Break;
      end;

      Y := GetHeightToLine(L - 1);
      if not Span.IsTab then
        X := W + Span.CalcTextWidth(Copy(Span.Text, 1, CursorPos - 1))
      else
        if CursorPos = 2 then
          X := W + Span.Width
        else
          X := W;

      Break;
    end;
    Dec(CursorPos, Span.Length);
    Inc(W, Span.Width);
    if (Span.NewLine or Span.NewParagraph) or
       (Span.LineBreak and (CursorPos > 1)) then
    begin
      if (not FCommingFromRight) and (FSelLength > 0) and (CursorPos = 1) and
         (not FNegativeSelection) then
      begin
        Y := GetHeightToLine(L - 1);
        X := W;
        if Span.NewLine then
          X := X + Span.CalcTextWidth(cNewLine);
        if Span.NewParagraph then
          X := X + Span.CalcTextWidth(cNewParagraph);
        Break;
      end;
      Inc(L);
      W := Span.Indent * FIndentSize;
      if Span.NewParagraph and (C < FItems.Count - 1) then
        W := FItems[C + 1].Indent * FIndentSize;

      if (C = FItems.Count - 1) and (FPosInSpan > FItems.Last.Length) then
        W := FGlobalIndent * FIndentSize;
      //W := 0;
    end;
    if Span.LineBreak and (CursorPos = 1) and
       (not (Span.NewLine or Span.NewParagraph)) then
    begin
      if FCommingFromRight then
      begin
        Y := GetHeightToLine(L);
        if (C < FItems.Count - 1) then
          X := FItems[C + 1].Indent * FIndentSize
        else
          X := FGlobalIndent;
      end else
      begin
        Y := GetHeightToLine(L);
        X := W;
      end;
      Break;
    end;

    if C = FItems.Count - 1 then
    begin
      Y := GetHeightToLine(L - 1);
      X := W;
     end;
  end;

  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);

  X := X + FClientRect.Left;
  Y := Y + FClientRect.Top;
end;

function TMWKEdit.GetIsChanged: Boolean;
begin
  Result := not FUndoRedo.Stable;
end;

function TMWKEdit.SetCaretPos(CalcHeight: Boolean): TPoint;
var
  X, Y, H: Integer;
  Start: Integer;
begin
  if (not Focused) then
  begin
    HideCaret(Handle);
    DestroyCaret;
  end;

  GetCaret(X, Y);

  if FRightToLeft then
    X := FClientRect.Right - X + FClientRect.Left;

  FCaretX := X - FClientRect.Left;

  Result := Point(X - FHorizontalScrollBar.Position, Y - FVerticalScrollBar.Position);
  if (not Focused) then
    Exit;

  if CalcHeight then
  begin
    if (FOldFontName <> FFontName) or (FOldSuper <> FSuper) or
       (FOldSub <> FSub) or (FOldFontHeight <> FontHeight) then
    begin
      CalcCaretPos(Start, FCaretHeight);
      FOldCaretStart := Start;
      FOldCaretHeight := FCaretHeight;
      FOldFontName := FFontName;
      FOldSuper := FSuper;
      FOldSub := FSub;
      FOldFontHeight := FFontHeight;
    end
    else
    begin
      Start := FOldCaretStart;
      FCaretHeight := FOldCaretHeight;
    end;
                       
    H := GetBaselineOfLine(FCurrentline - 1);
    if H > 0 then
      H := H + Start;

    if Y + H - FVerticalScrollBar.Position + FCaretHeight > FClientRect.Bottom then
      FCaretHeight := FClientRect.Bottom - (Y + H - FVerticalScrollBar.Position);

    CreateCaret(Handle, 0, 0, FCaretHeight);
    Windows.SetCaretPos(X - FHorizontalScrollBar.Position, Y + H - FVerticalScrollBar.Position);
    ShowCaret(Handle);
  end;
end;

procedure TMWKEdit.SetCaretTo(Where: TNewCaretPos);
var
  CursorPos: Integer;
  OldStart: Integer;
  Point: TPoint;
begin
  CursorPos := GetCursorPos;
  case Where of
    ncLeft:
    begin
      FCommingFromRight := True;
      SelStart := CursorPos - 1;
      FOldCaretX := -1;
    end;

    ncRight:
    begin
      FCommingFromRight := False;
      SelStart := CursorPos + 1;
      FOldCaretX := -1;
    end;

    ncUp:
    begin
      OldStart := FSelStart;
      GetCaretPos(Point);

      if FOldCaretX = -1 then
        FOldCaretX := Point.X;

      CalcMousePos(FOldCaretX, Point.Y - GetHeightOfLine(FCurrentLine - 1), False);
      SelStart := GetMousePosAbsolute;
      if (SelStart = OldStart) and (OldStart > 1) then
        SelStart := SelStart - 1;
    end;

    ncDown:
    begin
      GetCaretPos(Point);

      if FOldCaretX = -1 then
        FOldCaretX := Point.X;

      CalcMousePos(FOldCaretX, Point.Y + GetHeightOfLine(FCurrentLine - 1), False);
      SelStart := GetMousePosAbsolute;
    end;

    ncHome:
    begin
      SelStart := GetStartOfLinePos;
      FOldCaretX := -1;
    end;

    ncEnd:
    begin
      SelStart := GetEndOfLinePos(False);
      FOldCaretX := -1;
    end;

    ncFirst:
    begin
      FCommingFromRight := True;
      SelStart := 1;
      FOldCaretX := -1;
    end;

    ncLast:
    begin
      FCommingFromRight := False;
      SelStart := MaxInt;
      FOldCaretX := -1;
    end;

    ncPageDown:
    begin
      FOldCaretX := -1;
    end;

    ncPageUp:
    begin
      FOldCaretX := -1;
    end;

    ncPrevWord:
    begin
      SelStart := GetPrevWordPos;
      FOldCaretX := -1;
    end;

    ncNextWord:
    begin
      SelStart := GetNextWordPos;
      FOldCaretX := -1;
    end;
  end;
end;

procedure TMWKEdit.SetSelection(Where: TNewCaretPos);
var
  X: Integer;
  Point: TPoint;
  Negative: Boolean;
begin
  case Where of
    ncLeft:
    begin
      FCommingFromRight := True;
      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength - 1;
      end else
        SelLength := SelLength - 1;
      FOldCaretX := -1;
    end;

    ncRight:
    begin
      FCommingFromRight := False;
      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + 1;
      end else
        SelLength := SelLength + 1;
      FOldCaretX := -1;
    end;

    ncUp:
    begin
      if FSelLength = 0 then
        FNegativeSelection := False;

      Negative := FNegativeSelection;

      GetCaretPos(Point);

      if FOldCaretX = -1 then
        FOldCaretX := Point.X;

      CalcMousePos(FOldCaretX, Point.Y - GetHeightOfLine(FCurrentLine- 1), False);

      if (FSelLength > 0) then
        FNegativeSelection := Negative;

      MSelStart := GetMousePosAbsolute;

      if (FPosInMouseSpan > FItems[FCurrentMouseSpan].Length) and
         (FCurrentMouseSpan > 0) then
      begin
        Inc(FCurrentMouseSpan);
        FPosInMouseSpan := 1;
        FOldCaretX := -1;
      end;

      if FNegativeSelection then
      begin
        X := MSelStart - FSelStart - FSelLength;
        FSelStart := FSelStart + FSelLength;
        SelLength := X;
      end else
        SelLength := MSelStart - FSelStart;
    end;

    ncDown:
    begin
      if FSelLength = 0 then
        FCommingFromRight := True;
      GetCaretPos(Point);

      if FOldCaretX = -1 then
        FOldCaretX := Point.X;

      CalcMousePos(FOldCaretX, Point.Y + GetHeightOfLine(FCurrentLine - 1), False);

      MSelStart := GetMousePosAbsolute;
      if FNegativeSelection then
      begin
        X := MSelStart - FSelStart - FSelLength;
        FSelStart := FSelStart + FSelLength;
        SelLength := X;
      end else
          SelLength := MSelStart - FSelStart;
    end;

    ncHome:
    begin
      X := GetStartOfLinePos - GetCursorPos;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := SelLength + X;
      FOldCaretX := -1;
    end;

    ncEnd:
    begin
      X := GetEndOfLinePos(True) - GetCursorPos;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := SelLength + X;
      FOldCaretX := -1;
    end;

    ncFirst:
    begin
      X := 1 - GetCursorPos;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := SelLength + X;
      FOldCaretX := -1;
    end;

    ncLast:
    begin
      X := MaxInt;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := MaxInt;
      FOldCaretX := -1;
    end;

    ncPageDown:
    begin
      FOldCaretX := -1;
    end;

    ncPageUp:
    begin
      FOldCaretX := -1;
    end;

    ncPrevWord:
    begin
      X := GetPrevWordPos - GetCursorPos;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := SelLength + X;
      FOldCaretX := -1;
    end;

    ncNextWord:
    begin
      X := GetNextWordPos - GetCursorPos;

      if FNegativeSelection then
      begin
        FSelStart := FSelStart + FSelLength;
        SelLength := -SelLength + X;
      end else
        SelLength := SelLength + X;
      FOldCaretX := -1;
    end;
  end;
end;

procedure TMWKEdit.Delete(Next: Boolean);
begin
  if FReadOnly then
    Exit;
  if Next then
    FSelStart := GetCursorPos
  else
    FSelStart := GetCursorPos - 1;

  if FSelStart = 0 then
  begin
    FSelStart := 1;
    FOldCaretX := -1;
    Exit;
  end;

  FSelLength := 1;
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  DeleteSelected(True, True);
end;

procedure TMWKEdit.DeleteSelected;
begin
  FinishInsert;
  DeleteSelected(True, True);
  Invalidate;
end;

procedure TMWKEdit.AddCharacter(const Ch: WideChar);
var
  Span, NewSpan: TSpan;
  Rec: TSpanValues;
  IR: TInsertRec;
begin
  if FReadOnly then
    Exit;

  if not FMultiLine and ((Ch = #13) or (Ch = #14)) then
    Exit;
      
  if FRollBack = 0 then
  begin
    FInsertTimer.Enabled := False;
    if not FPendingInsertFinish then
    begin
      IR.Edit := Self;
      if FInsertState = isOverWrite then
        FUndoRedo.AddActionToUndo(UINSERT, atUndo, Integer(@IR), -1)
      else
        FUndoRedo.AddActionToUndo(UINSERT, atUndo, Integer(@IR), 1);
      FOldInsertAction := IR.OldInsertAction;
      FSelStartBeforeInsert := FSelStart;
      FTextBeforeInsert := nil;
    end;

    if FInsertState = isOverWrite then
    begin
      Inc(FUpdateCount);
      if FSelLength > 0 then
      begin
        SelLength := SelLength + 1;
        FTextBeforeInsert := CreateXML(True, '');
        FTextBeforeInsert.loadXML(SelectedText);
        SelLength := SelLength - 1;
      end
      else
      begin
        SelLength := 1;
        if not Assigned(FTextBeforeInsert) then
        begin
          FTextBeforeInsert := CreateXML(True, '');
          FTextBeforeInsert.loadXML(SelectedText);
        end else
          AddToXML(FTextBeforeInsert, True);
        SelLength := 0;
      end;
      Dec(FUpdateCount);
    end;

    FInsertTimer.Enabled := True;
  end;

  if not FPendingInsertFinish then
    if FInsertState = isOverWrite then
      FSelLengthBeforeInsert := -(FSelLength + 1)
    else
      FSelLengthBeforeInsert := FSelLength;

  if (FSelStart = 0) and (FSelLength = 0) then
    FSelStart := 1;

  FNegativeSelection := False;
  FOldCaretX := -1;

  Rec.IsBold := FBold;
  Rec.IsItalic := FItalic;
  Rec.IsUnderline := FUnderline;
  Rec.IsStrikeOut := FStrikeOut;
  Rec.IsSpaced := FSpaced;
  Rec.IsSuper := FSuper;
  Rec.IsSub := FSub;
  Rec.IsIndex := FIndex;
  Rec.Color := FFontColor;
  Rec.BackColor := FBackColor;
  Rec.FontName := FFontName;
  Rec.FontHeight := FFontHeight;
  Rec.IsNewLine := False;

  DeleteSelected(False, False);
  if not FPendingInsertFinish then
    FSelStartBeforeInsert := FSelStart;

  FPendingInsertFinish := True;

  if FItems.Count = 0 then
  begin // No text so far
    Span := TSpan.Create;
    Span.FontName := FontName;
    Span.FontHeight := FontHeight;
    Span.RightToLeft := FRightToLeft;
    Span.Values := Rec;
    Span.ShowTags := FShowTags;
    Span.ShowSpellErrors := FShowSpellErrors;
    Span.Indent := FGlobalIndent;
    Span.Text := Ch;
    FItems.Add(Span);
    FSelStart := 2;
    ReBuildLines(FStartLine, FEndLine);
    CalcSpellCheck(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnChange;
    DoOnSelectionChange;
    Exit;
  end;

  Span := FItems[FCurrentSpan];

  if (FPosInSpan > Span.Length) and (Span.NewLine or Span.NewParagraph) then
  begin // After last text + last character = NewLine/NewParagraph
    NewSpan := TSpan.Create;
    NewSpan.Assign(Span);
    NewSpan.Indent := FGlobalIndent;
    NewSpan.Values := Rec;
    NewSpan.Text := Ch;
    NewSpan.LineBreak := NewSpan.NewLine or NewSpan.NewParagraph;
    FItems.Add(NewSpan);
    FSelStart := SelStart + 1;
    ReBuildLines(FStartLine, FEndLine);
    CalcSpellCheck(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnChange;
    DoOnSelectionChange;
    Exit;
  end;

  if (FInsertState = isOverWrite) then
  begin // OverWrite
    if Span.SameValues(Rec) and not ((Ch = cTab) or Span.IsTab) then
    begin // Same formatting
      if (Ch = #13) or (Ch = #14) then
      begin
        if FPosInSpan + 1 <= Span.Length then
        begin
          NewSpan := TSpan.Create;
          NewSpan.Assign(Span);
          NewSpan.SetTextWithoutChange(Copy(Span.Text, FPosInSpan + 1, MaxInt));
          FItems.Insert(FCurrentSpan + 1, NewSpan);
        end;
        Span.Text := Copy(Span.Text, 1, FPosInSpan - 1) + Ch;
        if Ch = #13 then
          Span.NewParagraph := True
        else
          Span.NewLine := True;
      end
      else
      begin
        if Span.Text[FPosInSpan] in [WideChar(#13), WideChar(#14)] then
        begin
          Span.NewLine := False;
          Span.NewParagraph := False;
        end;
        Span.SetTextWithoutChange(Copy(Span.Text, 1, FPosInSpan - 1) + Ch +
          Copy(Span.Text, FPosInSpan + 1, MaxInt));
      end;
      RemoveEmptySpans;
      FSelStart := FSelStart + 1;
    end
    else
    begin // Different formatting
      if FPosInSpan > 1 then
      begin
        NewSpan := Span.Split(FPosInSpan);
        if Assigned(NewSpan) then
        begin
          if NewSpan.Length > 1 then
          begin
            NewSpan.SetTextWithoutChange(Copy(NewSpan.Text, 2, MaxInt));
            FItems.Insert(FCurrentSpan + 1, NewSpan);
          end else
            NewSpan.Free;
        end;
      end
      else
      begin
        Dec(FCurrentSpan);
        if Span.Length = 1 then
        begin
          Span.NewLine := False;
          Span.NewParagraph := False;
        end;
        Span.SetTextWithoutChange(Copy(Span.Text, 2, MaxInt))
      end;

      NewSpan := TSpan.Create;
      NewSpan.Assign(Span);
      NewSpan.FontName := FontName;
      NewSpan.FontHeight := FontHeight;
      NewSpan.Values := Rec;
      NewSpan.Text := Ch;
      FItems.Insert(FCurrentSpan + 1, NewSpan);

      RemoveEmptySpans;
      FSelStart := FSelStart + 1;
    end;

    ReBuildLines(FStartLine - 1, FEndLine + 1);
    CalcSpellCheck(False);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnChange;
    DoOnSelectionChange;
    Exit;
  end;

  if Span.SameValues(Rec) and not ((Ch = cTab) or Span.IsTab) then
  begin // Same formatting
    if (Ch = #13) or (Ch = #14) then
    begin
      NewSpan := TSpan.Create;
      NewSpan.Assign(Span);
      NewSpan.SetTextWithoutChange(Copy(Span.Text, FPosInSpan, MaxInt));
      FItems.Insert(FCurrentSpan + 1, NewSpan);
      Span.Text := Copy(Span.Text, 1, FPosInSpan - 1) + Ch;
    end else
      Span.SetTextWithoutChange(Copy(Span.Text, 1, FPosInSpan - 1) + Ch +
        Copy(Span.Text, FPosInSpan, MaxInt));
    RemoveEmptySpans;
    FSelStart := FSelStart + 1;
  end
  else
  begin // Different formatting
    if FPosInSpan > 1 then
    begin
      NewSpan := Span.Split(FPosInSpan);
      if Assigned(NewSpan) then
        FItems.Insert(FCurrentSpan + 1, NewSpan);
    end else
      Dec(FCurrentSpan);

    NewSpan := TSpan.Create;
    NewSpan.Assign(Span);
    NewSpan.Values := Rec;
    NewSpan.Text := Ch;
    FItems.Insert(FCurrentSpan + 1, NewSpan);
    RemoveEmptySpans;
    FSelStart := FSelStart + 1;
  end;
  ReBuildLines(FStartLine - 1, FEndLine + 1);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  DoOnChange;
  DoOnSelectionChange;
end;

procedure TMWKEdit.IncIndent;
var
  C: Integer;
  Range: TSpanRange;
  SaveIndent: Integer;
begin
  if FReadOnly or FPlain then
    Exit;
  FinishInsert;

  Range := GetSelectedParagraphs;

  FUndoRedo.AddActionToUndo(UINDENT, atUndo, Integer(Self), Integer(@Range));

  if Range.Start = -1 then
    Inc(FGlobalIndent)
  else
    for C := Range.Start to Range.Stop do
      SetParagraphIndent(C, GetParagraphIndent(C) + 1);

  FUndoRedo.AddActionToUndo(UINDENT, atRedo, Integer(Self), Integer(@Range));

  if Range.Start <> -1 then
  begin
    Range := CalculateParagraphRange;
    Range.Start := FindLineOfSpan(Range.Start);
    Range.Stop := FindLineOfSpan(Range.Stop) + 1;

    ReBuildLines(Range.Start, Range.Stop);
  end;
  
  CalcSpellCheck(False);
  SaveIndent := FGlobalIndent;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  FGlobalIndent := SaveIndent;
  if not ScrollInSight(SetCaretPos(True)) then
    Invalidate;
end;

function TMWKEdit.InsertOLEData(const DataObj: IDataObject;
  const GrfKeyState: Integer; PT: TPoint; var dwEffect: Integer): HRESULT;
var
  LRecStg: TStgMedium;
  Text: WideString;
  Title: WideString;
  LPtrLock: Pointer;
  LIntGlobalSize: Integer;
  RTF: TRTFToXML;
begin
  Result := E_FAIL;
  Text := '';

  if Succeeded(GetIDataObjectData(DataObj, CF_MWKTEXT, LRecStg)) then
    FOLEFormat := CF_MWKTEXT
  else
    if Succeeded(GetIDataObjectData(DataObj, CF_RTF, LRecStg)) then
      FOLEFormat := CF_RTF
    else
      if Succeeded(GetIDataObjectData(DataObj, CF_UNICODETEXT, LRecStg)) then
        FOLEFormat := CF_UNICODETEXT;

  if FOleFormat = 0 then
    Exit;

  try
    if LRecStg.tymed in [TYMED_HGLOBAL, TYMED_MFPICT] then
    begin
      LPtrLock := GlobalLock(LRecStg.hGlobal);
      LIntGlobalSize := GlobalSize(LRecStg.hGlobal);
    end else
    begin
      LPtrLock := nil;
      LIntGlobalSize := 0;
    end;

    try
      Text := GetSomeData(LPtrLock, LRecStg.hGlobal, FOLEFormat, LIntGlobalSize);
      Result := S_OK;
    finally
      if Assigned(LPtrLock) then
        GlobalUnlock(LRecStg.hGlobal);
    end;
  finally
    ReleaseStgMedium(LRecStg);
  end;

  if Result <> S_OK then
  begin
    dwEffect := DROPEFFECT_NONE;
    Exit;
  end;

  if FOLEFormat = CF_RTF then
  begin
    RTF := TRTFToXML.Create;
    Text := RTF.Parse(Text, Title).xml;
    RTF.Free;
  end;

  PT := ScreenToClient(PT);
  MouseDown(mbLeft, [], PT.X, PT.Y);

  SelectedText := Text;
end;

procedure TMWKEdit.InsertPicture(const FileName: WideString);
begin
  InsertPicture(LoadImage(FileName));
end;

procedure TMWKEdit.InsertSection;
var
  Start: Integer;
  InsertPos: Integer;
  Span, NewSpan: TSpan;
  L, OldStart, NewStart: Integer;
  IR: TInsertRec;
  Indent: Integer;

  procedure SetValues;
  var
    Rec: TSpanValues;
  begin
    FillChar(Rec, SizeOf(Rec), 0);
    Rec.Color := FFontColor;
    Rec.BackColor := FBackColor;
    Rec.FontName := FFontName;

    Span.FontName := FontName;
    Span.Values := Rec;
    Span.Updating := True;
    Span.FontHeight := FontHeight;
    Span.NewLine := False;
    Span.NewParagraph := False;
    Span.LineBreak := False;
    Span.ShowTags := FShowTags;
    Span.ShowSpellErrors := FShowSpellErrors;
    Span.RightToLeft := FRightToLeft;
    Span.Updating := False;
  end;

begin
  if FPlain or FReadOnly then
    Exit;

  Start := FindStartLine;

  L := 1;
  if FRollBack = 0 then
  begin
    FinishInsert;
    IR.Edit := Self;
    FUndoRedo.AddActionToUndo(UINSERTSECTION, atUndo, Integer(@IR), L);
  end;
  OldStart := FSelStart;
  L := FSelLength;
  DeleteSelected(False, False);

  Indent := FGlobalIndent;

  InsertPos := FCurrentSpan + 1;
  if (FPosInSpan > 1) and (FCurrentSpan <= FItems.Count - 1) and
     (FItems.Count > 0) then
  begin
    Span := FItems[FCurrentSpan];
    Indent := Span.Indent;
    NewSpan := Span.Split(FPosInSpan);
    if Assigned(NewSpan) then
      FItems.Insert(InsertPos, NewSpan);
  end else
    Dec(InsertPos);

  Span := TSectionSpan.Create;
  SetValues;
  Span.Indent := Indent;
  if InsertPos < 0 then
    InsertPos := 0;
  FItems.Insert(InsertPos, Span);

  Inc(FSelStart, 1);
  FSelLength := 0;

  ReBuildLines(Start, FEndLine + 1);
  CalcSpellCheck(False);

  if FRollBack = 0 then
  begin
    Inc(FUpdateCount);
    NewStart := FSelStart;
    SelStart := OldStart;
    SelLength := 1;
    IR.Text := nil;
    FUndoRedo.AddActionToUndo(UINSERTSECTION, atRedo, Integer(@IR), L);
    SelStart := NewStart;
    SelLength := 0;
    Dec(FUpdateCount);
  end;

  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);

  DoOnChange;
  DoOnSelectionChange;
end;

procedure TMWKEdit.InsertPicture(const Picture: TObject);
var
  Start: Integer;
  InsertPos: Integer;
  Span, NewSpan: TSpan;
  L, OldStart, NewStart: Integer;
  IR: TInsertRec;
  Indent: Integer;

  procedure SetValues;
  var
    Rec: TSpanValues;
  begin
    FillChar(Rec, SizeOf(Rec), 0);
    Rec.Color := FFontColor;
    Rec.BackColor := FBackColor;
    Rec.FontName := FFontName;

    Span.FontName := FontName;
    Span.Values := Rec;
    Span.Updating := True;
    Span.FontHeight := FontHeight;
    Span.NewLine := False;
    Span.NewParagraph := False;
    Span.LineBreak := False;
    Span.ShowTags := FShowTags;
    Span.ShowSpellErrors := FShowSpellErrors;
    Span.RightToLeft := FRightToLeft;
    Span.Updating := False;
  end;

begin
  if FPlain or FReadOnly or (Picture = nil) then
    Exit;

  Start := FindStartLine;

  L := 1;
  if FRollBack = 0 then
  begin
    FinishInsert;
    IR.Edit := Self;
    FUndoRedo.AddActionToUndo(UINSERTIMAGE, atUndo, Integer(@IR), L);
  end;
  OldStart := FSelStart;
  L := FSelLength;
  DeleteSelected(False, False);

  Indent := FGlobalIndent;

  InsertPos := FCurrentSpan + 1;
  if (FPosInSpan > 1) and (FCurrentSpan <= FItems.Count - 1) and
     (FItems.Count > 0) then
  begin
    Span := FItems[FCurrentSpan];
    Indent := Span.Indent;
    NewSpan := Span.Split(FPosInSpan);
    if Assigned(NewSpan) then
      FItems.Insert(InsertPos, NewSpan);
  end else
    Dec(InsertPos);

  Span := TGraphicSpan.Create;
  SetValues;
  Span.Indent := Indent;
  TGraphicSpan(Span).SetPicture(Picture);
  if InsertPos < 0 then
    InsertPos := 0;
  FItems.Insert(InsertPos, Span);

  Inc(FSelStart, 1);
  FSelLength := 0;

  ReBuildLines(Start, FEndLine + 1);
  CalcSpellCheck(False);

  if FRollBack = 0 then
  begin
    Inc(FUpdateCount);
    NewStart := FSelStart;
    SelStart := OldStart;
    SelLength := 1;
    IR.Text := nil;
    FUndoRedo.AddActionToUndo(UINSERTIMAGE, atRedo, Integer(@IR), L);
    SelStart := NewStart;
    SelLength := 0;
    Dec(FUpdateCount);
  end;

  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);

  DoOnChange;
  DoOnSelectionChange;
end;

procedure TMWKEdit.InsertXML(const XML: IXMLDOMDocument);
var
  Start: Integer;
  InsertPos: Integer;
  BreakCount: Integer;
  Span, NewSpan: TSpan;
  L, OldStart, NewStart: Integer;
  IR: TInsertRec;
  SR: TSpanRange;
begin
  SR := CalculateParagraphRange;

  L := System.Length(StripAllXML(XML));
  if FRollBack = 0 then
  begin
    FinishInsert;
    IR.Edit := Self;
    FUndoRedo.AddActionToUndo(UINSERT, atUndo, Integer(@IR), L);
  end;

  OldStart := FSelStart;
  L := FSelLength;
  DeleteSelected(False, False);

  InsertPos := FCurrentSpan + 1;
  if (FPosInSpan > 1) and (FCurrentSpan <= FItems.Count - 1) and
     (FItems.Count > 0) then
  begin
    Span := FItems[FCurrentSpan];
    NewSpan := Span.Split(FPosInSpan);
    if Assigned(NewSpan) then
      FItems.Insert(InsertPos, NewSpan);
  end else
    Dec(InsertPos);

  Inc(FSelStart, ParseXML(XML, InsertPos, BreakCount, False));

  FSelLength := 0;

  Start := FindLineOfSpan(SR.Start);
  if SR.Stop = MaxInt then
    FEndLine := MaxInt
  else
    FEndLine := Start + BreakCount + 1;
  ReBuildLines(Start, FEndLine);
  CalcSpellCheck(False);

  if FRollBack = 0 then
  begin
    Inc(FUpdateCount);
    NewStart := FSelStart;
    SelStart := OldStart;
    SelLength := System.Length(StripAllXML(XML));
    IR.Text := nil;
    FUndoRedo.AddActionToUndo(UINSERT, atRedo, Integer(@IR), L);
    SelStart := NewStart;
    SelLength := 0;
    Dec(FUpdateCount);
  end;

  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);

  DoOnChange;
  DoOnSelectionChange;
end;

procedure TMWKEdit.SetXMLAttributes(const Node: IXMLDOMNode; const Span: TSpan);
var
  Attributes: TStringList;
begin
  SetAttribute(Node, 'font', Span.FontName);

  if (Span.ShadowFontName <> '') then
    SetAttribute(Node, 'shadowfont', Span.ShadowFontName);

  SetAttribute(Node, 'size', Span.FontHeight);

  SetAttribute(Node, 'color', '#' + IntToHex(ColorToRGB(Span.Color), 6));

  SetAttribute(Node, 'bg-color', '#' + IntToHex(ColorToRGB(Span.BackgroundColor), 6));

  SetXMLIndent(Node, Span.Indent);

  Attributes := TStringList.Create;
  if Span.Bold then
    Attributes.Add('bold');

  if Span.Italic then
    Attributes.Add('italic');

  if Span.Underline then
    Attributes.Add('underline');

  if Span.StrikeOut then
    Attributes.Add('strikeout');

  if Span.Super then
    Attributes.Add('super');

  if Span.Sub then
    Attributes.Add('sub');

  if Span.Spaced then
    Attributes.Add('spaced');

  if Attributes.Count > 0 then
    SetAttribute(Node, 'deco', Attributes.CommaText);

  if Span.Index then
    SetAttribute(Node, 'index', True);
  Attributes.Free;
end;                           

procedure TMWKEdit.SetXMLIndent(const Node: IXMLDOMNode; Indent: Integer);
begin
  if (Node.nodeName = 'p') then
  begin
    if (Indent > 0) then
      SetAttribute(Node, 'idt', Indent)
    else
      RemoveAttribute(Node, 'idt');
  end;
end;

function TMWKEdit.CollectAsXML(Selected, Plain: Boolean): WideString;
var
  XML: IXMLDOMDocument;
begin
  Result := '';
  XML := CreateXML(True, 'mwkedit');

  if FItems.Count > 0 then
    AddToXML(XML, Selected);

  if Plain then
    Result := StripAllXML(XML)
  else
    Result := XML.xml;
end;

procedure TMWKEdit.DefineSelection(const Start, Length: Integer);
begin
  if ((FSelStart = Start) and (FSelLength = 0)) or (Start < 0) then
    Exit;

  if FSelLength > 0 then
    RemoveSelection;

  if Start = 0 then
    FSelStart := 1
  else
    FSelStart := Start;

  FNegativeSelection := (Length < 0);
  if Length < 0 then
  begin
    FSelStart := FSelStart + Length;
    FSelLength := -Length;
    while FSelStart < 1 do
    begin
      Inc(FSelStart);
      Dec(FSelLength);
    end;
    CalcSelection;
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    DoOnSelectionChange;
    Exit;
  end;

  FSelLength := Length;
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  DoOnSelectionChange;
end;

procedure TMWKEdit.Resize;
begin
  CalcAreas;
  CalcCurrentLine;
  SetCaretPos(True);
  FOldCaretX := -1;
  DoOnSelectionChange;
  inherited;
end;

procedure TMWKEdit.DoOnChange;
begin
  if FUpdateCount > 0 then
    Exit;
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TMWKEdit.DoOnSelectionChange;
begin
  if FUpdateCount > 0 then
    Exit;
  SetCaretPos(True);
  if Assigned(FOnSelectionChange) then
    FOnSelectionChange(Self);
end;

procedure TMWKEdit.DoOnVerticalScrollBarChange(Sender: TObject);
begin
  if FVerticalScrollBar.Position > FVerticalScrollBar.Max - FVerticalScrollBar.PageSize  + 1 then
  begin
    FVerticalScrollBar.Position := FVerticalScrollBar.Max - FVerticalScrollBar.PageSize + 1;
    Exit;
  end;
  SetCaretPos(True);
  Paint;
end;

procedure TMWKEdit.DoOnHorizontalScrollBarChange(Sender: TObject);
begin
  if FHorizontalScrollBar.Position > FHorizontalScrollBar.Max - FHorizontalScrollBar.PageSize  + 1 then
  begin
    FHorizontalScrollBar.Position := FHorizontalScrollBar.Max - FHorizontalScrollBar.PageSize + 1;
    Exit;
  end;
  SetCaretPos(True);
  Paint;
end;

procedure TMWKEdit.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);

  ControlStyle := ControlStyle + [csNeedsBorderPaint];

  with Params do
  begin
    Style := Style or (ES_AUTOHSCROLL or ES_AUTOVSCROLL) or
      BorderStyles[FBorderStyle];
    if Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TMWKEdit.CreateWnd;
begin
  inherited;
  if Assigned(FDropTarget) then
    FDropTarget.Active := FOLEDragDrop;
end;

function TMWKEdit.GetOnKeyPress: TWideKeyPressEvent;
begin
  Result := FOnKeyPress;
end;

procedure TMWKEdit.SetOnKeyPress(const Value: TWideKeyPressEvent);
begin
  FOnKeyPress := Value;
end;

function TMWKEdit.GetOnSelectionChange: TNotifyEvent;
begin
  Result := FOnSelectionChange;
end;

procedure TMWKEdit.SetOnSelctionChange(const Value: TNotifyEvent);
begin
  FOnSelectionChange := Value;
end;

function TMWKEdit.GetParagraphIndent(Paragraph: Integer): Integer;
var
  C, Para: Integer;
begin
  C := 0;
  Para := 0;
  while (C < FItems.Count) and (Para < Paragraph) do
  begin
    if FItems[C].NewParagraph then
      Inc(Para);
    Inc(C);
  end;

  if (C = FItems.Count) then
    Result := FGlobalIndent
  else
    Result := FItems[C].Indent;
end;

function TMWKEdit.GetPlain: Boolean;
begin
  Result := FPlain;
end;

procedure TMWKEdit.SetParagraphIndent(Paragraph, Indent: Integer);
var
  C, Para: Integer;
begin
  C := 0;
  Para := 0;
  while (C < FItems.Count) and (Para < Paragraph) do
  begin
    if FItems[C].NewParagraph then
      Inc(Para);
    Inc(C);
  end;

  if (C = FItems.Count) then
    Exit;

  repeat
    FItems[C].Indent := Indent;
    Inc(C);
  until (C = FItems.Count) or FItems[C].NewParagraph;
end;

procedure TMWKEdit.SetPlain(const Value: Boolean);
var
  C: Integer;
  Span: TSpan;
  Rec: TSpanValues;
begin
  FinishInsert;
  if FPlain = Value then
    Exit;

  if not Value and not FMultiLine then
    Exit;

  FPlain := Value;
  if FPlain then
  begin
    FillChar(Rec, SizeOf(Rec), 0);
    Rec.Color := FFontColor;
    Rec.BackColor := FBackColor;
    for C := 0 to FItems.Count - 1 do
    begin
      Span := FItems[C];
      Rec.IsNewLine := Span.NewLine;
      Span.Values := Rec;
    end;

    ReBuildLines(0, MaxInt);
    FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
    Invalidate;
  end;
end;

procedure TMWKEdit.SetPlainText(const Value: WideString);
var
  XML: IXMLDOMDocument;
begin
  FinishInsert;
  if not IsValidXML(Value, 'mwkedit') then
  begin
    XML := CreateXMLFromPlain(Value, nil);
    InsertXML(XML);
    Exit;
  end;

  XML := CreateFromString(Value);
  if Assigned(XML) and (XML.documentElement.nodeName = 'mwkedit') then
    InsertXML(XML);
end;

function TMWKEdit.GetReadOnly: Boolean;
begin
  Result := FReadOnly;
end;

procedure TMWKEdit.SetReadOnly(const Value: Boolean);
begin
  FReadOnly := Value;
end;

function TMWKEdit.LoadFromMSWord(const FileName: WideString): Integer;
var
  App: Variant;
  S: WideString;
  Cr: TCursor;
begin
  FinishInsert;
  Cr := Screen.Cursor;
  Result := 2;
  try
    Screen.Cursor := crHourGlass;
    try
      App := CreateOleObject('Word.Application');
    except
      Screen.Cursor := Cr;
      Exit;
    end;

    try
      Result := 0;
      try
        App.Documents.Open(FileName);
        App.Selection.WholeStory;
        S := AdjustLineBreaks(App.Selection.Text);
      except
        Result := 1;
      end;
    finally
      App.Quit;
      App := Unassigned;
    end;
    SetXML(S);
  finally
    Screen.Cursor := Cr;
  end;
  FUndoRedo.Clear;
  DoOnSelectionChange;
end;

procedure TMWKEdit.CalcLineCountAndStart;
var
  C: Integer;
begin
  FLineStarts.Clear;
  if FItems.Count > 0 then
    FLineStarts.Add(Pointer(0));

  FLineCount := 1;
  for C := 0 to FItems.Count - 1 do
    if FItems[C].LineBreak then
    begin
      Inc(FLineCount);
      FLineStarts.Add(Pointer(C + 1));
    end;

   FLastEmpty := (FItems.Count = 0) or
     ((FItems.Count > 0) and (FItems.Last.LineBreak));

  FTextHeight := GetHeightToLine(FLineCount);
end;

function TMWKEdit.GetFontColor: TColor;
begin
  Result := FFontColor;
end;

function TMWKEdit.GetFontHeight: Currency;
begin
  Result := FFontHeight;
end;

function TMWKEdit.GetFontName: TFontName;
begin
  Result := FFontName;
end;

function TMWKEdit.GetFontSize: Currency;
begin
  Result := FFontSize;
end;

function TMWKEdit.GetHeightOfCurrentFont: Integer;

  function CreateFont: HFont;
  var
    LogFont: TLogFontW;
    FontName: WideString;
  begin
    FillChar(LogFont, SizeOf(LogFont), 0);
    LogFont.lfHeight := Round(FFontHeight);
    LogFont.lfWidth := 0;
    LogFont.lfEscapement := 0;
    LogFont.lfOrientation := 0;
    if FBold then
      LogFont.lfWeight := FW_BOLD
    else
      LogFont.lfWeight := FW_NORMAL;
    LogFont.lfItalic := Byte(FItalic);
    LogFont.lfUnderline := Byte(FUnderline);
    LogFont.lfStrikeOut := Byte(FStrikeOut);
    logfont.lfCharSet := DEFAULT_CHARSET;
    LogFont.lfOutPrecision := OUT_DEFAULT_PRECIS;
    LogFont.lfClipPrecision := CLIP_DEFAULT_PRECIS;
    LogFont.lfQuality := DEFAULT_QUALITY;
    LogFont.lfPitchAndFamily := DEFAULT_PITCH or FF_DONTCARE;
    FontName := FFontName;

    Move(FontName[1], LogFont.lfFaceName, System.Length(FontName) * SizeOf(WideChar));
    LogFont.lfFaceName[System.Length(FontName)] := #0;
    Result := CreateFontIndirectW(LogFont);
  end;

var
  DC: HDC;
  Font, OldFont: HFont;
  Size: TSize;
begin
  DC := GetDC(0);
  try
    Font := CreateFont;
    OldFont := SelectObject(DC, Font);
    GetTextExtentPoint32W(DC, 'Mp', 2, Size);
    Result := Size.cy;
    SelectObject(DC, OldFont);
    DeleteObject(Font);
  finally
    ReleaseDC(0, DC);
  end;
end;

function TMWKEdit.GetHeightOfLine(const Index: Integer): Integer;
var
  C: Integer;
  Start, Stop: Integer;
begin
  Result := 0;
  if (Index < 0) or (Index >= FLineStarts.Count) then
    Exit;

  Start := GetLineStart(Index);
  if (Index < FLineStarts.Count - 1) then
    Stop := GetLineStart(Index + 1) - 1
  else
    Stop := FItems.Count - 1;

  for C := Start to Stop do
    if (C >= 0) and (C < FItems.Count) then
      Result := Max(Result, FItems[C].Height);
end;

function TMWKEdit.GetHeightToLine(const Line: Integer): Integer;
var
  C: Integer;
begin
  Result := 0;
  for C := 0 to Line do
    Result := Result + GetHeightOfLine(C);
end;

procedure TMWKEdit.SetFixedColor(const Value: Boolean);
begin
  FFixedColor := Value;
end;

procedure TMWKEdit.SetFixedFont(const Value: Boolean);
begin
  FFixedFont := Value;
end;

procedure TMWKEdit.SetFixedSize(const Value: Boolean);
begin
  FFixedSize := Value;
end;

procedure TMWKEdit.SetFontColor(const Value: TColor);
begin
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (FReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;
  FFontColor := Value;

  SetSpanValues(ctColor);
  ReBuildLines(FStartLine, FEndLine);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FFontColor := Value;
end;

procedure TMWKEdit.SetFontHeight(const Value: Currency);
begin
  if FFontHeight = Value then
    Exit;

  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (FReadOnly or FPlain or FFixedSize) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;

  FFontHeight := Value;

  SetSizeFromHeight;

  RealizeFontHeight;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);
end;

procedure TMWKEdit.SetFontName(const Value: TFontName);
begin
  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (FReadOnly or FPlain) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;

  FFontName := Value;
  Canvas.Font.Name := FFontName;

  SetSpanValues(ctFontName);
  ReBuildLines(FStartLine, FEndLine);
  CalcSpellCheck(False);
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);

  FFontName := Value;
end;

procedure TMWKEdit.SetFontSize(const Value: Currency);
var
  DC: HDC;
begin
  if FFontSize = Value then
    Exit;

  if not ((csReading in ComponentState) or
          (csDesigning in ComponentState)) then
  begin
    FinishInsert;
    if (FSelLength > 0) and not (FReadOnly or FPlain or FFixedSize) then
      FUndoRedo.AddActionToUndo(USTYLE, atUndo, Integer(Self), 0);
  end;

  FFontSize := Value;
  DC := GetDC(0);
  FFontHeight := -((FFontSize * GetDeviceCaps(DC, LOGPIXELSY)) / 72);
  ReleaseDC(0, DC);

  RealizeFontHeight;

  FUndoRedo.AddActionToUndo(USTYLE, atRedo, Integer(Self), 0);
end;

{function TMWKEdit.GetRightToLeft: Boolean;
begin
  Result := FRightToLeft;
end;

procedure TMWKEdit.SetRightToLeft(const Value: Boolean);
var
  C: Integer;
begin
  if FRightToLeft = Value then
    Exit;

  FRightToLeft := Value;
  for C := 0 to FItems.Count - 1 do
    FItems[C].RightToLeft := FRightToLeft;
  Invalidate;
end;}

procedure TMWKEdit.SetTabWidth(const Value: Integer);
begin
  if FTabWidth = Value then
    Exit;

  FTabWidth := Max(1, Value);
  ReBuildLines(0, MaxInt);
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
end;

function TMWKEdit.GetShowTags: Boolean;
begin
  Result := FShowTags;
end;

procedure TMWKEdit.SetShowTags(const Value: Boolean);
var
  C: Integer;
begin
  if Value = FShowTags then
    Exit;

  FShowTags := Value;

  for C := 0 to FItems.Count - 1 do
    FItems[C].ShowTags := FShowTags;
  RebuildLines(0, MaxInt);
  Invalidate;
end;

procedure TMWKEdit.SetSizeFromHeight;
var
  DC: HDC;
begin
  DC := GetDC(0);
  FFontSize := -((FFontHeight / GetDeviceCaps(DC, LOGPIXELSY)) * 72);
  ReleaseDC(0, DC);
end;

function TMWKEdit.GetOnToggleInsert: TNotifyEvent;
begin
  Result := FOnToggleInsert;
end;

procedure TMWKEdit.SetOnToggleInsert(const Value: TNotifyEvent);
begin
  FOntoggleInsert := Value;
end;

function TMWKEdit.GetDictionaries: TWideStringList;
begin
  if Assigned(FSpellChecker) then
    Result := FSpellChecker.Dictionaries
  else
    Result := nil;
end;

function TMWKEdit.GetDictionary: Integer;
begin
  if Assigned(FSpellChecker) then
    Result := FSpellChecker.CurrentDictionary
  else
    Result := -1;
end;

procedure TMWKEdit.SetDictionary(const Value: Integer);
begin
  if not Assigned(FSpellChecker) or (Value = FSpellChecker.CurrentDictionary) then
    Exit;

  if FSpellChecker.Dictionaries.Count > 0 then
  begin
    FSpellChecker.CurrentDictionary := Value;

    CalcSpellCheck(True);
  end;
end;

function TMWKEdit.GetCurrentLine: Integer;
begin
  Result := FCurrentLine;
end;

procedure TMWKEdit.AddToXML(const XML: IXMLDOMDocument; Selected: Boolean);

  procedure SetGraphic(const Node: IXMLDOMNode; Span: TGraphicSpan);
  var
    XML: IXMLDOMDocument3;
  begin
    if Span.IsSVG then
    begin
      SetAttribute(Node, 'svg', 'svg');
      XML := Span.GetSVG;
      Node.appendChild(XML.documentElement.cloneNode(True));
    end else
      Node.text := Span.GetAsText(False);
  end;

var
  C, D: Integer;
  Span: TSpan;
  Node: IXMLDOMNode;

  Paragraph: IXMLDOMNode;
  Line: IXMLDOMNode;

  S: WideString;
  L: Integer;

  Range: TSpanRange;
  NewLine: WideChar;
begin
  Paragraph := XML.documentElement.lastChild;

  if Selected then
    Range := CalculateSelectedSpanRange
  else
  begin
    Range.Start := 0;
    Range.Stop := FItems.Count - 1;
  end;

  if not Assigned(Paragraph) then
  begin
    Paragraph := AddChildNode(XML.documentElement, 'p');
    SetXMLAttributes(Paragraph, FItems[Range.Start]);
  end;

  C := Range.Start;

  while C <= Range.Stop do
  begin
    Span := FItems[C];

    if Selected and (Span.SelLength < Span.Length) then
      S := Copy(Span.Text, Span.SelStart, Span.SelLength)
    else
      S := Span.Text;

    L := System.Length(S);

    if (L > 0) and ((S[L] = #13) or (S[L] = #14)) then
    begin
      NewLine := S[L];
      S := Copy(S, 1, L - 1);
    end else
      NewLine := #0;

    Inc(C);
    while (C <= Range.Stop) and Span.CanAdd(FItems[C]) do
    begin
      Span := FItems[C];

      if Selected and (Span.SelLength < Span.Length) then
        S := S + Copy(Span.Text, Span.SelStart, Span.SelLength)
      else
        S := S + Span.Text;

      if (S <> '') and (S[System.Length(S)] in [WideChar(#13), WideChar(#14)]) then
      begin
        NewLine := S[System.Length(S)];
        S := Copy(S, 1, System.Length(S) - 1);
      end;

      Inc(C);
    end;
    Dec(C);

    if Span is TGraphicSpan then
    begin
      Node := AddChildNode(Paragraph, 'g');
      SetGraphic(Node, TGraphicSpan(Span));
      SetXMLAttributes(Node, Span);
      S := '';
    end;

    if Span is TSectionSpan then
    begin
      Node := AddChildNode(Paragraph, 'section');
      SetXMLAttributes(Node, Span);
      S := '';
    end;

    if S <> '' then
    begin
      if not Span.IsTab then
      begin
        Node := AddChildNode(Paragraph, 'span');

        D := CountFirstSpaces(S);
        if D > 0 then
          SetAttribute(Node, 'sb', D);

        D := CountLastSpaces(S);
        if D > 0 then
          SetAttribute(Node, 'sa', D);

        S := Trim(S);
        if S <> '' then
          Node.text := S;
      end else
        Node := AddChildNode(Paragraph, 'tab');

      SetXMLAttributes(Node, Span);
    end;

    if NewLine = #14 then
    begin
      Line := AddChildNode(Paragraph, 'br');
      SetXMLAttributes(Line, Span);
    end;

    if NewLine = #13 {and (C < FItems.Count - 1)} then
    begin
      SetXMLIndent(Paragraph, Span.Indent);
      Paragraph := AddChildNode(XML.documentElement, 'p');
      SetXMLAttributes(Paragraph, Span);
    end;
    Inc(C);
  end;

  if FItems.Count > 0 then
    SetXMLAttributes(Paragraph, FItems[FItems.Count - 1]);
end;

procedure TMWKEdit.BeginRollBack;
begin
  Inc(FRollBack);
end;

procedure TMWKEdit.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TMWKEdit.EndRollBack;
begin
  if FRollBack > 0 then
    Dec(FRollBack);
end;

procedure TMWKEdit.EndUpdate(WithPaint: Boolean);
begin
  if FUpdateCount > 0 then
    Dec(FUpdateCount);
  if (FUpdateCount = 0) and WithPaint then
    Invalidate;
end;

procedure TMWKEdit.Repaint;
begin
  if FUpdateCount = 0 then
    inherited;
end;

function TMWKEdit.GetShowSpellErrors: Boolean;
begin
  Result := FShowSpellErrors;
end;

procedure TMWKEdit.SetShowScrollBars(const Value: Boolean);
begin
  if Value = FShowScrollBars then
    Exit;

  if Value and not FMultiLine then
    Exit;
  FShowScrollBars := Value;

  FHorizontalScrollBar.Visible := FShowHorizontalScrollBar and FShowScrollBars;
  FVerticalScrollBar.Visible := FShowVerticalScrollBar and FShowScrollBars;
  ReBuildLines(0, MaxInt);
  CalcSelection;
  FCurrentSpan := GetCurrentSpanNum(FPosInSpan);
  Invalidate;
end;

procedure TMWKEdit.SetShowSpellErrors(const Value: Boolean);
var
  C: Integer;
begin
  if Value = FShowSpellErrors then
    Exit;

  FShowSpellErrors := Value;
  for C := 0 to FItems.Count - 1 do
    FItems[C].ShowSpellErrors := Value;
  Invalidate;
end;
{$ENDREGION}

{ TMWKEditDropTarget }
{$REGION TMWKDropTarget}

constructor TMWKEditDropTarget.Create(Edit: TMWKEdit);
begin
  inherited Create;
  FEdit := Edit;
  FEditHandle := 0;
  FActive := False;
end;

destructor TMWKEditDropTarget.Destroy;
begin
  Active := False;
  inherited;
end;

function TMWKEditDropTarget.DragEnter(const DataObj: IDataObject;
  GrfKeyState: Integer; PT: TPoint; var dwEffect: Integer): HResult;
begin
  Result := FEdit.SupportsOLEData(DataObj, GrfKeyState, PT, dwEffect);
  if Result = S_OK then
  begin
    Result := FEdit.ModifyOLEDropEffect(GrfKeyState, PT, dwEffect);
    if Result = S_OK then
    begin
      PT := FEdit.ScreenToClient(PT);
      FEdit.ShowDragCaret(PT);
    end;
  end else
    dwEffect := DROPEFFECT_NONE;
end;

function TMWKEditDropTarget.DragLeave: HResult;
begin
  Result := S_OK;
  FEdit.HideDragCaret;
end;

function TMWKEditDropTarget.DragOver(GrfKeyState: Integer; PT: TPoint;
  var dwEffect: Integer): HResult;
begin
  Result := FEdit.ModifyOLEDropEffect(GrfKeyState, PT, dwEffect);
  if Result = S_OK then
  begin
    PT := FEdit.ScreenToClient(PT);
    FEdit.ShowDragCaret(PT);
  end else
  begin
    dwEffect := DROPEFFECT_NONE;
    FEdit.HideDragCaret;
  end;
end;

function TMWKEditDropTarget.Drop(const DataObj: IDataObject;
  GrfKeyState: Integer; PT: TPoint; var dwEffect: Integer): HResult;
begin
  FEdit.HideDragCaret;

  Result := FEdit.SupportsOLEData(DataObj, GrfKeyState, PT, dwEffect);
  if Result = S_OK then
  begin
    Result := FEdit.ModifyOLEDropEffect(GrfKeyState, PT, dwEffect);
    if Result = S_OK then
    try
      Result := FEdit.InsertOLEData(DataObj, GrfKeyState, PT, dwEffect);
    except
      Result := E_FAIL;
      ShowException(ExceptObject, ExceptAddr);
    end;
  end;
end;

function TMWKEditDropTarget.GetActive: Boolean;
begin
  Result := FActive;
end;

procedure TMWKEditDropTarget.SetActive(const Value: Boolean);
begin
  if FActive <> Value then
  begin
    FActive := Value;
    if not Value then
    begin
      OleCheck(RevokeDragDrop(FEditHandle));
      OleCheck(CoLockObjectExternal(Self, False, True));
    end else
    begin
      FEditHandle := FEdit.Handle;
      OleCheck(RegisterDragDrop(FEdit.Handle, Self));
      OleCheck(CoLockObjectExternal(Self, True, False));
    end;
  end;
end;

{$ENDREGION}

{ TFormatEnum }
{$REGION TFormatEnum}

constructor TFormatEnum.Create(const DataObject: IDataObject);
var
  LRecFormat: TFormatETC;
  LifEnum: IEnumFormatETC;
begin
  FFormats := nil;
  if Succeeded(dataObject.EnumFormatEtc(DATADIR_GET, LifEnum)) then
  begin
    while LifEnum.Next(1, LRecFormat, nil) = S_OK do
    begin
      SetLength(FFormats, Succ(Length(FFormats)));
      FFormats[Pred(Length(FFormats))] := LRecFormat;
    end;
  end;
end;

destructor TFormatEnum.Destroy;
begin
  FFormats := nil;
  inherited;
end;

function TFormatEnum.GetFormatETC(const cfFormat: TClipFormat): TFormatETC;
var
  LBoolOK: Boolean;
  LIntLoop: Integer;
begin
  LBoolOK := False;
  if Length(FFormats) > 0 then
    for LIntLoop := 0 to Pred(Length(FFormats)) do
      if FFormats[LIntLoop].cfFormat = cfFormat then
      begin
        LBoolOK := True;
        Result := FFormats[LIntLoop];
        Break;
      end;
  if not LBoolOK then
    FillChar(Result, SizeOf(Result), #$FF);
end;

function TFormatEnum.HasFormat(const cfFormat: TClipFormat): Boolean;
var
  LIntLoop: Integer;
begin
  Result := False;
  if Length(FFormats) > 0 then
    for LIntLoop := 0 to Pred(Length(FFormats)) do
      if FFormats[LIntLoop].cfFormat = cfFormat then
      begin
        Result := True;
        Break;
      end;
end;
{$ENDREGION}

{ TIndentAction }

constructor TIndentAction.Create;
begin
  inherited;
  Indents := TList.Create;
end;

destructor TIndentAction.Destroy;
begin
  Indents.Free;
  inherited;
end;

{ TSpellCheckerList }

procedure TSpellCheckerList.Add(const AName: TSpellChecker;
  AClass: TSpellCheckerClass);
var
  NewRec: PSpellCheckerEntry;
begin
  New(NewRec);
  with NewRec^ do
  begin
    SpellCheckerClass := AClass;
    Name := AName;
  end;
  inherited Add(NewRec);
end;

destructor TSpellCheckerList.Destroy;
var
  C: Integer;
begin
  for C := 0 to Count - 1 do
    Dispose(PSpellCheckerEntry(Items[C]));

  inherited;
end;

function TSpellCheckerList.GetName(const Index: Integer): TSpellChecker;
begin
  if (Index >= 0) and (Index < Count) then
    Result := PSpellCheckerEntry(Items[Index])^.Name
  else
    Result := '';
end;

function TSpellCheckerList.IndexOf(const Name: TSpellChecker): Integer;
begin
  for Result := 0 to Count - 1 do
    if SameText(Name, GetName(Result)) then
      Exit;
  Result := -1;
end;

procedure TSpellCheckerList.Remove(AClass: TSpellCheckerClass);
var
  C: Integer;
  P: PSpellCheckerEntry;
begin
  for C := Count - 1 downto 0 do
  begin
    P := PSpellCheckerEntry(Items[C]);
    if P^.SpellCheckerClass.InheritsFrom(AClass) then
    begin
      Dispose(P);
      Delete(C);
    end;
  end;
end;

initialization
  OleInitialize(nil);
finalization
  OleUninitialize;
  SpellCheckers.Free;
end.
